BglInput = function(filePath) {
  x = read.table(file=filePath, header=FALSE, sep=" ")
  x = as.matrix(x[-1, -(1:2)])
  x = matrix(nrow=nrow(x), data=as.integer(x))
  return(t(x))
}

BglDummySnpIds = function(n) {
  snpStr = rep("", n)
  for (k in 1:n) {
    snpStr[k] = paste("SNP",k, sep="")
  }
  return(snpStr)
}

BglFormatMatrix = function(db, affectStat = NULL, rs = NULL) {
  db = MatrixSlice(db, matType = "IntMat")
  z = matrix(nrow=nrow(db), data=as.character(db))
  z[db == 2] = "?"
  z = t(z)

  if (is.null(rs)) {
    snpStr = BglDummySnpIds(nrow(z))
  } else {
    snpStr = rs
  }

  z = cbind(rep("M", nrow(z)), snpStr, z)
  
	header.str = c("I", "id")
  v1 = rep(1:(nrow(db)/2), each=2)
  for (j in v1) header.str = c(header.str, toString(j))
  
  if (!is.null(affectStat)) {
    if (!is.list(affectStat)) {
      affectStat = list(trait = affectStat)
    }
    
    for (k in 1:length(affectStat)) {
      aff.str = c("A", names(affectStat)[k])
      v = affectStat[[k]]
      for (j in 1:length(v)) {
        aff.str = c(aff.str, toString(v[j]))
      }
      z = rbind(aff.str, z)
    }
  }    
  z = rbind(header.str, z)
  return(z)
}

BglOutput = function(db, file.path, scramble = TRUE, verbose = FALSE,
                     affectStat = NULL, rs = NULL) {
  db = MatrixSlice(db, matType = "IntMat")
  if (scramble) {
    db = RandPhasing(db)
  }
  z = BglFormatMatrix(db, affectStat, rs)
  write.table(z, file=file.path, quote = FALSE, sep = " ",
              row.names = FALSE, col.names = FALSE)
  return(z)
}

BglWrapper = function(db, pathBase = NULL, phasedRefPanel = NULL, rs = NULL,
                          removeIntermediate = TRUE, lowmem = "false",
                          gprobs = "false", tmpDir = NULL,
                          callPrefix = "", dontRun = FALSE,
                          readOnly = FALSE) {
  if (is.null(pathBase)) {
    pathBase = sprintf("bgl_tmp_%d_", as.integer(runif(1) * 100000))
  }
  
  db = MatrixSlice(db, matType = "IntMat")
  nHap = nrow(db)
  
  inpFname = paste(pathBase, ".unphInp", sep = "")
  logFname = paste(pathBase, ".log.log", sep = "")
  inpMat = BglOutput(db, inpFname, rs = rs)

  javaArgs = "-Xmx1000m -jar beagle.jar"
  bglArgs = sprintf("unphased=%s missing=? log=%s.log lowmem=%s gprobs=%s",
                    inpFname, pathBase, lowmem, gprobs)
                    
  phasedRefPanelFnm = NULL
  markersFnm = NULL
  
  if (!is.null(phasedRefPanel)) {
    phasedRefPanelFnm = paste(pathBase, ".refPanel", sep = "")
    
    if (!readOnly) {
      BglOutput(phasedRefPanel, phasedRefPanelFnm, scramble = FALSE, rs = rs)
    }
    bglArgs = paste(bglArgs, " phased=", phasedRefPanelFnm, sep = "")
    
    if (is.null(rs)) {
      snpStr = BglDummySnpIds(nrow(inpMat))
    } else {
      snpStr = rs
    }    
    markersFnm = paste(pathBase, ".markers", sep = "")
    markersMat = cbind(snpStr, 1:length(snpStr),
                       rep(0, length(snpStr)),
                       rep(1, length(snpStr)))
                       
    if (!readOnly) {
      write.table(markersMat, file=markersFnm, quote = FALSE, sep = " ",
                  row.names = FALSE, col.names = FALSE)
    }
    bglArgs = paste(bglArgs, " markers=", markersFnm, sep = "")
  }
  
  if (!is.null(tmpDir)) {
    javaArgs = paste("-Djava.io.tmpdir=", tmpDir, " ", javaArgs, sep = "")
  }

  bglCmd = paste(callPrefix, "java ", javaArgs, " ", bglArgs, sep="")
  if (dontRun) {
    return(bglCmd)
  } else if (!readOnly) {
    system(bglCmd)
  }

  bglOut = paste(inpFname, ".phased", sep="")
  haps = BglInput(bglOut)
  
  while (ncol(haps) > ncol(db)) {
    haps = haps[, -ncol(haps)]
  }
  
  if (removeIntermediate) {
    system(paste("rm", bglOut))
    system(paste("rm ", inpFname, "*", sep = ""))
    system(paste("rm ", logFname, "*", sep = ""))
    if (!is.null(phasedRefPanelFnm)) system(paste("rm", phasedRefPanelFnm))
    if (!is.null(markersFnm)) system(paste("rm", markersFnm))
  }
  return(haps)
}