DISC_DBL_TYPE = "disc-dbl"
PACKED_INT_TYPE = "packed-int"

RecursiveDiscretizePack = function(x, minLen = 2000, maxVal = 2,
                                   minVal = -2, ...) {
  if (is.list(x)) {         
    if (!is.null(x$type) && x$type %in% c(DISC_DBL_TYPE, PACKED_INT_TYPE)) {
      return(x)
    }
    for (k in 1:length(x)) {
      x[[k]] = RecursiveDiscretizePack(x[[k]], minLen, ...)
    }
  } else if (is.integer(x) && length(x) >= minLen) {
    x = PackInts(x)
  } else if (is.double(x) && length(x) >= minLen) {
    i = is.na(x)
    i2 = (x != Inf) & (x != -Inf)
    i2[i] = FALSE
    i = i2
    
    if (sum(i) > 0 && min(x[i]) > minVal && max(x[i]) < maxVal) {
      x = DiscretizeDbl(x, ...)
    }
  }
  return(x)
}
                               

DiscretizeDbl = function(x, discSz = 100, pack = TRUE) {
  if (!is.double(x)) {
    return(x)
  }

  discSz = as.integer(discSz)

  NACode = as.integer(discSz + 3)
  PosInfCode = as.integer(discSz + 2)
  NegInfCode = as.integer(discSz + 1)

  d0 = dim(x)
  len = length(x)
  i0 = as.integer(0)

  xi = rep(NACode, len)
  i = !is.na(x)
  xi[i][x[i] == -Inf] = NegInfCode
  xi[i][x[i] == Inf] = PosInfCode

  i2 = ((x < Inf) & (x > -Inf))
  i2[is.na(x)] = FALSE

  if (sum(i2) == 0) {
    rg = c(0, 0)
  } else {
    rg = range(x[i2])
    x[i2] = (x[i2] - rg[1])
    w = rg[2] - rg[1]
    if (w > 0.0) {
      x[i2] = x[i2] / w
    }

    # We subtract 0.99999 rather than 1 because
    # if we discretize, un-discretize, and discretize
    # again, a rounding error will cause some points to
    # get progressively smaller.  This fudge factor
    # seems to fix this issue

    mlt = (discSz - 0.999999)
    #xi[i2] = as.integer(x[i2] * mlt) + as.integer(1)
    xi[i2] = as.integer(round(x[i2] * mlt)) + as.integer(1)
    xi[i2][xi[i2] > discSz] = discSz
  }

  untrans = seq(from = rg[1], to = rg[2], length = discSz)

  if (pack) {
    xi = PackInts(xi)
  }
  
  return(list(x = xi, dim = d0, length = len,
              NACode = NACode,
              NegInfCode = NegInfCode,
              PosInfCode = PosInfCode,
              range = rg,
              packed = pack,
              untrans = untrans,
              type = DISC_DBL_TYPE))
}

UnDiscretize = function(x) {
  if (!is.list(x) || is.null(x$type) || x$type != DISC_DBL_TYPE) {
    return(x)
  }
  z = x$untrans

  xi = x$x
  if (x$packed) {
    xi = UnpackInts(xi)
  }

  xd = rep(NA, x$length)

  i = ((xi >= 1) & (xi <= length(z)))
  xd[i] = z[xi[i]]

  xd[xi == x$NACode] = NA
  xd[xi == x$NegInfCode] = -Inf
  xd[xi == x$PosInfCode] = Inf

  if (!is.null(x$dim)) {
    dim(xd) = x$dim
  }
  return(xd)
}

PackInts = function(x) {
  if (!is.integer(x)) {
    return(x)
  }

  shft = as.integer(min(x))
  x = x - shft

  m = as.integer(max(x)) + as.integer(1)
  
  sz = 1
  m0 = m
  while (TRUE) {
    if (2e9 / m < m0) {
      break
    } else {
      sz = sz + 1
      m0 = m0 * m
    }
  }
  if (sz == 3 && m < 256) {
    m = as.integer(256)
  }
  i = seq(from = 1, to = length(x), by = sz)
  z = rep(as.integer(0), length(i))

  mlt = as.integer(1)
  for (j in 1:sz) {
    updInds = which(i <= length(x))
    if (length(updInds) == 0) break

    i0 = i[updInds]
    z[updInds] = z[updInds] + mlt * x[i0]

    mlt = mlt * m
    
    i = i + as.integer(1)
  }
  return(list(x = z, m = m, sz = sz, length = length(x),
              shft = shft, type = PACKED_INT_TYPE))
}

UnpackInts = function(x) {
  if (is.null(x$type) || x$type != PACKED_INT_TYPE) {
    return(x)
  }

  z = rep(as.integer(0), x$length)
  sz = x$sz
  m = x$m
  len = x$length
  shft = x$shft
  x = x$x
  i = seq(from = 1, to = len, by = sz)

  mlt = as.integer(1)

  for (j in 1:sz) {
    updInds = which(i <= len)
    if (length(updInds) == 0) break

    i0 = i[updInds]
    z[i0] = as.integer(x[updInds] / mlt) %% m

    mlt = mlt * m
    
    i = i + as.integer(1)
  }
  return(z + shft)
}

DiscretizeFileContents = function(inpPath, outPath = NULL, deleteFile = FALSE, ...) {
  if (is.null(outPath)) {
    outPath = inpPath
  }
  fi = file.info(inpPath)
  if (is.na(fi$size)) {
    return(NULL)
  }

  v = new.env()
  load(file = inpPath, envir = v)
  
  nms = ls(v)
  for (n in nms) {
    v[[n]] = RecursiveDiscretizePack(v[[n]], ...)
  }  
  
  if (deleteFile) {
    print("deleting file")
    cmd = paste("rm", inpPath)
    system(cmd)
  }
  save(list = ls(v), file = outPath, envir = v)
}

DiscretizeTests = function() {
  source("deps.r")
  
  r = runif(100000)
  r[1:3] = c(NA, Inf, -Inf)

  z = runif(10000)
  x = list(x = runif(100000), y = 0)
  i = as.integer(-10) + as.integer(runif(100000) * 100)
  j = as.integer(runif(100000)  < 0.5)

  inpPath = "disc_test1.Rdata"
  outPath = "disc_test2.Rdata"
  save(r, z, x, i, j, file = inpPath)
  DiscretizeFileContents(inpPath, outPath, pack = TRUE)

  v = new.env()
  load(outPath, envir = v)

  if (is.null(v$x$x$type) || v$x$x$type != DISC_DBL_TYPE) {
    stop("failed to discretize")
  }
  if (is.null(v$r$type) || v$r$type != DISC_DBL_TYPE) {
    stop("failed to discretize")
  }

  r2 = UnDiscretize(v$r)
  i2 = UnpackInts(v$i)

  r3 = range(r2[-(1:3)] - r[-(1:3)])

  print(file.info(inpPath)$size / file.info(outPath)$size)


  inpPath = "disc_test1.Rdata"
  outPath = "disc_test2.Rdata"
  save(j, file = inpPath)
  DiscretizeFileContents(inpPath, outPath, pack = TRUE)
  print(file.info(inpPath)$size / file.info(outPath)$size)

  #####################################
  # Test for accumulating rounding errors

  x = rnorm(100)

  for (pack in c(FALSE, TRUE)) {
    x2 = x

    for (k in 1:3) {
      xd = DiscretizeDbl(x2, pack = pack)
      x2 = UnDiscretize(xd)
      plot(x, x2, main = paste(pack, toString(k)))
      abline(a = 0, b = 1)
      abline(h = xd$untrans)
      scan()
    }
  }


  ########################################
  # Test packing and un-packing
  source("deps.r")

  x = as.integer(runif(100) * 110) - as.integer(20)
  xp = PackInts(x)
  x2 = UnpackInts(xp)
  plot(x, x2)
  abline(a = 0, b = 1)

  if (max(abs(x - x2)) > 0) {
    stop("pack/unpack test failed")
  }

  ########################################
  # Test recursive packing
  v = list(ancs = c(), k = 0)
  v$ancs = vector(mode = "list", length = 3)
  for (k in 1:3) {
    v$ancs[[k]]$x = list(ancPrbs = runif(10000),
                         popSz = 3)
    v$ancs[[k]]$y = list(ancPrbs = runif(100),
                         popSz = 3)
  }
  vd = RecursiveDiscretize(v)

  
  ########################################
  # Size test that checks whether this is 
  # worthwhile
  x = array(0, c(10000, 3, 5))
  x[] = runif(length(x))
  x[1:100] = -Inf

  xd = DiscretizeDbl(x, pack = FALSE)
  xd2 = DiscretizeDbl(x, pack = TRUE)
  x2 = UnDiscretize(xd)

  save(x, file = "disc_test1.Rdata")

  save(xd, file = "disc_test2.Rdata")
  save(x2, file = "disc_test3.Rdata")
  save(xd2, file = "disc_test4.Rdata")

  print(file.info("disc_test1.Rdata")$size)
  print(file.info("disc_test2.Rdata")$size)
  print(file.info("disc_test3.Rdata")$size)
  print(file.info("disc_test4.Rdata")$size)
}