#ifndef ARADMX_FWD_BKWD_HPP_
#define ARADMX_FWD_BKWD_HPP_


#include "util.hpp"
#include "aradmx.hpp"

// #define EXACT_INF_DEBUG_CHECKS 1
namespace Aradmx {
namespace FwdBkwd {

typedef const int * Hap;
typedef pair<Hap, Hap> HapPair;

void HapFromSxp(SEXP sxp_obj, Hap * h);

bool CheckMap(int msg_sz, const int * msg_map, int map_len);

class ModelParams {
public:

	ModelParams() {
		drop_lookups_ = NULL;
		drop_lookups_len_ = 0;

	  packed_trees_ = NULL;
	  packed_trees_len_ = 0;
	  tree_posns_ = NULL;
	  tree_posns_len_ = 0;
	  n_clip_trees_ = 0;
	  n_states_ = 0;

	  l2r_map_ = r2l_map_ = NULL;
	  l2r_map_len_ = r2l_map_len_ = 0;
	}

	virtual void FillDropLookups(SEXP drop_lookups_sxp);

	// Fill entries from R data structures
	virtual void Fill(SEXP drop_lookups_sxp, SEXP max_drop_sxp,
	          SEXP win_sz_sxp,              // The state matrix
	          SEXP L2RMap, SEXP R2LMap,     // mappings to absorb messages
	          SEXP anc_transprb_sxp, SEXP anc_priors_sxp, // Recombination and ancestral prior parameters
	          SEXP bin_trees_sxp);    // Parent sets and their observed counts/multinomial parameters

	int * GetMutableTree(int snp_idx, int hid_state, int jump_counter);
	const int * GetTree(int snp_idx, int hid_state, int jump_counter) const;
	const int * GetTreeDebug(int snp_idx, int hid_state, int jump_counter) const;

	bool CheckFwdBkwdMappings() const {
	  if (!CheckMap(n_states_, l2r_map_, l2r_map_len_)) {
	    error("Fill : left-to-right map has an error\n");
	  } else if (!CheckMap(n_states_, r2l_map_, r2l_map_len_)) {
	    error("Fill : right-to-left map has an error\n");
	  }
	  return(true);
	}

	virtual void ObservePhasing(int snp_idx, int phase_or, double * r) const = 0;

	virtual void FirstSnpZero(double * r) = 0;

	virtual void GetAncPrbs(int snp_idx, const double * cur_pttl, double * anc_prbs) const = 0;

	virtual void CondProbs(const Util::IntVec& pop_lab, int snp_idx, double * r) = 0;

	virtual void IncTreeCounts(double obs_wt, int snp_idx, const double * cur_pttl) {
		; //do nothing.  This is used in the EM algorithms
	}

	double * AllocMsgBuf() const {
		return(new double[n_states_]);
	}

	double FwdAlgo(const Util::IntVec& pop_lab, double obs_wt,
			           int checkpt_freq, vector<double *> * fwd_msgs);

	double FwdBkwdAlgo(const Util::IntVec& pop_lab, double obs_wt,
			               bool check_pointed, double * anc_prbs);


	double FwdSumBkwdSample(const Util::IntVec& pop_lab,
			                    bool check_pointed,
			                    Util::IntVec * state_seq);
	double FwdSumBkwdSamplePhaseLik(const Util::IntVec& pop_lab,
			                            const Util::IntVec& phase_seq,
			                            bool check_pointed);

	void DeleteHigherIndexMsgs(int snp_idx, vector<double *> * fwd_msgs);
	void BackfillForwardMsgs(const Util::IntVec& pop_lab,
			                     int snp_idx, vector<double *> * fwd_msgs);
	void DeleteForwardMsgs(vector<double *> * fwd_msgs);

public:
	unsigned char * drop_lookups_;
	int drop_lookups_len_;

  int * packed_trees_;
  int packed_trees_len_;
  int * tree_posns_;
  int tree_posns_len_, n_clip_trees_;

  int n_pops_, n_snp_;

  int win_sz_, max_drop_;

  double * anc_jmp_;
  double * grp_prior_;

  int * l2r_map_, * r2l_map_;
  int l2r_map_len_, r2l_map_len_;

  int n_states_;
  DropLookupIndexer edgecut_indexer_;
  PackedTreeIndexer tree_indexer_;
};

int PhasedTreeTraverse(Hap hap, const int * split_stream, int win_sz);

int TreeTraverse(const int * hap1, const int * hap2, const int * orientations,
		             const int * split_stream, int win_sz, int or_offset);

int TreeTraverseUnphApprox(HapPair hap, int strand_idx,
		                       int ph_or_left, int ph_or_right, int or_change_pt,
		                       const int * split_stream);

double RescaleMsg(double * msg, int sz);

// Absorb the table 'msg' into the table 'pttl' by going through each
// entry in pttl and looking up which elements of msg are compatible with it.
// We sum over these compatible elements and mutliply in the result.

void AbsorbMsg(const double * msg, double * pttl, const int * msg_map, int map_len);

// similar to 'AbsorbMsg' above except we pass in a state-index rather than an entire
// message
void AbsorbSampleMsg(double * msg, int state_idx, double * pttl,
		                 const int * msg_map, int map_len);

int SampleState(double unif_var, const double * pttl, int pttl_sz);

}  // namespace Aradmx {
}  // namespace FwdBkwd {


#endif /*ARADMX_FWD_BKWD_HPP_*/
