#include "aradmx_pack_trees.hpp"
#include "util.hpp"

namespace Aradmx {
namespace PackTrees {

void FillCanary(int group_idx, int snp_idx, int drop_idx, int * split_stream) {
	split_stream[0] = -kCanaryCode;
	split_stream[1] = group_idx;
	split_stream[2] = snp_idx;
	split_stream[3] = drop_idx;
}

void PackCounts(int offset, int * split_stream, const Util::RealVec * count_table,
		            bool count_pack, int * packed_tree, int * open_posn) {
	// This function assumes that if a canary is added to 'packed_tree' then
	// 'packed_trees' or 'open_posn'is already shifted past it; this function
	// does _not_ check for a canary.

  packed_tree[*open_posn] = split_stream[offset];
  if (split_stream[offset] < 0) {

    if (count_pack) {
	    float * packed_counts = (float *)(packed_tree + ((*open_posn)+1));

	    if (count_table) {
	    	int p1 = split_stream[offset+1];
	    	int p2 = split_stream[offset+2];

	    	if (p1 < 0 || p1 >= count_table->size()) {
	    		Util::ThrowError("PackCounts : count-table index out of range\n");
	    	} else if (p2 < 0 || p2 >= count_table->size()) {
	    		Util::ThrowError("PackCounts : count-table index out of range\n");
	    	}

	      packed_counts[0] = (*count_table)[p1];
	      packed_counts[1] = (*count_table)[p2];
	    } else {
	      packed_counts[0] = packed_counts[1] = 0.0;
	    }
    } else {
	    double * leaf_prb = (double *)(packed_tree + ((*open_posn)+1));

	    if (count_table) {
	    	int p1 = split_stream[offset+1];
	    	int p2 = split_stream[offset+2];

	    	if (p1 < 0 || p1 >= count_table->size()) {
	    		Util::ThrowError("PackCounts : count-table index out of range\n");
	    	} else if (p2 < 0 || p2 >= count_table->size()) {
	    		Util::ThrowError("PackCounts : count-table index out of range\n");
	    	}

	      *leaf_prb = (*count_table)[p2]
	                   / ((*count_table)[p1] + (*count_table)[p2]);
	    } else {
	      *leaf_prb = -1.0;
	    }
    }
    *open_posn += 3;
  } else {
  	int p0 = *open_posn;
    *open_posn += 3;

  	for (int k = 0; k < 2; ++k) {
	    packed_tree[p0 + k + 1] = *open_posn;
	    int stream_offset = split_stream[offset + k + 1];
	    PackCounts(stream_offset, split_stream, count_table, count_pack,
	    		       packed_tree, open_posn);
  	}
  }
}

int NumNodes(int offset, int * split_stream) {
	if (offset == 0 && HasCanary(split_stream)) {
		offset = kCanarySize;
	}

  if (split_stream[offset] < 0) {
    return 1;
  } else {
    int r = NumNodes(split_stream[offset+1], split_stream);
    r += NumNodes(split_stream[offset+2], split_stream);
    return(r+1);
  }
}

void CountsToProbs(int offset, double pcount, int * split_stream) {
	if (offset == 0 && HasCanary(split_stream)) {
		offset = kCanarySize;
	}

  if (split_stream[offset] < 0) {
    float * counts = (float *)(split_stream + 1 + offset);
    double c1 = counts[0] + pcount;
    double c2 = counts[1] + pcount;
    double dnm = c1 + c2;
    if (dnm <= 0.0) {
    	Util::ThrowError("CountsToProbs : leaf node probability invalid (1)\n");
    }

    double prb = c2 / dnm;
    if (prb < 0.0 || prb > 1.0) {
    	Util::ThrowError("CountsToProbs : leaf node probability invalid (2)\n");
    }

    *((double *)(split_stream + 1 + offset)) = prb;

  } else {
    CountsToProbs(split_stream[offset + 1], pcount, split_stream);
    CountsToProbs(split_stream[offset + 2], pcount, split_stream);
  }
}

SEXP GetModelsForAnc(SEXP mdl_trees, int anc_idx) {
	if (anc_idx < 0 || anc_idx >= LENGTH(mdl_trees)) {
		Util::ThrowError("GetSplitLists : ancestry index out of range\n");
	}
	return(VECTOR_ELT(mdl_trees, anc_idx));
}

SEXP GetModelElt(SEXP mdl_trees, int anc_idx, int mdl_idx, const char* nm) {
	SEXP v = VECTOR_ELT(mdl_trees, anc_idx);
	if (mdl_idx < 0 || mdl_idx >= LENGTH(v)) {
		Util::ThrowError("GetModelElt : model index out of range\n");
	}
	SEXP u = VECTOR_ELT(v, mdl_idx);
	return(Util::SxpList(u).Sxp(nm));
}

SEXP GetModelEltForSnp(SEXP mdl_trees, int anc_idx, int mdl_idx,
		                     int snp_idx, const char* nm) {
	SEXP v = GetModelElt(mdl_trees, anc_idx, mdl_idx, nm);
	if (snp_idx < 0 || snp_idx >= LENGTH(v)) {
		Util::ThrowError("GetModelEltForSnp : snp index out of range\n");
	}
	return(VECTOR_ELT(v, snp_idx));
}

SEXP GetSplitLists(SEXP mdl_trees, int anc_idx, int mdl_idx) {
	return(GetModelElt(mdl_trees, anc_idx, mdl_idx, "splitLists"));
}


SEXP GetSplitListsForSnp(SEXP mdl_trees, int anc_idx, int mdl_idx,
		                     int snp_idx) {
	return(GetModelEltForSnp(mdl_trees, anc_idx, mdl_idx, snp_idx,
			"splitLists"));
}

SEXP GetLeafCountsForSnp(SEXP mdl_trees, int anc_idx, int mdl_idx,
		                     int snp_idx) {
	return(GetModelEltForSnp(mdl_trees, anc_idx, mdl_idx, snp_idx,
                           "leafCounts"));
}

SEXP GetLeafCounts(SEXP mdl_trees, int anc_idx, int mdl_idx) {
	return(GetModelElt(mdl_trees, anc_idx, mdl_idx, "leafCounts"));
}

void NumPopsModelsSnps(SEXP mdl_trees, int* n_pop, int* n_snp, int* n_mdl,
		                   bool check_model) {
  int n_pops = LENGTH(mdl_trees);
  int n_mdls = LENGTH(GetModelsForAnc(mdl_trees, 0));
  int n_snps = LENGTH(GetSplitLists(mdl_trees, 0, 0));

  if (check_model) {
	  for (int pop_idx = 0; pop_idx < n_pops; ++pop_idx) {
	  	if (n_mdls != LENGTH(GetModelsForAnc(mdl_trees, pop_idx))) {
	  		Util::ThrowError("inconsistent number of trees per model\n");
	  	}
	  	for (int mdl_idx = 0; mdl_idx < n_mdls; ++mdl_idx) {
	  		SEXP v = GetSplitLists(mdl_trees, pop_idx, mdl_idx);
		  	if (n_snps != LENGTH(v)) {
		  		Util::ThrowError("inconsistent number of implied snp's per model\n");
		  	}
        for (int snp_idx = 0; snp_idx < n_snps; ++snp_idx) {
        	v = GetSplitListsForSnp(mdl_trees, pop_idx, mdl_idx, snp_idx);
			  	if (!IS_INTEGER(v)) {
			  		Util::ThrowError("split list is not integer\n");
			  	}
        }
	  	}
	  }
  }
  if (n_snp) *n_snp = n_snps;
  if (n_pop) *n_pop = n_pops;
  if (n_mdl) *n_mdl = n_mdls;
}

}  // namespace AbbnPackTrees
}  // namespace Aradmx

extern "C" {

SEXP AradmxPackInitTrees_RI(SEXP bin_trees_sxp, SEXP tree_posns_sxp,
		                   SEXP count_pack_sxp, SEXP zero_counts_sxp,
		                   SEXP add_canaries_sxp) {
	using namespace Util;
	using namespace Aradmx::PackTrees;

#ifdef CONSOLE_STACK_TRACE
  Util::ConsolePrint("AradmxPackInitTrees_RI : Entering\n");
#endif

  // bin_trees_sxp : list of list of list with indexing

  // We pack into the leaf nodes the two counts of 0 and 1 alleles
  // using two floats if count_pack_sxp is TRUE, and we pack
  // in the probability of the '1' allele as a single double
  // if count_pack_sxp is FALSE

  // If 'add_canaries' is true, we put the following into the stream
  // before the splits :
  //  position 0 : AbbnPackTrees::kCanaryCode   (indicates a canary is present)
  //  position 1 : group-index
  //  position 2 : snp-index
  //  position 3 : drop-index

  int n_pops = 0, n_drop = 0, n_snp = 0;
  NumPopsModelsSnps(bin_trees_sxp, &n_pops, &n_snp, &n_drop, true);
  PackedModelIndexer indexer(bin_trees_sxp);

  if (!IS_INTEGER(tree_posns_sxp)) {
  	Util::ThrowError("AradmxPackInitTrees_RI : tree_posns_sxp "
  			             "must be integer\n");
  }
  int * posns = INTEGER(tree_posns_sxp);
  int posns_len = n_snp * n_pops * n_drop;
  if (LENGTH(tree_posns_sxp) < posns_len) {
  	Util::ThrowError("AradmxPackInitTrees_RI : tree_posns_sxp "
  			             "has wrong length\n");
  }

  int cur_posn = 0;

  bool add_canaries = GetInt(add_canaries_sxp, 0, NULL);
  bool pack_counts = GetInt(count_pack_sxp, 0, NULL);
  bool zero_counts = GetInt(zero_counts_sxp, 0, NULL);

#ifdef CONSOLE_STACK_TRACE
  Util::ConsolePrint("AradmxPackInitTrees_RI : Calculating buffer sizes\n");
#endif

  // loop over each marker, each ancestral population, and each model index
  int num_trees = 0;
  for (indexer.StartIterator(); !indexer.Finished(); indexer.IncIterator(), ++num_trees) {
    // Get the tree structure and counts
  	SEXP v = indexer.CurrentSplitListsForSnp();
  	if (!IS_INTEGER(v)) Util::ThrowError("split list is not integer vector");
  	int * split_stream = INTEGER(v);

    int sz = NumNodes(0, split_stream);

    int cur_idx = indexer.CurrentPackedPositionIndex();
    if (cur_idx >= posns_len) Util::ThrowError("index out of range (1)");
    posns[cur_idx] = cur_posn;
    cur_posn += sz * 3;

    if (add_canaries) cur_posn += kCanarySize;
  }
  if (num_trees != posns_len) {
  	Util::ConsolePrint("AradmxPackInitTrees_RI : dit not loop over all trees\n");
  }

#ifdef CONSOLE_STACK_TRACE
  Util::ConsolePrint("AradmxPackInitTrees_RI : Packing\n");
#endif

  if (cur_posn <= 0) {
  	Util::ThrowError("no model data!");
  }
  SEXP ret_sxp;
  PROTECT(ret_sxp = NEW_INTEGER(cur_posn));

  int * packed_trees = INTEGER(ret_sxp);
  RealVec counts;

  // loop over each marker, each ancestral population, and each model index
  for (indexer.StartIterator(); !indexer.Finished(); indexer.IncIterator()) {
    // Get the tree structure and counts
  	SEXP v = indexer.CurrentSplitListsForSnp();
  	if (!IS_INTEGER(v)) Util::ThrowError("split list is not integer vector");
  	int * split_stream = INTEGER(v);

    int sz = NumNodes(0, split_stream);

    int cur_idx = indexer.CurrentPackedPositionIndex();
    if (cur_idx >= posns_len) Util::ThrowError("index out of range (1)");
    cur_posn = posns[cur_idx];
    int * tr = packed_trees + cur_posn;

    //if (add_canaries) {
    //	FillCanary(anc_idx, snp_idx, mdl_idx, tr);
    //  tr += kCanarySize;
    //}

    int open_posn = 0;
    RealVec * count_table = NULL;

    if (!zero_counts) {
    	counts.clear();
    	SEXP leaf_cts = indexer.CurrentLeafCountsForSnp();
    	if (!IS_NUMERIC(leaf_cts)) {
    		Util::ThrowError("leaf counts must be of type double");
    	}
    	Util::SxpToDblVec(leaf_cts, &counts);
    	count_table = &counts;
    }

    // pass in NULL for the counts so we init to zeroes
    PackCounts(0, split_stream, count_table, pack_counts, tr, &open_posn);
  }
  UNPROTECT(1);

#ifdef CONSOLE_STACK_TRACE
  Util::ConsolePrint("AradmxPackInitTrees_RI : Exiting\n");
#endif
  return(ret_sxp);
}

SEXP AradmxCtTreesToPrbTrees_RI(SEXP bin_trees_sxp, SEXP packed_trees_sxp,
		                            SEXP tree_posns_sxp,
		                            SEXP pcount_sxp) {
	using namespace Util;
	using namespace Aradmx::PackTrees;

  double pcount = GetNumeric(pcount_sxp, 0.0, NULL);
  int n_pops = 0, n_drop = 0, n_snp = 0;
  NumPopsModelsSnps(bin_trees_sxp, &n_pops, &n_snp, &n_drop, true);
  PackedModelIndexer indexer(bin_trees_sxp);

  int * posns = INTEGER(tree_posns_sxp);
  if (LENGTH(tree_posns_sxp) < n_snp * n_pops * n_drop) {
  	Util::ThrowError("CountTreesToProbTrees : tree_posns_sxp has "
  			             "wrong length\n");
  }

  int * packed_trees = INTEGER(packed_trees_sxp);

  for (indexer.StartIterator(); !indexer.Finished(); indexer.IncIterator()) {
    int cur_idx = indexer.CurrentPackedPositionIndex();
    int cur_posn = posns[cur_idx];
    CountsToProbs(0, pcount, packed_trees + cur_posn);
  }

  return(R_NilValue);
}

SEXP AradmxPackDropLookups_RI(SEXP dropLookupsSxp) {
  int n_elt = LENGTH(dropLookupsSxp);
  int * dl = INTEGER(dropLookupsSxp);

  int new_n_elt = (n_elt/4) + ((n_elt % 4 != 0) ? 1 : 0);

  SEXP r_sxp;
  PROTECT(r_sxp = NEW_INTEGER(new_n_elt));
  unsigned char * dlp = (unsigned char *)INTEGER(r_sxp);

  for (int i = 0; i < n_elt; ++i, ++dl, ++dlp) {
    *dlp = (unsigned char)(*dl);
  }
  UNPROTECT(1);
  return(r_sxp);
}

}  // extern "C"
