#include "binary_mat.hpp"

namespace BinaryMatrix {

void FmLeftUpdByRow(const CastableMatI& a, const CastableMatI& msk, int row_idx, int fact_idx,
		                FactoredMatrix * b) {
	double num = 0.0, dnm = 0.0;
	for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
    if (msk.GetElt(row_idx, ci)) {
    	double u = b->GetU(row_idx, fact_idx);
    	double v = b->GetV(ci, fact_idx);
    	double e = a.GetEltDbl(row_idx, ci) - b->GetEltDbl(row_idx, ci) + u * v;
    	num += e * v;
    	dnm += v * v;
    }
	}
	if (dnm > 0.0) {
		b->SetU(row_idx, fact_idx, num / dnm);
	}
}

void FmRightUpdByCol(const CastableMatI& a, const CastableMatI& msk, int col_idx, int fact_idx,
		                FactoredMatrix * b) {
	double num = 0.0, dnm = 0.0;
	for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
    if (msk.GetElt(ri, col_idx)) {
    	double u = b->GetU(ri, fact_idx);
    	double v = b->GetV(col_idx, fact_idx);
    	double e = a.GetEltDbl(ri, col_idx) - b->GetEltDbl(ri, col_idx) + u * v;
    	num += e * u;
    	dnm += u * u;
    }
	}
	if (dnm > 0.0) {
		b->SetV(col_idx, fact_idx, num / dnm);
	}
}


void FmRightUpdTiedByCol(const CastableMatI& a, const CastableMatI& msk, int col_idx, int fact_idx,
		                     FactoredMatrix * b) {
	// We loop over col_idx and col_idx + 1 in the outer loop
	// The function assumes that col_idx is the first column in the pair
	double num = 0.0, dnm = 0.0;
	for (int ci = col_idx; ci < col_idx + 2; ++ci) {
	for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
    if (msk.GetElt(ri, ci)) {
    	double u = b->GetU(ri, fact_idx);
    	double e = a.GetEltDbl(ri, ci) - b->GetEltDbl(ri, ci) + u * b->GetV(ci, fact_idx);
    	num += e * u;
    	dnm += u * u;
    }
	}}
	if (dnm > 0.0) {
		b->SetV(col_idx, fact_idx, num / dnm);
		b->SetV(col_idx + 1, fact_idx, num / dnm);
	}
}


void FmLeftUpd(const CastableMatI& a, const CastableMatI& msk, int fact_idx,
		           FactoredMatrix * b) {
	for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
		FmLeftUpdByRow(a, msk, ri, fact_idx, b);
	}
}

void FmRightUpd(const CastableMatI& a, const CastableMatI& msk, int fact_idx,
		                FactoredMatrix * b) {
	for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
		FmRightUpdByCol(a, msk, ci, fact_idx, b);
	}
}

void FmRightUpdTied(const CastableMatI& a, const CastableMatI& msk, int fact_idx,
		                FactoredMatrix * b) {
	for (int ci = 0, nc = a.ncol(); ci < nc; ci += 2) {
		FmRightUpdTiedByCol(a, msk, ci, fact_idx, b);
	}
}

void SSLeftGrad(const CastableMatI& a, const CastableMatI& msk,
                const FactoredMatrix& b, FactoredMatrix * r) {
	int nr = a.nrow(), nc = a.ncol(), nf = b.nfact();
	for (int fi = 0; fi < nf; ++fi) {
	for (int ri = 0; ri < nr; ++ri) {
		double gr = 0.0;
		for (int ci = 0; ci < nc; ++ci) {
	    if (msk.GetElt(ri, ci)) {
	    	double e = a.GetEltDbl(ri, ci) - b.GetEltDbl(ri, ci);
	    	gr += e * b.GetV(ci, fi);
	    }
		}
		r->SetU(ri, fi, -2.0 * gr);
	}}
}

void SSRightGrad(const CastableMatI& a, const CastableMatI& msk,
                     const FactoredMatrix& b, FactoredMatrix * r) {
	int nr = a.nrow(), nc = a.ncol(), nf = b.nfact();
	for (int fi = 0; fi < nf; ++fi) {
	for (int ci = 0; ci < nc; ++ci) {
		double gr = 0.0;
		for (int ri = 0; ri < nr; ++ri) {
	    if (msk.GetElt(ri, ci)) {
	    	double e = a.GetEltDbl(ri, ci) - b.GetEltDbl(ri, ci);
	    	gr += e * b.GetU(ri, fi);
	    }
		}
		r->SetV(ci, fi, -2.0 * gr);
	}}
}

// This function is used for the accelerated subspace algorithm
void SSLeftSqrs(const CastableMatI& a, const CastableMatI& msk,
                const FactoredMatrix& b, FactoredMatrix * r) {
	int nr = a.nrow(), nc = a.ncol(), nf = b.nfact();
	for (int fi = 0; fi < nf; ++fi) {
	for (int ri = 0; ri < nr; ++ri) {
		double sq = 0.0;
		for (int ci = 0; ci < nc; ++ci) {
	    if (msk.GetElt(ri, ci)) {
	    	double e = b.GetV(ci, fi);
	    	sq += e * e;
	    }
		}
		r->SetU(ri, fi, sq);
	}}
}

// This function is used for the accelerated subspace algorithm
void SSRightSqrs(const CastableMatI& a, const CastableMatI& msk,
                 const FactoredMatrix& b, FactoredMatrix * r) {
	int nr = a.nrow(), nc = a.ncol(), nf = b.nfact();
	for (int fi = 0; fi < nf; ++fi) {
	for (int ci = 0; ci < nc; ++ci) {
		double sq = 0.0;
		for (int ri = 0; ri < nr; ++ri) {
	    if (msk.GetElt(ri, ci)) {
	    	double e = b.GetU(ri, fi);
	    	sq += e * e;
	    }
		}
		r->SetV(ci, fi, sq);
	}}
}

double FmSqrErr(const CastableMatI& a, const CastableMatI& msk,
                const FactoredMatrix& b) {
	double r = 0.0;
	int nr = a.nrow(), nc = a.ncol();
	for (int ci = 0; ci < nc; ++ci) {
	for (int ri = 0; ri < nr; ++ri) {
    if (msk.GetElt(ri, ci)) {
    	double e = a.GetEltDbl(ri, ci) - b.GetEltDbl(ri, ci);
    	r += e * e;
    }
	}}
	return(r);
}

}  // namespace


extern "C" {

	SEXP MaskedPcaLRUpdate(SEXP mat_sxp, SEXP msk_sxp, SEXP fact_idx_sxp,
			                   SEXP fm_sxp, SEXP tied_cols_sxp) {
	  using namespace BinaryMatrix;
	  using namespace Util;

	  int fact_idx = GetInt(fact_idx_sxp, 0, NULL);
	  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(mat_sxp));
	  PointerDeleter<CastableMatI> msk(OwningCastableMatFactory(msk_sxp));
	  PointerDeleter<FactoredMatrix> fm(new FactoredMatrix(fm_sxp));

	  if (m1.is_null() || msk.is_null() || fm.is_null()) {
	    error("Failed to load matrices\n");
	  }

	  FmLeftUpd(*m1.get(), *msk.get(), fact_idx, fm.get());
	  if (GetInt(tied_cols_sxp, 0, NULL)) {
	    FmRightUpdTied(*m1.get(), *msk.get(), fact_idx, fm.get());
	  } else {
	  	FmRightUpd(*m1.get(), *msk.get(), fact_idx, fm.get());
	  }
	  return(fm_sxp);
	}

	SEXP MaskedPcaSubSpaceGrad(SEXP mat_sxp, SEXP msk_sxp, SEXP fm_sxp,
			                       SEXP fm_grad_sxp) {
	  using namespace BinaryMatrix;
	  using namespace Util;

	  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(mat_sxp));
	  PointerDeleter<CastableMatI> msk(OwningCastableMatFactory(msk_sxp));
	  PointerDeleter<FactoredMatrix> fm(new FactoredMatrix(fm_sxp));
	  PointerDeleter<FactoredMatrix> fm_grad(new FactoredMatrix(fm_grad_sxp));

	  if (m1.is_null() || msk.is_null() || fm.is_null()) {
	    error("Failed to load matrices\n");
	  }
		SSLeftGrad(*m1.get(), *msk.get(), *fm.get(), fm_grad.get());
		SSRightGrad(*m1.get(), *msk.get(), *fm.get(), fm_grad.get());
		return(fm_grad_sxp);
	}

	SEXP MaskedPcaSubSpaceSqrs(SEXP mat_sxp, SEXP msk_sxp, SEXP fm_sxp,
			                       SEXP fm_grad_sxp) {
	  using namespace BinaryMatrix;
	  using namespace Util;

	  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(mat_sxp));
	  PointerDeleter<CastableMatI> msk(OwningCastableMatFactory(msk_sxp));
	  PointerDeleter<FactoredMatrix> fm(new FactoredMatrix(fm_sxp));
	  PointerDeleter<FactoredMatrix> fm_grad(new FactoredMatrix(fm_grad_sxp));

	  if (m1.is_null() || msk.is_null() || fm.is_null()) {
	    error("Failed to load matrices\n");
	  }
		SSLeftSqrs(*m1.get(), *msk.get(), *fm.get(), fm_grad.get());
		SSRightSqrs(*m1.get(), *msk.get(), *fm.get(), fm_grad.get());
		return(fm_grad_sxp);
	}

	SEXP MaskedPcaSqrErr(SEXP mat_sxp, SEXP msk_sxp, SEXP fm_sxp) {
	  using namespace BinaryMatrix;
	  using namespace Util;

	  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(mat_sxp));
	  PointerDeleter<CastableMatI> msk(OwningCastableMatFactory(msk_sxp));
	  PointerDeleter<FactoredMatrix> fm(new FactoredMatrix(fm_sxp));

	  if (m1.is_null() || msk.is_null() || fm.is_null()) {
	    error("Failed to load matrices\n");
	  }

	  SEXP r_sxp = NEW_NUMERIC(1);
		REAL(r_sxp)[0] = FmSqrErr(*m1.get(), *msk.get(), *fm.get());
		return(r_sxp);
	}

}  // extern "C"
