#ifndef UTIL_HPP__
#define UTIL_HPP__

#include <R.h>
#include <Rdefines.h>
#include <Rmath.h>
#include <R_ext/Print.h>

#include <vector>

using namespace std;

namespace Util {

void LogToFile(const char * fname, const char * str, int indent_amt = 0);
void LogToFileSprintf(const char * fname, int indent_amt, const char * format,
		                  ...);

void ConsolePrint(const char * format, ...);
void ThrowError(const char * format, ...);

__inline void ZeroBit(int idx, int * x) {
  *x &= ~(1 << idx);
}

template<class T> class PointerDeleter {
public:
  PointerDeleter(T * ptr) : ptr_(ptr) { }
  T * get() { return(ptr_); }
  bool is_null() { return(ptr_ == NULL); }

  ~PointerDeleter() {
    if (ptr_) {
      delete ptr_;
    }
  }
private:
  T * ptr_;
};

template<class T> class VectorEltDeleter {
public:
	VectorEltDeleter(vector<T *>& v) : v_(v) { }
  ~VectorEltDeleter() {
  	typename vector<T *>::iterator v_iter;
  	for (v_iter = v_.begin(); v_iter != v_.end(); ++v_iter) {
  		if (*v_iter) {
  			delete *v_iter;
  			*v_iter = NULL;
  		}
  	}
  }
private:
  vector<T *>& v_;
};



template<class T> void RemoveRange(int lwr, int upr, vector<T> * seq) {
  vector<T> v;
  for (int i = 0, n = seq->size(); i < n; ++i) {
    if (i < lwr || i >= upr) {
      v.push_back((*seq)[i]);
    }
  }
  seq->swap(v);
}

template<class T> void ExtractRange(int lwr, int upr, vector<T> * seq) {
  vector<T> v;
  for (int i = 0, n = seq->size(); i < n; ++i) {
    if (i >= lwr && i < upr) {
      v.push_back((*seq)[i]);
    }
  }
  seq->swap(v);
}

template<class T> void PushVector(const vector<T>& v, vector<T> * seq) {
  for (int i = 0, n = v.size(); i < n; ++i) {
    seq->push_back(v[i]);
  }
}

template<class T> T VecSum(const vector<T>& v) {
	T r = 0;
	typename vector<T>::const_iterator i = v.begin();
  for (; i != v.end(); ++i) {
  	r += *i;
  }
  return(r);
}

template<class T> void VecScale(T scale_amt, vector<T> * v) {
	typename vector<T>::iterator i = v->begin();
  for (; i != v->end(); ++i) {
  	*i *= scale_amt;
  }
}

template<class T> T DotProd(const vector<T>& v, const vector<T>& v2) {
	T r = 0;
	typename vector<T>::const_iterator i = v.begin();
	typename vector<T>::const_iterator j = v2.begin();
  for (; i != v.end(); ++i, ++j) {
  	r += (*i) * (*j);
  }
  return(r);
}

typedef pair<double, double> RealPair;
typedef pair<int, int> IntPair;
typedef vector<double> RealVec;
typedef vector<RealVec> RealVec2;
typedef vector<RealVec2> RealVec3;

typedef vector<int> IntVec;
typedef vector<IntVec> IntVec2;
typedef vector<IntVec2> IntVec3;


int GetInt(SEXP p, int default_val, int* err_code);

double GetNumeric(SEXP p, double default_val, int* err_code);

SEXP ListEltByName(SEXP vec, const char * str, int * elt_idx);

void SaveIntVecToSxp(SEXP vec, int idx, const IntVec& v);
void SaveDblVecToSxp(SEXP vec, int idx, const RealVec& v);

bool SxpToIntVec(SEXP vec, IntVec* v);

void SxpVecEltToIntVec(SEXP vec, int idx, IntVec* v);
void SxpVecEltToIntVecs(SEXP vec, int idx, IntVec2* v);

bool SxpToDblVec(SEXP vec, RealVec* v);

void SxpVecEltToDblVec(SEXP vec, int idx, RealVec* v);

void SxpVecEltToDblVecs(SEXP vec, int idx, RealVec2* v);
int ListEltIdxByName(SEXP v, const char * name);

bool IsMatrix(SEXP x);
void GetMatrixDim(SEXP m, int * n_row, int * n_col);
void SetMatrixDim(SEXP m, int n_row, int n_col);

void DblOrDie(SEXP sxp_obj, const char * nm);
void IntOrDie(SEXP sxp_obj, const char * nm);

SEXP DeepCopy(SEXP sxp_obj);

class SxpList {
public:
	SxpList(SEXP sxp) : sxp_(sxp) { }
	int Index(const char * nm) const {
		return(ListEltIdxByName(sxp_, nm));
	}
	int Int(const char * nm) const {
		SEXP sxp = ListEltByName(sxp_, nm, NULL);
		if (sxp == R_NilValue) {
			error("does not have name %s!", nm);
		}
		if (IS_NUMERIC(sxp)) {
			return((int)REAL(sxp)[0]);
		} else if (IS_LOGICAL(sxp)) {
			return((int)LOGICAL(sxp)[0]);
		} else if (IS_INTEGER(sxp)) {
			return((int)INTEGER(sxp)[0]);
		} else {
			error("Element [%s] cannot be converted to int!", nm);
		}
		return(NULL);
	}
	double Dbl(const char * nm) const {
		SEXP sxp = ListEltByName(sxp_, nm, NULL);
		if (sxp == R_NilValue) {
			error("does not have name %s!", nm);
		}
		if (IS_NUMERIC(sxp)) {
			return(REAL(sxp)[0]);
		} else if (IS_LOGICAL(sxp)) {
			return((double)LOGICAL(sxp)[0]);
		} else if (IS_INTEGER(sxp)) {
			return((double)INTEGER(sxp)[0]);
		} else {
			error("Element [%s] cannot be converted to double!", nm);
		}
		return(NULL);
	}
	int * IntVec(const char * nm) const {
		SEXP sxp = ListEltByName(sxp_, nm, NULL);
		if (sxp == R_NilValue) {
			error("does not have name %s!", nm);
		}
		if (IS_LOGICAL(sxp)) {
			return((int *)LOGICAL(sxp));
		} else if (IS_INTEGER(sxp)) {
			return(INTEGER(sxp));
		} else {
			error("Element [%s] cannot be converted to int!", nm);
		}
		return(NULL);
	}
	double * DblVec(const char * nm) const {
		SEXP sxp = ListEltByName(sxp_, nm, NULL);
		if (sxp == R_NilValue) {
			error("does not have name %s!", nm);
		}
		if (IS_NUMERIC(sxp)) {
			return(REAL(sxp));
		} else {
			error("Element [%s] cannot be converted to double!", nm);
		}
		return(NULL);
	}
	int Length(const char * nm) const {
		return(LENGTH(ListEltByName(sxp_, nm, NULL)));
	}
	SEXP Sxp(const char * nm) const {
		SEXP sxp = ListEltByName(sxp_, nm, NULL);
		if (sxp == R_NilValue) {
			error("does not have name %s!", nm);
		}
		return(sxp);
	}
private:
	SEXP sxp_;
};

}

#endif