MakeTimer = function(timeFreq) {
  t0 = proc.time()[[3]]
  r = list(startTime = t0, numTicks = 0, lastTick = t0, tickFreq = timeFreq)
  return(r)
}

TimerSetLastTick = function(timerObj) {
  timerObj$lastTick = proc.time()[[3]]
  return(timerObj)
}

TimerCheckPrint = function(timerObj, ...) {
  t0 = proc.time()[[3]]
  
  if (t0 - timerObj$lastTick > timerObj$tickFreq) {
    print(...)
    timerObj$lastTick = t0
    timerObj$numTicks = timerObj$numTicks + 1
  }
  return(timerObj)
}

FindArg = function(argName = "", defaultVal = NULL, castFunc = NULL) {
  v = commandArgs(trailingOnly=T)
  for (k in seq(length = length(v))) {
    nm = substr(v[k], 1, nchar(argName))
    if (nm == argName) {
      s0 = strsplit(v[k], "=")
      r = s0[[1]][[2]]
      if (!is.null(castFunc)) {
        r = castFunc(r)
      }
      return(r)
    }
  }
  
  if (!is.null(castFunc)) {
    defaultVal = castFunc(defaultVal)
  }  
  return(defaultVal)
}

FindArgList = function(argName = "", defaultVal = NULL, delim = "\\|",
                       castFunc = NULL) {
  r = FindArg(argName = argName, defaultVal = defaultVal, castFunc = NULL)
  r2 = strsplit(r, delim)[[1]]
  if (!is.null(castFunc)) r2 = sapply(r2, castFunc, USE.NAMES = FALSE)
  return(r2)
}

FindIntArg = function(argName, defaultVal = NULL) {
  return(FindArg(argName, defaultVal = defaultVal, castFunc = as.integer))
}

FindDblArg = function(argName, defaultVal = NULL) {
  return(FindArg(argName, defaultVal = defaultVal, castFunc = as.double))
}

FileExists = function(fnm) {
  return(!is.na(file.info(fnm)$size))
}

RandRow = function(x) {
  if (is.null(x) || !is.matrix(x)) return(NULL)
  if (nrow(x) == 0) return(c())
  
  i = sample(1:nrow(x), 1)
  return(x[i,])
}

IsWindows = function() {
  return(.Platform$OS.type == "windows")
}

IsIncreasing = function(x, strict = FALSE) {
  n = length(x)
  if (strict) {
    return(sum(x[-1] <= x[-n]) == 0)
  } else {
    return(sum(x[-1] < x[-n]) == 0)
  }
}

EvenSubset = function(x) {
  i = 1:length(x)
  i = i[(i %% 2) == 0]
  return(x[i])
}

GetFunctions = function(env = globalenv()) {
  # Gets the source and names of all functions in the environment
  vars = ls(env)
  
  v = vector(mode = "list")
  isFunc = rep(FALSE, length(vars))
  for (j in 1:length(vars)) {
    if (is.function(env[[vars[j]]])) {
      v[[vars[j]]] = env[[vars[j]]]
    }
  }
  return(v)
}

PlotTextSeq = function(x, ylim = NULL, nc = 80, nr = 20, main = "") {
  mainMsg = main
  while (nchar(mainMsg) < nc / 2) {
    mainMsg = paste(" ", mainMsg, sep = "")
  }
  print(mainMsg)

  if (is.null(ylim)) ylim = range(x)

  v = rep(" ", nc)
  for (k in 1:nr) {
    delt = (ylim[2] - ylim[1]) / nr
    
    i = (x > delt * (nr - k) + ylim[1])
    i = i & (x < delt * (nr + k) + ylim[1])

    i = as.integer(nc * which(i) / length(x))
    i = sort(unique(i))

    if (length(i) > 0) {
      i = i[(i >= 1) & (i <= nc)]
      v[i] = "*"
    }
    print(paste(v, collapse = ""))
  }
}

PlotTextScatter = function(x, y, xlim = NULL, pch = "*",
                           ylim = NULL, nc = 80, nr = 20, main = "") {
  mainMsg = main
  while (nchar(mainMsg) < nc / 2) {
    mainMsg = paste(" ", mainMsg, sep = "")
  }
  print(mainMsg)

  if (is.null(ylim)) ylim = range(y)
  if (is.null(xlim)) xlim = range(x)
  
  if (length(pch) < length(x)) {
    pch = rep(pch, length(x))
  }
  
  xbd = seq(from = xlim[1], to = xlim[2], length = (nc+1))
  ybd = seq(from = ylim[1], to = ylim[2], length = (nr+1))

  m = matrix(nrow = nr, ncol = nc, data = 0)
  
  for (k in 1:nr) {
    i1 = (y >= ybd[k]) & (y <= ybd[k+1])
    for (j in 1:nc) {     
      i2 = (x >= xbd[j]) & (x <= xbd[j+1])
      idx = which(i1 & i2)
      if (length(idx) > 0) m[k, j] = sample(rep(idx, each = 2), 1)
    }
    v = rep(" ", nc)
    i = (m[k,] > 0)
    v[i] = pch[m[k, i]]
    print(paste(v, collapse = ""))
  }
}

OverlayHistogram = function(vals, cols, bgcol = NULL, alpha = 0.2,
                            sharedBreaks = TRUE, breaks = NULL, ...) {
  h0 = hist(vals, plot = FALSE, breaks = breaks)

  ylim = range(h0$density)

  unqCols = unique(cols)
  histCols = unqCols
  histObjs = vector(mode = "list")

  for (j in 1:length(unqCols)) {
    if (sharedBreaks) {
      h1 = hist(vals[cols == unqCols[j]], plot = FALSE, breaks = h0$breaks)
    } else {
      h1 = hist(vals[cols == unqCols[j]], plot = FALSE, breaks = breaks)
    }
    h1$counts = h1$density

    ylim = range(ylim, h1$density)
    histObjs[[j]] = h1
    histCols[j] = rgb(t(col2rgb(unqCols[j])) / 255, alpha = alpha)
  }

  if (is.null(bgcol)) {
    plot(range(vals), ylim, col = "white", ...)
  } else {
    h0 = hist(vals, probability = TRUE, col = bgcol,
              bg = bgcol, ylim = ylim, ...)
  }

  a0 = -45
  for (j in 1:length(unqCols)) {
    plot(histObjs[[j]], col = histCols[j], add = TRUE, angle = a0, ...)
    a0 = a0 + 45
  }
}

OverlayHistogramTest = function() {
  x1 = runif(1000)
  x2 = rnorm(1000) + 1
  x3 = rnorm(1000) - 1

  x = c(x1, x2, x3)
  cv = rep(c("blue", "green", "red"), each = 1000)

  par(mfrow = c(1, 2))
  OverlayHistogram(x, cv, breaks = 30)
  OverlayHistogram(x, cv, sharedBreaks = FALSE, breaks = 30)
}

ReadIntMat = function(fnm, ...) {
  z = read.table(file = fnm, ...)
  z = as.matrix(z)
  z = matrix(nrow = nrow(z), data = as.integer(z))
  return(z)
}

ConcatFiles = function(fileNames, tmpFile1, tmpFile2, outFile,
                       verbose = FALSE, recursive = TRUE) {
  # concatenates many files in O(n log n) time or just calls 'cat'
  # once if the number of files is less than 4
  if (length(fileNames) == 1) {
    if (verbose) print(paste("copying", fileNames[1], "to", outFile))
    system(paste("cp", fileNames[1], outFile))
    return(TRUE)
    
  } else if (length(fileNames) <= 4) {
    cmd = "cat"
    for (j in 1:length(fileNames)) {
      cmd = paste(cmd, fileNames[j])
    }
    cmd = paste(cmd, ">", outFile)
    if (verbose) print(cmd)
    system(cmd)
    return(TRUE)
    
  } else if (recursive) {
    halfPt = as.integer(length(fileNames)/2)
    f1 = fileNames[1:halfPt]
    f2 = fileNames[-(1:halfPt)]
    
    ConcatFiles(f1, paste(tmpFile1, "0", sep = ""),
                paste(tmpFile2, "1", sep = ""), tmpFile1,
                verbose = verbose, recursive = recursive)
    ConcatFiles(f2, paste(tmpFile2, "0", sep = ""),
                paste(tmpFile2, "1", sep = ""), tmpFile2,
                verbose = verbose, recursive = recursive)
                
    cmd = paste("cat", tmpFile1, tmpFile2, ">", outFile)
    if (verbose) print(cmd)
    system(cmd)
                       
  } else {

    NextIdx = function(i) {
      if (i == 1) return(2)
      if (i == 2) return(1)
    }

    tmpIdx = 1
    tmpFiles = c(tmpFile1, tmpFile2)

    system(paste("cat", fileNames[1], fileNames[2], ">", tmpFiles[tmpIdx]))

    for (fileIdx in 3:length(fileNames)) {
      if (verbose) print(paste("cat", tmpFiles[tmpIdx], "to", fileNames[fileIdx]))

      nextIdx = NextIdx(tmpIdx)
      system(paste("cat", tmpFiles[tmpIdx], fileNames[fileIdx], ">", tmpFiles[nextIdx]))
      tmpIdx = nextIdx
    }
    system(paste("cp", tmpFiles[tmpIdx], outFile))
    system(paste("rm", tmpFile1))
    system(paste("rm", tmpFile2))
  }
}

  