#include "aradmx_fwd_bkwd.hpp"

// #define EXACT_INF_DEBUG_CHECKS 1
namespace Aradmx {
namespace FwdBkwd {

void HapFromSxp(SEXP sxp_obj, Hap * h) {
	*h = INTEGER(sxp_obj);
}

void ModelParams::FillDropLookups(SEXP drop_lookups_sxp) {
	if (drop_lookups_sxp == R_NilValue) {
		drop_lookups_ = NULL;
		drop_lookups_len_ = 0;
	} else {
	  drop_lookups_ = (unsigned char *)INTEGER(drop_lookups_sxp);
	  drop_lookups_len_ = 4 * LENGTH(drop_lookups_sxp);
	}

  if (drop_lookups_len_ < max_drop_ * n_pops_ * n_snp_) {
  	Util::ThrowError("FillDropLookups : drop_lookups_len_ is wrong\n");
  }
}

void ModelParams::Fill(SEXP drop_lookups_sxp, SEXP max_drop_sxp,
                       SEXP win_sz_sxp, SEXP l2r_map_sxp, SEXP r2l_map_sxp,
                       SEXP anc_transprb_sxp, SEXP anc_priors_sxp,
                       SEXP bin_trees_sxp) {
	using namespace Util;

  int win_sz = GetInt(win_sz_sxp, 0, NULL);

  // The tree structure for population X at snp Y is held in  binTrees[[ X ]][[ Z ]][[ 1 ]][[ Y ]]
  // The leaf counts are in binTrees[[ X ]][[ Z ]][[ 2 ]][[ Y ]]
  // where 'Z' indexes the amount of edge clipping applied.

  SxpList bt(bin_trees_sxp);
  SEXP trees_sxp = bt.Sxp("trees");
  if (trees_sxp == R_NilValue) {
  	Util::ThrowError("ModelParams::Fill : bin_trees_sxp has no element "
  			             "named 'trees'\n");
  }

  int n_snp = bt.Int("nSnp");
  int n_pops = bt.Int("nPop");

  if (LENGTH(anc_priors_sxp) != n_pops) {
  	Util::ThrowError("n_pops != LENGTH(anc_priors_sxp) (%d and %d)\n",
  			             n_pops, LENGTH(anc_priors_sxp));
  }

  n_snp_ = n_snp;
  win_sz_ = win_sz;
  n_pops_ = n_pops;
  max_drop_ = GetInt(max_drop_sxp, 0, NULL);

  IntOrDie(l2r_map_sxp, "l2r_map_sxp");
  IntOrDie(r2l_map_sxp, "r2l_map_sxp");
  DblOrDie(anc_priors_sxp, "anc_priors_sxp");
  DblOrDie(anc_transprb_sxp, "anc_transprb_sxp");
  IntOrDie(trees_sxp, "trees_sxp");

  grp_prior_ = REAL(anc_priors_sxp);
  anc_jmp_ = REAL(anc_transprb_sxp);
  l2r_map_ = INTEGER(l2r_map_sxp);
  r2l_map_ = INTEGER(r2l_map_sxp);
  l2r_map_len_ = LENGTH(l2r_map_sxp);
  r2l_map_len_ = LENGTH(r2l_map_sxp);

  // The trees are in a packed format
  packed_trees_ = (trees_sxp == R_NilValue) ? NULL : INTEGER(trees_sxp);
  packed_trees_len_ = (trees_sxp == R_NilValue) ? 0 : LENGTH(trees_sxp);
  tree_posns_ = bt.IntVec("posns");
  tree_posns_len_ = bt.Length("posns");
  n_clip_trees_ = bt.Int("numClipTrees");

  edgecut_indexer_.Init(n_pops_, n_snp_, max_drop_);
  tree_indexer_.Init(n_pops_, n_snp_, n_clip_trees_);

  if (LENGTH(anc_transprb_sxp) < 2 * n_snp_) {
  	Util::ThrowError("Fill : jump probabilities wrong length "
  			             "(is %d and should be %d)\n",
 	  		             LENGTH(anc_transprb_sxp), 2 * n_snp_);
  } else if (LENGTH(anc_priors_sxp) != n_pops_) {
  	Util::ThrowError("Fill : ancestral prior wrong length\n");
  }
  this->FillDropLookups(drop_lookups_sxp);
}

const int * ModelParams::GetTree(int snp_idx, int hid_state, int jump_counter) const {
//#ifdef EXACT_INF_DEBUG_CHECKS
	return(this->GetTreeDebug(snp_idx, hid_state, jump_counter));
//#endif
  int drop_idx = 0;
  if (jump_counter < win_sz_ - 1) {
    drop_idx = drop_lookups_[edgecut_indexer_.FlatIndex(
    		                                  hid_state, snp_idx, jump_counter)];
  }
  int cur_posn = tree_posns_[tree_indexer_.PackedPositionIndex(
  		                                         hid_state, drop_idx, snp_idx)];
  return(packed_trees_ + cur_posn);
}

const int * ModelParams::GetTreeDebug(int snp_idx, int hid_state, int jump_counter) const {
	// similar to ModelParams::GetTree above but checks array bounds
	int drop_idx = 0;
  if (jump_counter < win_sz_ - 1) {
  	int i = edgecut_indexer_.FlatIndex(hid_state, snp_idx, jump_counter);

  	if (i < 0 || i >= drop_lookups_len_) {
  		Util::ThrowError("GetTree : drop_lookup index out of range\n");
  	}
    drop_idx = drop_lookups_[i];
  }

  if (drop_idx < 0 || drop_idx >= n_clip_trees_) {
  	Util::ThrowError("GetTree : drop_idx = %d\n", drop_idx);
  }
  int t_idx = tree_indexer_.PackedPositionIndex(hid_state, drop_idx, snp_idx);

  if (t_idx < 0 || t_idx >= tree_posns_len_) {
  	Rprintf("drop_idx = %d (of %d/%d), snp_idx = %d, hid_state = %d (of %d)\n",
  			    drop_idx, max_drop_, n_clip_trees_, snp_idx, hid_state, n_pops_);
  	Util::ThrowError("GetTree : tree position index out of range "
  			             "(%d must be in [0, %d)\n",
  		             	 t_idx, tree_posns_len_);
  }
  int cur_posn = tree_posns_[t_idx];

  if (cur_posn < 0 || cur_posn >= packed_trees_len_) {
  	Util::ThrowError("GetTree : tree position out of range\n");
  }

  const int * tr = packed_trees_ + cur_posn;
  return(tr);
}


double ModelParams::FwdAlgo(const Util::IntVec& pop_lab, double obs_wt,
		                        int checkpt_freq, vector<double *> * fwd_msgs) {
	if (n_states_ == 0) error("FwdAlgo : n_states_ = 0\n");

  double scale_fact = 0.0;

  if (fwd_msgs == NULL) {
  	Util::ThrowError("FwdAlgo : fwd_msgs is NULL\n");
  } else if (fwd_msgs->size() < n_snp_) {
  	Util::ThrowError("FwdAlgo : buffer of forward messages not allocated yet\n");
  }

  // Foward pass
  for (int snp_i = 0; snp_i < n_snp_; ++snp_i) {
  	(*fwd_msgs)[snp_i] = this->AllocMsgBuf();
    this->CondProbs(pop_lab, snp_i, (*fwd_msgs)[snp_i]);

    if (snp_i > 0) {
      // Absorb a message from the right
      AbsorbMsg((*fwd_msgs)[snp_i - 1], (*fwd_msgs)[snp_i],
      		      l2r_map_, l2r_map_len_);
      scale_fact += log(RescaleMsg((*fwd_msgs)[snp_i],  n_states_));
    }

    // delete the previous message if it does not fall on the grid
    if (snp_i - 1 > 0 && ((snp_i - 1) % checkpt_freq) != 0) {
    	delete (*fwd_msgs)[snp_i - 1];
    	(*fwd_msgs)[snp_i - 1] = NULL;
    }
  }
  return(scale_fact);
}

void ModelParams::DeleteHigherIndexMsgs(int snp_idx,
		                                    vector<double *> * fwd_msgs) {
	for (int snp_i2 = snp_idx + 1; snp_i2 < n_snp_; ++snp_i2) {
		if ((*fwd_msgs)[snp_i2]) {
			delete (*fwd_msgs)[snp_i2];
			(*fwd_msgs)[snp_i2] = NULL;
		} else {
			break;
		}
	}
}

void ModelParams::BackfillForwardMsgs(const Util::IntVec& pop_lab,
		                                  int snp_idx, vector<double *> * fwd_msgs) {
	int start_marker = snp_idx - 1;
	for (; start_marker >= 0; --start_marker) {
	  if ((*fwd_msgs)[start_marker]) {
	  	break;
	  }
	}
	if (start_marker < 0) {
		Util::ThrowError("ModelParams::BackfillForwardMsgs: found no checkpointed "
				             "message\n");
	}

	// recompute messages up to the current marker
	for (int snp_i2 = start_marker + 1; snp_i2 <= snp_idx; ++snp_i2) {
		(*fwd_msgs)[snp_i2] = this->AllocMsgBuf();
    this->CondProbs(pop_lab, snp_i2, (*fwd_msgs)[snp_i2]);

    if (snp_i2 > 0) {
      // Absorb a message from the
      AbsorbMsg((*fwd_msgs)[snp_i2-1], (*fwd_msgs)[snp_i2],
      		      l2r_map_, l2r_map_len_);
      RescaleMsg((*fwd_msgs)[snp_i2], n_states_);
    }
	}
}

void ModelParams::DeleteForwardMsgs(vector<double *> * fwd_msgs) {
  for (int snp_i = 0; snp_i < n_snp_; ++snp_i) {
  	if ((*fwd_msgs)[snp_i]) {
  		delete (*fwd_msgs)[snp_i];
  		(*fwd_msgs)[snp_i] = NULL;
  	}
  }
}

double ModelParams::FwdBkwdAlgo(const Util::IntVec& pop_lab, double obs_wt,
		                            bool check_pointed, double * anc_prbs) {
	if (n_states_ == 0) error("FwdAlgo : n_states_ = 0\n");

  vector<double *> fwd_msgs(n_snp_, (double *)NULL);
  double * bkwd_msg = this->AllocMsgBuf();
  double * tmp_pttl = this->AllocMsgBuf();

  int checkpt_freq = (int)sqrt((double)n_states_);
  if (!check_pointed || checkpt_freq < 1) checkpt_freq = 1;
  double scale_fact = this->FwdAlgo(pop_lab, obs_wt, checkpt_freq, &fwd_msgs);

  // Backward pass
  for (int snp_i = n_snp_ - 1; snp_i >= 0; snp_i--) {
    this->CondProbs(pop_lab, snp_i, tmp_pttl);
    double * cur_pttl = fwd_msgs[snp_i];

    if (cur_pttl == NULL) {
    	// delete messages with higher index
    	this->DeleteHigherIndexMsgs(snp_i, &fwd_msgs);
    	this->BackfillForwardMsgs(pop_lab, snp_i, &fwd_msgs);
    	cur_pttl = fwd_msgs[snp_i];
    }

    if (snp_i < n_snp_ - 1) {
      // Absorb a message from the right to get a final potential
      AbsorbMsg(bkwd_msg, cur_pttl, r2l_map_, r2l_map_len_);
    }
    RescaleMsg(cur_pttl, n_states_);

    this->GetAncPrbs(snp_i, cur_pttl, anc_prbs);
    this->IncTreeCounts(obs_wt, snp_i, cur_pttl);

    // If we are in the rightmost SNP we did not need to absorb
    // any messages to get the expected counts.  In either case
    // they are stored and normalized here.
    if (snp_i < n_snp_ - 1) {
      // Finally absorb the message from the right a second time
      AbsorbMsg(bkwd_msg, tmp_pttl, r2l_map_, r2l_map_len_);
      RescaleMsg(tmp_pttl, n_states_);
    }

    // Swap the pointers so that this is now the backward message
    double * z = tmp_pttl;
    tmp_pttl = bkwd_msg;
    bkwd_msg = z;
  }

  delete tmp_pttl;
  delete bkwd_msg;
  this->DeleteForwardMsgs(&fwd_msgs);
  return scale_fact;
}

double ModelParams::FwdSumBkwdSample(const Util::IntVec& pop_lab,
		                                 bool check_pointed,
		                                 Util::IntVec * state_seq) {
	if (n_states_ == 0) Util::ThrowError("FwdAlgo : n_states_ = 0\n");

  vector<double *> fwd_msgs(n_snp_, (double *)NULL);
  double * bkwd_msg = this->AllocMsgBuf();
  double * tmp_pttl = this->AllocMsgBuf();

  int checkpt_freq = (int)sqrt((double)n_states_);
  if (!check_pointed || checkpt_freq < 1) checkpt_freq = 1;
  double scale_fact = this->FwdAlgo(pop_lab, 1.0, checkpt_freq, &fwd_msgs);

  // Backward pass
  for (int snp_i = n_snp_ - 1; snp_i >= 0; snp_i--) {
    this->CondProbs(pop_lab, snp_i, tmp_pttl);
    double * cur_pttl = fwd_msgs[snp_i];

    if (cur_pttl == NULL) {
    	// delete messages with higher index
    	this->DeleteHigherIndexMsgs(snp_i, &fwd_msgs);
    	this->BackfillForwardMsgs(pop_lab, snp_i, &fwd_msgs);
    	cur_pttl = fwd_msgs[snp_i];
    }

    if (snp_i < n_snp_ - 1) {
      // Absorb a message from the right to get a final potential
    	AbsorbSampleMsg(bkwd_msg, (*state_seq)[snp_i + 1], cur_pttl,
    			            r2l_map_, r2l_map_len_);
    }
    RescaleMsg(cur_pttl, n_states_);

    (*state_seq)[snp_i] = SampleState(unif_rand(), cur_pttl, n_states_);

    // sample a state

    // If we are in the rightmost SNP we did not need to absorb
    // any messages to get the expected counts.  In either case
    // they are stored and normalized here.
    if (snp_i < n_snp_ - 1) {
      // Finally absorb the message from the right a second time
      AbsorbMsg(bkwd_msg, tmp_pttl, r2l_map_, r2l_map_len_);
      RescaleMsg(tmp_pttl, n_states_);
    }

    // Swap the pointers so that this is now the backward message
    double * z = tmp_pttl;
    tmp_pttl = bkwd_msg;
    bkwd_msg = z;
  }

  delete tmp_pttl;
  delete bkwd_msg;
  this->DeleteForwardMsgs(&fwd_msgs);
  return scale_fact;
}

double ModelParams::FwdSumBkwdSamplePhaseLik(const Util::IntVec& pop_lab,
		                                         const Util::IntVec& phase_seq,
		                                         bool check_pointed) {
	if (n_states_ == 0) Util::ThrowError("FwdAlgo : n_states_ = 0\n");

  vector<double *> fwd_msgs(n_snp_, (double *)NULL);
  double * bkwd_msg = this->AllocMsgBuf();
  double * tmp_pttl = this->AllocMsgBuf();

  int checkpt_freq = (int)sqrt((double)n_states_);
  if (!check_pointed || checkpt_freq < 1) checkpt_freq = 1;

  double scale_fact = 0.0;

  // Foward pass
  for (int snp_i = 0; snp_i < n_snp_; ++snp_i) {
  	fwd_msgs[snp_i] = this->AllocMsgBuf();
    this->CondProbs(pop_lab, snp_i, fwd_msgs[snp_i]);
    this->ObservePhasing(snp_i, phase_seq[snp_i], fwd_msgs[snp_i]);

    if (snp_i > 0) {
      // Absorb a message from the right
      AbsorbMsg(fwd_msgs[snp_i - 1], fwd_msgs[snp_i],
      		      l2r_map_, l2r_map_len_);
      scale_fact += log(RescaleMsg(fwd_msgs[snp_i],  n_states_));
    }

    // delete the previous message if it does not fall on the grid
    if (snp_i - 1 > 0 && ((snp_i - 1) % checkpt_freq) != 0) {
    	delete fwd_msgs[snp_i - 1];
    	fwd_msgs[snp_i - 1] = NULL;
    }
  }

  delete tmp_pttl;
  delete bkwd_msg;
  this->DeleteForwardMsgs(&fwd_msgs);
  return scale_fact;
}


int PhasedTreeTraverse(const int * hap, const int * split_stream, int win_sz) {
  // returns index into split_stream of the leaf node we end up in

  int stream_pos = 0;
  for (;;) {
    if (split_stream[stream_pos] == -1) {
      return stream_pos;
    } else {
      int split_par = split_stream[stream_pos];
      split_par = hap[split_par];

      int stream_idx = stream_pos + 1 + split_par;
      if (split_par < 0 || split_par > 1 || stream_idx < 0) {
      	Util::ThrowError("TreeTraverse : stream_idx=%d, split_par=%d, stream_pos=%d\n",
      			             stream_idx, split_par, stream_pos);
      }

      stream_pos = split_stream[stream_idx];
    }
  }

  return(-1);
}

int TreeTraverse(const int * hap1, const int * hap2, const int * orientations,
		             const int * split_stream, int win_sz, int or_offset) {
  // We apply an orientation to each element of hap1 and hap2
  // returns index into split_stream of the leaf node we end up in

  int stream_pos = 0;
  for (;;) {
    if (split_stream[stream_pos] == -1) {
      return stream_pos;
    } else {
      int split_par = split_stream[stream_pos];
      if (split_par + or_offset < 0 || split_par + or_offset >= win_sz) {
      	Util::ThrowError("TreeTraverse : or_offset = %d, split_par = %d, sum = %d\n",
        		  or_offset, split_par, split_par + or_offset);
      }

      int hap_or = orientations[split_par + or_offset];
      if (hap_or < 0 || hap_or > 1) {
      	Util::ThrowError("TreeTraverse : orientation out of range (%d)\n", hap_or);
      }
      split_par = ((hap_or) ? hap2 : hap1)[split_par];

      int stream_idx = stream_pos + 1 + split_par;
      if (split_par < 0 || split_par > 1 || stream_idx < 0) {
      	Util::ThrowError("TreeTraverse: stream_idx=%d, split_par=%d, stream_pos=%d\n",
      			  stream_idx, split_par, stream_pos);
      }

      stream_pos = split_stream[stream_idx];
    }
  }

  return(-1);
}

int TreeTraverseUnphApprox(HapPair hap, int strand_idx,
		                       int ph_or_left, int ph_or_right, int or_change_pt,
		                       const int * split_stream) {
	// For markers with index <= or_change_pt, we get the allele from 'hap2' rather
	// than 'hap1'.  This is used in our approximation for unphased data.
	//
  // returns index into split_stream of the leaf node we end up in
	const int * hap1 = hap.first;
	const int * hap2 = hap.second;

	if (strand_idx) {
		ph_or_left = 1 - ph_or_left;
		ph_or_right = 1 - ph_or_right;
	}
	const int * h_rt = (ph_or_right) ? hap2 : hap1;
	const int * h_lft = (ph_or_left) ? hap2 : hap1;

  int stream_pos = 0;
  for (;;) {
    if (split_stream[stream_pos] == -1) {
      return stream_pos;
    } else {
      int split_par = split_stream[stream_pos];
      split_par = ((split_par <= or_change_pt) ? h_lft : h_rt)[split_par];

      int stream_idx = stream_pos + 1 + split_par;
      if (split_par < 0 || split_par > 1 || stream_idx < 0) {
      	Util::ThrowError("TreeTraverse : stream_idx=%d, split_par=%d, stream_pos=%d\n",
      			  stream_idx, split_par, stream_pos);
      }

      stream_pos = split_stream[stream_idx];
    }
  }

  return(-1);
}

double RescaleMsg(double * msg, int sz) {
  //if (sz <= 0) error("RescaleMsg : sz = %d\n", sz); //return(0.0);
  double tot = 0.0;
  double * x = msg;
  for (int i = 0; i < sz; ++i, ++x) {
    tot += *x;
  }
  if (tot <= 0.0) Util::ThrowError("RescaleMsg : message sums to <= 0\n");

  x = msg;
  for (int i = 0; i < sz; ++i, ++x) {
    *x /= tot;
  }
  return(tot);
}

bool CheckMap(int msg_sz, const int * msg_map, int map_len) {
  // Absorb a message from the left
  int map_pos = 0;
  while (map_pos < map_len) {
    int len = msg_map[map_pos];

    if (len + map_pos + 1 > map_len) {
    	Util::ThrowError("CheckMap : map construction wrong\n");
      return(false);
    }

    const int * mp = msg_map + (map_pos + 1);
    for(int i = 0; i < len; ++i, ++mp){
      if (*mp < 0 || *mp >= msg_sz) {
      	Util::ThrowError("CheckMap : index out of range\n");
        return(false);
      }
    }
    map_pos += len + 1;
  }
  return(true);
}

// Absorb the table 'msg' into the table 'pttl' by going through each
// entry in pttl and looking up which elements of msg are compatible with it.
// We sum over these compatible elements and mutliply in the result.

void AbsorbMsg(const double * msg, double * pttl, const int * msg_map,
		           int map_len) {
  // Absorb a message from the left
  int map_pos = 0;
  while (map_pos < map_len) {
    int len = msg_map[map_pos];
    double cur_sum = 0.0;

    const int * mp = msg_map + (map_pos + 1);
    for(int i = 0; i < len; ++i, ++mp){
      cur_sum += msg[*mp];
    }
    map_pos += len + 1;
    (*pttl) *= cur_sum;

    pttl++;
  }
}

void AbsorbSampleMsg(double * msg, int state_idx, double * pttl,
		                 const int * msg_map, int map_len) {
  // Absorb a message from the left
  int map_pos = 0;
  while (map_pos < map_len) {
    int len = msg_map[map_pos];
    double cur_sum = 0.0;

    const int * mp = msg_map + (map_pos + 1);
    for(int i = 0; i < len; ++i, ++mp){
      if ((*mp) == state_idx) cur_sum += msg[*mp];
    }
    map_pos += len + 1;
    (*pttl) *= cur_sum;

    pttl++;
  }
}

int SampleState(double unif_var, const double * pttl, int pttl_sz) {
  for (int state_idx = 0; state_idx < pttl_sz; ++pttl, ++state_idx) {
  	unif_var -= *pttl;
  	if (unif_var <= 0.0) {
      return(state_idx);
  	}
  }
  return(-1);
}

bool CheckDropIndexInc(const ModelParams * p, int snp_idx, int hid_state) {
	// p->drop_lookups_ has dimensions [number-populations, window-size, number-markers]

  int drop_idx_last = -1;

  for (int jump_counter = 0; jump_counter < p->win_sz_; ++jump_counter) {
  	int drop_idx = 0;

	  if (snp_idx - jump_counter > 0 && jump_counter < p->win_sz_ - 1) {
	  	int dl_idx = p->edgecut_indexer_.FlatIndex(hid_state, snp_idx, jump_counter);
	  	if (dl_idx < 0 || dl_idx >= p->drop_lookups_len_) {
	  		Util::ThrowError("CheckDropIndexInc : drop_lookup index out of range\n");
	  	}
	    drop_idx = p->drop_lookups_[dl_idx];
	  }
	  if (snp_idx - jump_counter > 0 && drop_idx_last >= 0
	  		&& drop_idx > drop_idx_last) {
	  	Util::ThrowError("CheckDropIndexInc : drop index not decreasing (%d to %d) at jump_counter = %d, snp_idx = %d\n",
	  			drop_idx_last, drop_idx, jump_counter, snp_idx);
	  	return(false);
	  } else {
	  	drop_idx_last = drop_idx;
	  }
  }
  return(true);
}

}  // namespace
}  // namespace
