#ifndef PHASEEM_C_SRC_PHASE_EM_HAP_OPT_HPP__
#define PHASEEM_C_SRC_PHASE_EM_HAP_OPT_HPP__ 1

#include <R.h>
#include <Rdefines.h>
#include <Rmath.h>
#include <R_ext/Print.h>

#include "util.hpp"
#include "binary_mat.hpp"
#include "phase_em.hpp"

namespace PhaseEM {

using namespace Util;
using namespace BinaryMatrix;

class PhaseEmHapOptBase {
public:
	PhaseEmHapOptBase(PhaseEmModel& p)
	  : params_(p), rescale_freq_(3), move_thresh_(0.3),
	    move_liks_(p.n_snp(), R_NegInf), verbose_(false) {

    for (int si = 0; si < 2; ++si) {
    	inf_obj_[si] = new PhaseEmInference(p);
      case_inds_[si] = -1;
    }
	}

	void set_move_thresh(double x) { move_thresh_ = x; }
  void set_verbose(bool x) { verbose_ = x; }

	void set_case_inds(int case_idx1, int case_idx2) {
		case_inds_[0] = case_idx1;
		case_inds_[1] = case_idx2;
		inf_obj_[0]->set_case_index(case_idx1);
		inf_obj_[1]->set_case_index(case_idx2);
	}

  const RealVec& move_liks() const { return(move_liks_); }

	virtual void ForwardAlgo() {
		for (int si = 0; si < 2; ++si) {
			inf_obj_[si]->ForwardAlgo();
		}
	}

	virtual void BackwardAlgo() = 0;
	virtual void ApplyMoves() = 0;

  virtual ~PhaseEmHapOptBase() {
  	delete inf_obj_[0];
  	delete inf_obj_[1];
  }

protected:
	virtual double Lik(const RealVec& bkwd_msg, const RealVec& fwd_msg) {
		double r = DotProd(bkwd_msg, fwd_msg);
		return(log(r));
	}
	virtual bool MakeMove(int * snp_idx) = 0;

	void RescaleMsg(RealVec * msg) {
	  double scale_amt = 1.0 / VecSum<double>(*msg);
		VecScale<double>(scale_amt, msg);
	}

	void InitMsg(RealVec * m) {
		for (RealVec::iterator i = m->begin(); i != m->end(); ++i) {
			*i = 1.0;
		}
	}

public:
	PhaseEmModel& params_;

	int rescale_freq_;
	double move_thresh_;
	PhaseEmInference * inf_obj_[2];
	RealVec move_liks_;

	int case_inds_[2];
	bool verbose_;
};

class PhaseEmHapOpt : public PhaseEmHapOptBase {
public:
	PhaseEmHapOpt(PhaseEmModel& p, const CastableMatI& gtype_codes)
	  : PhaseEmHapOptBase(p),
	    made_cross_(p.n_snp(), false),
	    opt_alleles_(p.n_snp()),
	    gtype_codes_(gtype_codes),
	    allow_crossover_(true), allow_single_site_(true) {

    for (int si = 0; si < 2; ++si) {
      for (int gi = 0; gi < 2; ++gi) {
        bkwd_msgs_[si][gi] = RealVec(p.n_cases(), 1.0);
      }
    }
	}

	void set_allow_single_site(bool x) { allow_single_site_ = x; }
	void set_allow_crossover(bool x) { allow_crossover_ = x; }

	void BackwardAlgo() {
		for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
			this->InitMsg(&bkwd_msgs_[strand_idx][0]);
		}

		for (int snp_idx = params_.n_snp() - 1; snp_idx >= 0; --snp_idx) {
			for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
				for (int a0 = 1; a0 >= 0; --a0) {
				  inf_obj_[strand_idx]->BackwardStepSetAllele(snp_idx, a0, bkwd_msgs_[strand_idx][0],
						                                          &bkwd_msgs_[strand_idx][a0]);
				}
			}
			this->MakeMove(&snp_idx);

			if (snp_idx % rescale_freq_ == 0) {
				for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
				  this->RescaleMsg(&bkwd_msgs_[strand_idx][0]);
				}
			}
		}
	}

	virtual void ApplyMoves() {
    bool cur_or = true;
    for (int i = 0; i < params_.n_snp(); ++i) {
    	if (cur_or) {
    		params_.SetAllele(case_inds_[0], i, opt_alleles_[i].first);
    		params_.SetAllele(case_inds_[1], i, opt_alleles_[i].second);
    	} else {
    		params_.SetAllele(case_inds_[0], i, opt_alleles_[i].second);
    		params_.SetAllele(case_inds_[1], i, opt_alleles_[i].first);
    	}

    	if (made_cross_[i]) cur_or = !cur_or;
    }
	}

  virtual ~PhaseEmHapOpt() { }

protected:
	void NoMove(int snp_idx, double move_lik);
	void SetMove(int snp_idx, int made_cross, int a0, int a1, double move_lik);
	virtual bool MakeMove(int * snp_idx);
	virtual double LikRatio(int snp_idx);

public:
	RealVec bkwd_msgs_[2][2];
	vector<bool> made_cross_;
	vector<pair<int, int> > opt_alleles_;
	const CastableMatI& gtype_codes_;
	bool allow_crossover_, allow_single_site_;
};

class CrossoverHapOpt : public PhaseEmHapOptBase {
public:
	CrossoverHapOpt(PhaseEmModel& p)
	  : PhaseEmHapOptBase(p),
	    made_cross_(p.n_snp(), false) {

    for (int si = 0; si < 2; ++si) {
      bkwd_msgs_[si] = RealVec(p.n_cases(), 1.0);
    }
	}

	void BackwardAlgo() {
		for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
		  this->InitMsg(&bkwd_msgs_[strand_idx]);
		}

		for (int snp_idx = params_.n_snp() - 1; snp_idx >= 0; --snp_idx) {
			for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
				inf_obj_[strand_idx]->BackwardStep(snp_idx, bkwd_msgs_[strand_idx],
						                              &bkwd_msgs_[strand_idx]);
			}
			this->MakeMove(&snp_idx);

			if (snp_idx % rescale_freq_ == 0) {
				for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
				  this->RescaleMsg(&bkwd_msgs_[strand_idx]);
				}
			}
		}
	}

	void ApplyMoves() {
    bool cur_or = true;
    for (int i = 0; i < params_.n_snp(); ++i) {
    	if (made_cross_[i]) cur_or = !cur_or;

    	if (!cur_or) {
    		int a0 = params_.GetAllele(case_inds_[0], i);
    		int a1 = params_.GetAllele(case_inds_[1], i);
    		params_.SetAllele(case_inds_[0], i, a1);
    		params_.SetAllele(case_inds_[1], i, a0);
    	}
    }
	}

  virtual ~CrossoverHapOpt() { }

protected:
	double LikRatio(int snp_idx);
	virtual bool MakeMove(int * snp_idx);

public:
	RealVec bkwd_msgs_[2];
	vector<bool> made_cross_;
};

class RunTracker {
public:
	RunTracker() : best_swap_idx_(-1), best_swap_score_(R_NegInf) {}

	int best_swap_idx() const {
		return(best_swap_idx_);
	}

	int best_swap_score() const {
		return(best_swap_score_);
	}

	void MarkRun(int snp_idx, double lik_ratio) {
		best_swap_idx_ = snp_idx;
    best_swap_score_ = lik_ratio;
	}

	bool RunStarted(int snp_idx, double lik_ratio, double start_thresh) {
		if (lik_ratio > start_thresh && (best_swap_idx_ < 0 || lik_ratio > best_swap_score_)) {
			return(true);
		} else {
			return(false);
		}
	}

	bool RunEnded(int snp_idx, double lik_ratio) {
		if (best_swap_idx_ < 0) {
			return(false);
		} else if (snp_idx == 0 || lik_ratio < 0) {
			return(true);
		} else {
			return(false);
		}
	}

	void EndRun(int * snp_idx) {
		*snp_idx = best_swap_idx_;
		best_swap_idx_ = -1;
    best_swap_score_ = R_NegInf;
	}

	virtual ~RunTracker() {}

private:
  int best_swap_idx_;
  double best_swap_score_;
};

class BacktrackHapOpt : public PhaseEmHapOpt {
public:
	BacktrackHapOpt(PhaseEmModel& p, const CastableMatI& gtype_codes)
	  : PhaseEmHapOpt(p, gtype_codes), run_tracker_() {

    for (int si = 0; si < 2; ++si) {
	    for (int a0 = 0; a0 < 2; ++a0) {
	      best_bkwd_msgs_[si][a0] = RealVec(p.n_cases(), 1.0);
	    }
    }
	}

  virtual ~BacktrackHapOpt() { }

protected:
	void MarkRun(int snp_idx, double lik_ratio) {
		run_tracker_.MarkRun(snp_idx, lik_ratio);
    for (int si = 0; si < 2; ++si) {
    	for (int a0 = 0; a0 < 2; ++a0) {
        best_bkwd_msgs_[si][a0] = bkwd_msgs_[si][a0];
    	}
    }
	}

	void LoadBestMsgs() {
		for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
			for (int a0 = 0; a0 < 2; ++a0) {
			  bkwd_msgs_[strand_idx][a0] = best_bkwd_msgs_[strand_idx][a0];
			}
	  }
	}

	virtual bool MakeMove(int * snp_idx) {
		int si = *snp_idx;
		double cur_lr = this->LikRatio(si);

		move_liks_[si] = cur_lr;
		made_cross_[si] = false;

		if (run_tracker_.RunStarted(si, cur_lr, move_thresh_)) {
      this->MarkRun(si, cur_lr);
      this->PhaseEmHapOpt::NoMove(si, cur_lr);

		} else if (run_tracker_.RunEnded(si, cur_lr)) {
			run_tracker_.EndRun(snp_idx);
		  this->LoadBestMsgs();
			this->PhaseEmHapOpt::MakeMove(snp_idx);
			return(true);

		} else {
      this->PhaseEmHapOpt::NoMove(si, cur_lr);
		}
		return(false);
	}

public:
	RealVec best_bkwd_msgs_[2][2];
	RunTracker run_tracker_;
};


class BacktrackXOverHapOpt : public CrossoverHapOpt {
public:
	BacktrackXOverHapOpt(PhaseEmModel& p)
	  : CrossoverHapOpt(p), run_tracker_() {

    for (int si = 0; si < 2; ++si) {
      best_bkwd_msgs_[si] = RealVec(p.n_cases(), 1.0);
    }
	}

  virtual ~BacktrackXOverHapOpt() { }

protected:
	void MarkRun(int snp_idx, double lik_ratio) {
		run_tracker_.MarkRun(snp_idx, lik_ratio);
    for (int si = 0; si < 2; ++si) {
    	best_bkwd_msgs_[si] = bkwd_msgs_[si];
    }
	}

	void LoadBestMsgs() {
		for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
			bkwd_msgs_[strand_idx] = best_bkwd_msgs_[strand_idx];
	  }
	}

	virtual bool MakeMove(int * snp_idx);

public:
	RealVec best_bkwd_msgs_[2];
	RunTracker run_tracker_;
};


CrossoverHapOpt * CrossoverHapOptFactory(PhaseEmModel& p,
		                                     bool back_tracking);


PhaseEmHapOpt * HapOptFactory(PhaseEmModel& p, const CastableMatI& gtype_codes,
		                          bool back_tracking);

}  // namespace PhaseEM


extern "C" {
SEXP PhaseEmGenotypeCodes_RI(SEXP allele_mat_sxp, SEXP gtype_mat_sxp);

SEXP PhaseEmCrossoverScores_RI(SEXP params_sxp, SEXP cases_sxp,
					                  SEXP case_idx1_sxp, SEXP case_idx2_sxp);

SEXP PhaseEmCrossoverOpt_RI(SEXP params_sxp, SEXP cases_sxp,
					                  SEXP case_inds_sxp, SEXP move_thresh_sxp,
					                  SEXP back_tracking_sxp);

SEXP PhaseEmHapOptScores_RI(SEXP params_sxp, SEXP cases_sxp,
					                  SEXP case_idx1_sxp, SEXP case_idx2_sxp,
					                  SEXP allow_crossover_sxp,
					                  SEXP gtype_codes_sxp);

SEXP PhaseEmHapOpt_RI(SEXP params_sxp, SEXP cases_sxp,
					            SEXP case_inds_sxp, SEXP move_thresh_sxp,
					            SEXP allow_crossover_sxp,
					            SEXP allow_single_site_sxp,
					            SEXP gtype_codes_sxp,
					            SEXP back_tracking_sxp);

}  // extern "C" {

#endif