#ifndef BIN_TREE_HPP_
#define BIN_TREE_HPP_

#include "util.hpp"
#include "binary_mat.hpp"
#include "aradmx.hpp"

namespace BinTree {

using namespace Util;
using namespace BinaryMatrix;

class BaseTree {
public:
	static const int kNoSplit = -1;

	class TreeNode {
	public:
		int split_var_;
		int ch0_idx_;
		int ch1_idx_;
		int depth_;
		IntVec pt_set_;

		TreeNode() : split_var_(kNoSplit), ch0_idx_(-1), ch1_idx_(-1), depth_(0) { }

		bool IsLeaf() const {
			if (ch0_idx_ == -1 && ch1_idx_ != -1) {
				Util::ThrowError("BaseTree::TreeNode : invalid child indices\n");
			}
			return(ch0_idx_ == -1);
		}
	};

	BaseTree(const BinaryMat& dat, const IntVec& pop_labels, int target_var);

	void set_pop_wtd_score(bool v) {
		pop_wtd_score_ = v;
	}
	void set_max_split_depth(int depth) {
		max_split_depth_ = depth;
	}

	void set_min_leaf_sz(int sz) {
		min_leaf_sz_ = sz;
	}

	void set_target_var(int idx) {
		target_var_ = idx;
	}

	void set_split_pcount(double x) {
		split_pcount_ = x;
	}

	void AddSplitVar(int idx) {
		split_vars_.push_back(idx);
	}

	void ClearSplitVars() {
		split_vars_.clear();
	}

	const IntVec& split_vars() const { return split_vars_; }

	void ClearTree();

	bool IsPopSplitVar(int split_var) const {
		return(split_var < -1);
	}

	int SplitVarToPopIdx(int split_var) const {
		return(-split_var - 2);
	}

	int PopIdxToSplitVar(int pop_idx) const {
		return(-pop_idx - 2);
	}

	bool pop_wtd_score() const { return(pop_wtd_score_); }

	// For the 'SinglePop' versions we compute the score using only samples from a single
	// population specified in pop_idx.  We do not apply the population weight.  A value
	// of -1 means to use all populations pooled (ignore the population label and population
	// weights).
	virtual double GetSinglePopNoSplitScore(int node_idx, int pop_idx) const;
	virtual double GetNoSplitScore(int node_idx) const;
	virtual double GetSinglePopSplitScore(int node_idx, int split_var, int pop_idx) const;
	virtual double GetSplitScore(int node_idx, int split_var) const;

	virtual void SplitNode(int node_idx, int split_var);

	virtual void NodeTargetCounts(int node_idx, double * n0, double * n1) const;

	virtual void GrowTree() = 0;

	void GetSplitStream(IntVec * splits, RealVec * leaf_counts) const;
	void GetSinglePopSplitStream(IntVec * splits, RealVec * leaf_counts, int pop_idx) const;

protected:
	const BinaryMat& dat_;
	const IntVec& pop_labels_;
	IntVec pt_labels_;
	vector<TreeNode> tree_;
	int target_var_;
	IntVec split_vars_;

	int n_obs_;

	RealVec pop_wts_;

	double split_pcount_;
	int min_leaf_sz_;
	int max_split_depth_;
	bool pop_wtd_score_;
};


class GreedyTreeBuilder : public BaseTree {
public:
	GreedyTreeBuilder(const BinaryMat& dat, const IntVec& pop_labels, int target_var)
	  : BaseTree(dat, pop_labels, target_var) {
		split_penalty_ = 0.0;
	}

	void set_split_penalty(double x) {
		split_penalty_ = x;
	}

	virtual void GrowTree();

protected:
	int BestNodeSplit(int node_idx) const;

protected:
	double split_penalty_;
};


class RandomTreeBuilder : public GreedyTreeBuilder {
public:
	RandomTreeBuilder(const BinaryMat& dat, const IntVec& pop_labels, int target_var)
	  : GreedyTreeBuilder(dat, pop_labels, target_var) { }

	virtual void GrowTree() {
		for (int iter_idx = 0;; ++iter_idx) {
			bool made_split = false;
			RealVec split_prbs(this->split_vars().size() + 1);

			for (int nd_idx = 0, sz = tree_.size(); nd_idx < sz; ++nd_idx) {
				int split_var = this->SampleNodeSplit(nd_idx, &split_prbs);

				if (split_var != BaseTree::kNoSplit) {
					this->SplitNode(nd_idx, split_var);
					made_split = true;
				}
			}
			if (!made_split) {
				break;
			}
		}
	}

protected:
	int SampleNodeSplit(int node_idx, RealVec * prb_buffer) const;
};

}

extern "C" {
	SEXP GrowBinTrees_RI(SEXP bin_mat, SEXP p_pop_labels,
								    SEXP p_max_index_diffs, SEXP p_mdl_params_sxp,
								    SEXP p_split_penalty, SEXP p_pop_idx, SEXP p_verbose,
								    SEXP split_lists, SEXP split_counts);

	SEXP IncLeafCounts_RI(SEXP bin_mat, SEXP wts_sxp, SEXP zero_counts_sxp,
			                   SEXP p_mdl_params_sxp, SEXP tree_sxp);

	SEXP FirstUnclippedTreeIndex_RI(SEXP trees_sxp, SEXP max_drop_sxp);

	SEXP PackIntsToUnsignedChars_RI(SEXP ints_sxp);
}

#endif /*BIN_TREE_HPP_*/
