#ifndef BINARY_MAT_HPP__
#define BINARY_MAT_HPP__

#include <R.h>
#include <Rdefines.h>
#include <Rmath.h>
#include <R_ext/Print.h>

#include "util.hpp"

#define BM_BOUNDCHECKS

namespace BinaryMatrix {

using namespace Util;

class FactoredMatrix {
public:
  FactoredMatrix(SEXP sxp) {
    SEXP u = ListEltByName(sxp, "u", NULL);
    SEXP v = ListEltByName(sxp, "v", NULL);
    SEXP nr = ListEltByName(sxp, "nr", NULL);
    SEXP nc = ListEltByName(sxp, "nc", NULL);
    SEXP nfact = ListEltByName(sxp, "nfact", NULL);

    u_ = REAL(u);
    v_ = REAL(v);
    nr_ = GetInt(nr, 0, NULL);
    nc_ = GetInt(nc, 0, NULL);
    nfact_ = GetInt(nfact, 0, NULL);

    int unr = -1, unc = -1, vnr = -1, vnc = -1;
    Util::GetMatrixDim(u, &unr, &unc);
    Util::GetMatrixDim(v, &vnr, &vnc);
    if (unr != nr_ || unc != nfact_ || vnr != nfact_ || vnc != nc_) {
    	error("FactoredMatrix : bad input dimensions\n");
    }
  }

  double GetEltDbl(int ri, int ci) const {
    double r = 0.0;
    for (int fact_idx = 0; fact_idx < nfact_; ++fact_idx) {
      r += this->GetU(ri, fact_idx) * this->GetV(ci, fact_idx);
    }
    return(r);
  }

  double GetU(int row_idx, int fact_idx) const {
#ifdef BM_BOUNDCHECKS
  	if (row_idx >= nr_ || fact_idx >= nfact_) error("FactoredMatrix::GetU : bad input\n");
#endif
    return(u_[row_idx + nr_ * fact_idx]);
  }

  // Even though internally the 'V' may be stored as [fact-idx, col-idx] we reverse
  // the order of the parameters
  double GetV(int col_idx, int fact_idx) const {
#ifdef BM_BOUNDCHECKS
  	if (col_idx >= nc_ || fact_idx >= nfact_) error("FactoredMatrix::GetV : bad input\n");
#endif
    return(v_[fact_idx + nfact_ * col_idx]);
  }

  void SetU(int row_idx, int fact_idx, double val) {
#ifdef BM_BOUNDCHECKS
  	if (row_idx >= nr_ || fact_idx >= nfact_) error("FactoredMatrix::SetU : bad input\n");
#endif
    u_[row_idx + nr_ * fact_idx] = val;
  }
  void SetV(int col_idx, int fact_idx, double val) {
#ifdef BM_BOUNDCHECKS
  	if (col_idx >= nc_ || fact_idx >= nfact_) error("FactoredMatrix::GetV : bad input\n");
#endif
    v_[fact_idx + nfact_ * col_idx] = val;
  }

  int nfact() const { return nfact_; }
public:
  int nr_, nc_, nfact_;
  double * u_, * v_;
};

class BaseMat {
public:
  BaseMat() : nr_(-1), nc_(-1) { }
  BaseMat(int nrow, int ncol) : nr_(nrow), nc_(ncol) { }

  int nrow() const { return(nr_); }
  int ncol() const { return(nc_); }
public:
  int nr_, nc_;
};

class ByteMat : public BaseMat {
public:
  ByteMat(unsigned char * dat, int nrow, int ncol, int bytes_per_col)
    : BaseMat(nrow, ncol), dat_(dat), bytes_per_col_(bytes_per_col) { }

  ByteMat(SEXP sxp);
  static bool IsByteMat(SEXP sxp_obj);

  void SetElt(int ri, int ci, int val) {
    dat_[ri + bytes_per_col_ * ci] = val;
  }
  void SetEltDbl(int ri, int ci, double val) {
    this->SetElt(ri, ci, val);
  }

  int GetElt(int ri, int ci) const {
    return(dat_[ri + bytes_per_col_ * ci]);
  }
  double GetEltDbl(int ri, int ci) const {
    return(this->GetElt(ri, ci));
  }

  virtual ~ByteMat() {
    dat_ = NULL;
  }

public:
  unsigned char * dat_;
  int bytes_per_col_;
};

class IntMat : public BaseMat {
public:
  IntMat(int * dat, int nrow, int ncol)
    : BaseMat(nrow, ncol), dat_(dat) { }

  IntMat(SEXP sxp_obj) : BaseMat() {
    GetMatrixDim(sxp_obj, &nr_, &nc_);
    dat_ = INTEGER(sxp_obj);
  }

  static bool IsIntMat(SEXP sxp) {
    return((IS_INTEGER(sxp) || IS_LOGICAL(sxp)) && Util::IsMatrix(sxp));
  }

  void SetElt(int ri, int ci, int val) {
    dat_[ri + nr_ * ci] = val;
  }
  void SetEltDbl(int ri, int ci, double val) {
    this->SetElt(ri, ci, val);
  }
  int GetElt(int ri, int ci) const {
    return(dat_[ri + nr_ * ci]);
  }
  double GetEltDbl(int ri, int ci) const {
    return(this->GetElt(ri, ci));
  }
  virtual ~IntMat() { dat_ = NULL; }

public:
  int * dat_;
};

class DblMat : public BaseMat {
public:
  DblMat(double * dat, int nrow, int ncol)
    : BaseMat(nrow, ncol), dat_(dat) { }

  DblMat(SEXP sxp_obj) : BaseMat() {
    GetMatrixDim(sxp_obj, &nr_, &nc_);
    dat_ = REAL(sxp_obj);
  }

  static bool IsDblMat(SEXP sxp) {
    return(IS_NUMERIC(sxp) && Util::IsMatrix(sxp));
  }

  void SetElt(int ri, int ci, int val) {
    this->SetEltDbl(ri, ci, val);
  }
  void SetEltDbl(int ri, int ci, double val) {
    dat_[ri + nr_ * ci] = val;
  }
  int GetElt(int ri, int ci) const {
    return(this->GetEltDbl(ri, ci));
  }
  double GetEltDbl(int ri, int ci) const {
    return(dat_[ri + nr_ * ci]);
  }
  virtual ~DblMat() { dat_ = NULL; }

public:
  double * dat_;
};

class BinaryMat : public BaseMat {
public:
  BinaryMat(unsigned char * dat, int nrow, int ncol, int bytes_per_col)
    : BaseMat(nrow, ncol), dat_(dat), bytes_per_col_(bytes_per_col) { }

  BinaryMat(SEXP sxp);
  static bool IsBinaryMat(SEXP sxp_obj);

  void SetElt(int ri, int ci, int val);
  void SetEltDbl(int ri, int ci, double val) {
    this->SetElt(ri, ci, val);
  }
  int GetElt(int ri, int ci) const;
  double GetEltDbl(int ri, int ci) const {
    return(this->GetElt(ri, ci));
  }

  virtual ~BinaryMat() {
    dat_ = NULL;
  }

public:
  class RowIterator {
  public:
    RowIterator(const BinaryMat& bm, int col_idx)
      : x_(bm.dat_ + col_idx * bm.bytes_per_col_) {
      x_idx_ = buf_idx_ = idx_ = 0;
      this->SetBuf(x_[0]);
    }

    void Inc();
    unsigned char Val() const { return(buf_[buf_idx_]); }

  private:
    void SetBuf(unsigned char x);

  private:
    const unsigned char * x_;
    int idx_, buf_idx_, x_idx_;
    unsigned char buf_[8];
  };

public:
  unsigned char * dat_;
  // number of rows, columns and integers used per row
  int bytes_per_col_;

  friend class RowIterator;
};

class OnesMat : public BaseMat {
public:
	OnesMat(int nrow, int ncol) : BaseMat(nrow, ncol) { }

	OnesMat(SEXP sxp);
  static bool IsOnesMat(SEXP sxp_obj);

  void SetElt(int ri, int ci, int val) {
  	error("SetElt : called on OnesMat object\n");
  }
  void SetEltDbl(int ri, int ci, double val) {
  	error("SetEltDbl : called on OnesMat object\n");
  }
  int GetElt(int ri, int ci) const { return 1; }
  double GetEltDbl(int ri, int ci) const { return 1.0; }
  virtual ~OnesMat() { }
};

template <class T> class ScaledMat : public T {
public:
	ScaledMat(unsigned char * dat, int nrow, int ncol, int bytes_per_col)
    : T(dat, nrow, ncol, bytes_per_col) {
		error("ScaledMat<class T> : can only be instantiated from existing R object\n");
	}

	ScaledMat(SEXP sxp) : T(sxp) {
	  SEXP rshift = ListEltByName(sxp, "rshift", NULL);
	  SEXP cshift = ListEltByName(sxp, "cshift", NULL);
	  SEXP rmlt = ListEltByName(sxp, "rmlt", NULL);
	  SEXP cmlt = ListEltByName(sxp, "cmlt", NULL);

	  rshift_ = REAL(rshift);
	  cshift_ = REAL(cshift);
	  rmlt_ = REAL(rmlt);
	  cmlt_ = REAL(cmlt);

	  if (LENGTH(rshift) != this->nr_ || LENGTH(rmlt) != this->nr_) {
	  	error("ScaledMat : check row shift/scale\n");
	  } else if (LENGTH(cshift) != this->nc_ || LENGTH(cmlt) != this->nc_) {
	  	error("ScaledMat : check column shift/scale\n");
	  }
	}

  double GetEltDbl(int ri, int ci) const {
#ifdef BM_BOUNDCHECKS
  	if (ri < 0 || ri >= this->nr_ || ci < 0 || ci >= this->nc_) {
  		error("ScaledMat::GetEltDbl : bad row/column indices\n");
  	}
#endif
  	double v = this->GetElt(ri, ci);
  	v += this->GetShift(ri, ci);
  	v *= this->GetScale(ri, ci);
    return(v);
  }

  double GetShift(int ri, int ci) const {
  	return(rshift_[ri] + cshift_[ci]);
  }
  double GetScale(int ri, int ci) const {
  	return(rmlt_[ri] * cmlt_[ci]);
  }

  virtual ~ScaledMat() {
    rshift_ = cshift_ = rmlt_ = cmlt_ = NULL;
  }

public:
  double * rshift_, * cshift_, * rmlt_, * cmlt_;
};

bool IsBinaryMatScaled(SEXP sxp_obj);

bool IsByteMatScaled(SEXP sxp_obj);

class CastableMatI {
public:
  virtual int GetElt(int ri, int ci) const = 0;
  virtual double GetEltDbl(int ri, int ci) const = 0;
  virtual void SetElt(int ri, int ci, int val) = 0;
  virtual void SetEltDbl(int ri, int ci, double val) = 0;
  virtual int nrow() const = 0;
  virtual int ncol() const = 0;
  virtual ~CastableMatI() { }

public:
  // These are derived from the interface and do not need to be implemented
  void Fill(bool val) {
    int v0 = (val) ? 1 : 0;
    int nr = this->nrow();
    int nc = this->ncol();

    for (int ri = 0; ri < nr; ++ri) {
      for (int ci = 0; ci < nc; ++ci) {
        this->SetElt(ri, ci, v0);
      }
    }
  }

  bool EqualDims(const CastableMatI& m) const {
    if (this->nrow() != m.nrow()) return(false);
    if (this->ncol() != m.ncol()) return(false);
    return(true);
  }
};



template <class T> class CastableMat : public CastableMatI {
public:
  CastableMat(T * m) : m_(m) { }
  virtual ~CastableMat() {
    m_ = NULL;
  }

  void CheckMatrix() const {
    if (!m_) error("CastableMat : m_ is NULL\n");
  }

  virtual int nrow() const {
    return(m_->nrow());
  }

  virtual int ncol() const {
    return(m_->ncol());
  }

  virtual int GetElt(int ri, int ci) const {
    return(m_->GetElt(ri, ci));
  }

  virtual void SetElt(int ri, int ci, int val) {
    m_->SetElt(ri, ci, val);
  }

  virtual double GetEltDbl(int ri, int ci) const {
    return(m_->GetEltDbl(ri, ci));
  }

  virtual void SetEltDbl(int ri, int ci, double val) {
    m_->SetEltDbl(ri, ci, val);
  }

protected:
  void DeleteMatrix() {
    if (m_) delete m_;
    m_ = NULL;
  }
private:
  CastableMat() : m_(NULL) { }
  T * m_;
};


template <class T> class OwningCastableMat : public CastableMat<T> {
public:
  OwningCastableMat(T * m) : CastableMat<T>(m) { }
  virtual ~OwningCastableMat() {
    this->DeleteMatrix();
  }
};


void RbindMats(const CastableMatI& a, const CastableMatI& b,
               CastableMatI * c);
void CbindMats(const CastableMatI& a, const CastableMatI& b,
               CastableMatI * c);

void ConvertMat(const CastableMatI& a, const IntVec& row_inds,
                const IntVec& col_inds, CastableMatI * b);
void ConvertMatDbl(const CastableMatI& a, const IntVec& row_inds,
                   const IntVec& col_inds, CastableMatI * b);

int MatrixHash(const CastableMatI& a);

void TransposeMat(const CastableMatI& a, CastableMatI * b);
CastableMatI * OwningCastableMatFactory(SEXP sxp_obj);

}  // namespace

extern "C" {

SEXP MarginSums(SEXP mat_sxp, SEXP margin_idx_sxp, SEXP r_sxp);

SEXP ConvertMatrix(SEXP mat1_sxp, SEXP row_inds_sxp, SEXP col_inds_sxp,
                   SEXP mat2_sxp);
SEXP ConvertMatrixDbl(SEXP mat1_sxp, SEXP row_inds_sxp, SEXP col_inds_sxp,
                      SEXP mat2_sxp);
SEXP RbindMatrix(SEXP mat1_sxp, SEXP mat2_sxp, SEXP mat3_sxp);
SEXP CbindMatrix(SEXP mat1_sxp, SEXP mat2_sxp, SEXP mat3_sxp);

SEXP FillMatrix(SEXP val_sxp, SEXP mat_sxp);

SEXP SetMatrixSlice(SEXP mat1_sxp, SEXP row_inds_sxp, SEXP col_inds_sxp,
                    SEXP mat2_sxp);

SEXP TransposeMatrix(SEXP mat1_sxp, SEXP mat2_sxp);

SEXP MultMatMatMasked(SEXP mat1_sxp, SEXP mat2_sxp, SEXP msk_sxp,
		                  SEXP r_sxp);
SEXP MaskedPcaLRUpdate(SEXP mat_sxp, SEXP msk_sxp, SEXP fact_idx_sxp,
		                   SEXP fm_sxp, SEXP tied_cols_sxp);
SEXP MaskedPcaSubSpaceGrad(SEXP mat_sxp, SEXP msk_sxp, SEXP fm_sxp,
		                       SEXP fm_grad_sxp);

//SEXP ComponentAnd(SEXP mat1_sxp, SEXP mat2_sxp);

}

#endif