#ifndef SWITCH_ERR_HPP__
#define SWITCH_ERR_HPP__

#include <R.h>
#include <Rdefines.h>
#include <Rmath.h>
#include <R_ext/Print.h>

#include "util.hpp"
#include "binary_mat.hpp"

using namespace BinaryMatrix;

namespace SwitchError {

class SwitchErrorCalculator {
public:
	// bins = {x_1, x_2, x_3, ..., x_n}
	// where the elements must be sorted and x_n is the largest
	// distance considered.
  // n_errs and n_tests must be the same length as 'bins'
	static void Calc(const CastableMatI& true_ph,
			             const CastableMatI& ph_known,
			             const CastableMatI& test_ph,
			             int * n_errs,
			             int * n_tests) {
		if (!CheckInputs(true_ph, ph_known, test_ph)) {
			error("SwitchErrorCalculator : bad input\n");
			return;
		}

		int nc = true_ph.ncol();

    for (int ri = 0, nr = true_ph.nrow(); ri < nr; ri += 2) {
      int ci = -1;
      while (ci < nc) {
        ci = NextKnownHet(ri, ci, true_ph, ph_known);
	      if (ci < 0) {
	      	break;
	      }

	      int ci2 = NextKnownHet(ri, ci, true_ph, ph_known);
	      if (ci2 >= 0) {
	        ++(*n_tests);
	        if (!PhaseCorrect(ri, ci, ci2, true_ph, test_ph)) {
	          ++(*n_errs);
	      	}
	      } else {
	      	break;
	      }
      }
    }
	}

protected:
	SwitchErrorCalculator() { }

protected:
	static bool CheckInputs(const CastableMatI& true_ph,
						              const CastableMatI& ph_known,
						              const CastableMatI& test_ph) {
		if (!true_ph.EqualDims(ph_known) || !true_ph.EqualDims(test_ph)) {
    	error("SwErrByDist : input matrices of unequal dimensions\n");
    	return(false);
    } else {
      return(true);
    }
	}

	static bool PhaseCorrect(int row_idx, int col_idx1, int col_idx2,
			                     const CastableMatI& true_ph,
			                     const CastableMatI& test_ph) {
    int a00 = true_ph.GetElt(row_idx, col_idx1);
    int a01 = true_ph.GetElt(row_idx, col_idx2);
    int a10 = true_ph.GetElt(row_idx + 1, col_idx1);
    int a11 = true_ph.GetElt(row_idx + 1, col_idx2);
    int b00 = test_ph.GetElt(row_idx, col_idx1);
    int b01 = test_ph.GetElt(row_idx, col_idx2);
    int b10 = test_ph.GetElt(row_idx + 1, col_idx1);
    int b11 = test_ph.GetElt(row_idx + 1, col_idx2);

    if (a00 != b00) {
    	std::swap(b00, b10);
    	std::swap(b01, b11);
    }
    if (a00 != b00 || a01 != b01 || a10 != b10 || a11 != b11) {
    	return(false);
    } else {
    	return(true);
    }
	}

	static int NextKnownHet(int row_idx, int col_idx,
			                    const CastableMatI& true_ph,
                          const CastableMatI& ph_known) {
    int nc = true_ph.ncol();

    for (int ci = col_idx + 1; ci < nc; ++ci) {
    	if (ph_known.GetElt(row_idx, ci) &&
    			true_ph.GetElt(row_idx, ci) != true_ph.GetElt(row_idx + 1, ci)) {
    		return(ci);
    	}
    }
    return(-1);
	}
};

class SwErrByDist : public SwitchErrorCalculator {
public:
	// bins = {x_1, x_2, x_3, ..., x_n}
	// where the elements must be sorted and x_n is the largest
	// distance considered.
  // n_errs and n_tests must be the same length as 'bins'
	static void Calc(const CastableMatI& true_ph,
			             const CastableMatI& ph_known,
			             const CastableMatI& test_ph,
			             const RealVec& posns,
			             const RealVec& bins,
			             IntVec * n_errs,
			             IntVec * n_tests) {
		if (!CheckInputs(true_ph, ph_known, test_ph,
				             posns, bins, *n_errs, *n_tests)) {
			return;
		}

    double mx_dist = SwErrByDist::MaxDist(bins);

    int nr = true_ph.nrow();
    int nc = true_ph.ncol();

    for (int ri = 0; ri < nr; ri += 2) {
      int ci = -1;

      while (true) {
      	ci = NextKnownHet(ri, ci, true_ph, ph_known);
      	if (ci < 0) {
	      	break;
	      }

      	int ci2 = ci;
        while (true) {
        	ci2 = NextKnownHet(ri, ci2, true_ph, ph_known);
        	if (ci2 < 0 || posns[ci2] - posns[ci] > mx_dist) {
            break;
          } else {
          	int bin_idx = FindBin(bins, posns[ci2] - posns[ci]);
          	if (bin_idx < 0) {
          		error("SwErrByDist : failed to find bin (bins out of order ?)\n");
          	}
          	++(*n_tests)[bin_idx];

          	if (!PhaseCorrect(ri, ci, ci2, true_ph, test_ph)) {
          		++(*n_errs)[bin_idx];
          	}
          }
      	}
      }
    }
	}

protected:
	SwErrByDist() { }

	static double MaxDist(const RealVec& bins) {
		return(bins[bins.size() - 1]);
	}

	static int FindBin(const RealVec& bins, double dist) {
		// assumes 'bins' is sorted and performs a binary search
    int lb = 0;
    int ub = bins.size() - 1;
    while (true) {
    	if (lb == ub) {
    		return(lb);
    	}
    	int mi = (lb + ub) / 2;
    	if (bins[mi] < dist) {
    		lb = mi + 1;
    	} else {
    		ub = mi;
    	}
    }
	}


	// If 'bins' = x_0 < x_1 < ... < x_n then we require that
	// n_errs.size() == n_tests.size() == n
	static bool CheckInputs(const CastableMatI& true_ph,
						              const CastableMatI& ph_known,
						              const CastableMatI& test_ph,
						              const RealVec& posns,
						              const RealVec& bins,
						              const IntVec& n_errs,
						              const IntVec& n_tests) {
		if (!SwitchErrorCalculator::CheckInputs(true_ph, ph_known, test_ph)) {
			return(false);
		} else if (bins.size() == 0) {
			error("SwErrByDist : bins.size() = 0\n");
	  } else if (n_errs.size() != n_tests.size() || n_errs.size() != bins.size()) {
    	error("SwErrByDist : n_errs or n_tests is of wrong length\n");
    } else if ((int)posns.size() != true_ph.ncol()) {
    	error("SwErrByDist : posns.size() != # markers\n");
    } else if (true_ph.nrow() % 2 != 0) {
    	error("SwErrByDist : number of rows must be even\n");
    } else {
    	return(true);
    }
    return(false);
	}
};

}  // namespace SwitchError

extern "C" {
	SEXP SwitchErrorCalc(SEXP true_ph_sxp, SEXP ph_known_sxp, SEXP test_ph_sxp,
				               SEXP n_errs_sxp, SEXP n_tests_sxp);

  SEXP SwitchErrByDist(SEXP true_ph_sxp, SEXP ph_known_sxp, SEXP test_ph_sxp,
  		                 SEXP posns_sxp, SEXP bins_sxp,
  		                 SEXP n_errs_sxp, SEXP n_tests_sxp);
}  // extern "C" {

# endif