PhaseEmControl = function() {
  return(list(tiedMut = TRUE,
              nCasesParamUpd = 500,
              hapOptAllowCrossover = TRUE,
              hapOptAllowSingleSite = TRUE,
              crossoverThresh = 0.3,
              hapOptThresh = 0.3,
              backTrackingOpt = TRUE))
}

PhaseEmModel = function(m, ctrl = NULL) {
  if (is.null(ctrl)) {
    ctrl = PhaseEmControl()
  }
  # makes an empty model
  nSnp = MatrixNcol(m)
  r = list(n_snp = nSnp,
           jump_probs = rep(0.01, nSnp),
           mut_probs = rep(0.01, nSnp),
           ctrl = ctrl)
  class(r) = "PhaseEm"
  return(r)
}

PhaseEmExpCounts = function(mdl, m, caseInds = NULL) {
  if (!is.null(caseInds)) {
    caseInds = as.integer(caseInds - 1)
  } else {
    caseInds = as.integer(1:MatrixNrow(m) - 1)
  }

  pLiks = rep(0.0, length(caseInds))
  jumpCounts = matrix(nrow = 2, ncol = mdl$n_snp, data = -1.0)
  mutCounts = matrix(nrow = 2, ncol = mdl$n_snp, data = -1.0)

  .Call("PhaseEmExpCounts_RI", mdl, m, caseInds,
        jumpCounts, mutCounts, pLiks, PACKAGE = "PHASEEM")

  jumpCounts[2, mdl$n_snp] = length(caseInds)
  jumpCounts[1, mdl$n_snp] = 0

  return(list(jumpCounts = jumpCounts, mutCounts = mutCounts,
              pLiks = pLiks))
}

PhaseEmParamUpd = function(mdl, x, caseInds = NULL, nIter = 1,
                           returnAll = FALSE, verbose = FALSE) {
  nr = MatrixNrow(x)
  v = vector(mode = "list")
  caseInds0 = 1:nr
  
  for (iterIdx in 1:nIter) {
    if (verbose) print(sprintf("iteration %d", iterIdx))
    
    if (is.null(caseInds)) {
      if (mdl$ctrl$nCasesParamUpd < nr) {
        caseInds0 = sample(1:nr, mdl$ctrl$nCasesParamUpd)
      }
    } else {
      caseInds0 = caseInds
    }
  
    cts = PhaseEmExpCounts(mdl, x, caseInds0)
    dnm = cts$jumpCounts[1, ] + cts$jumpCounts[2, ]
    mdl$jump_probs = cts$jumpCounts[2, ] / dnm
    
    if (mdl$ctrl$tiedMut) {
      mdl$mut_probs[] = sum(cts$mutCounts[2, ]) / sum(cts$mutCounts)
    } else {
      dnm = cts$mutCounts[1, ] + cts$mutCounts[2, ]
      mdl$mut_probs = cts$mutCounts[2, ] / dnm
    }
    
    if (returnAll) {
      v[[iterIdx]] = list(mdl = mdl, caseInds = caseInds0, cts = cts)
    }
  }

  if (returnAll) {
    return(v)
  } else {
    return(mdl)
  }
}

PhaseEmCrossoverScores = function(mdl, m, caseIdx1, caseIdx2) {
  # computes the crossover confidence measure
  if (caseIdx1 <= 0 || caseIdx1 > MatrixNrow(m)) {
    stop("caseIdx1 invalid")
  } else if (caseIdx2 <= 0 || caseIdx2 > MatrixNrow(m)) {
    stop("caseIdx2 invalid")
  }
  r = .Call("PhaseEmCrossoverScores_RI", mdl, m, caseIdx1 - 1, caseIdx2 - 1,
            PACKAGE = "PHASEEM")
  return(r)
}

PhaseEmGenotypeCodes = function(x, missMat = NULL) {
  r = ByteMat(MatrixNrow(x), MatrixNcol(x))
  .Call("PhaseEmGenotypeCodes_RI", x, r, PACKAGE = "PHASEEM")
  
  if (!is.null(missMat)) {
    .Call("PhaseEmSetMissGenotypeCodes_RI", missMat, r, PACKAGE = "PHASEEM")
  }
  return(r)
}

HapOpt = function(mdl, ...) {
  UseMethod("HapOpt", mdl)
}

HapOpt.PhaseEm = function(mdl, ...) {
  return(PhaseEmHapOpt(mdl, ...))  
}

PhaseEmHapOpt = function(mdl, x, gtypes = NULL,
                         caseInds = NULL,
                         allowCrossover = NULL, 
                         allowSingleSite = NULL,
                         moveThresh = NULL,
                         backTracking = NULL,
                         deepCopy = TRUE) {
  if (deepCopy) {
    x = MatrixSlice(x)
  }
  if (is.null(gtypes)) {
    gtypes = PhaseEmGenotypeCodes(x)
  }

  if (is.null(backTracking)) {
    if (!("backTrackingOpt" %in% names(mdl$ctrl))) {
      backTracking = TRUE
    } else {
      backTracking = mdl$ctrl$backTrackingOpt
    }
  }
  
  nr = MatrixNrow(x)
  if (is.null(caseInds)) {
    caseInds = matrix(0:(nr-1), nrow = 2)
  } else {
    caseInds = matrix(as.integer(caseInds-1), nrow = 2)
  }
  
  if (is.null(moveThresh)) moveThresh = mdl$ctrl$hapOptThresh
  if (is.null(allowCrossover)) allowCrossover = mdl$ctrl$hapOptAllowCrossover
  if (is.null(allowSingleSite)) allowSingleSite = mdl$ctrl$hapOptAllowSingleSite
  
  .Call("PhaseEmHapOpt_RI", mdl, x, caseInds, moveThresh, allowCrossover,
        allowSingleSite, gtypes, backTracking, PACKAGE = "PHASEEM")
  return(x)
}

PhaseEmCrossoverOpt = function(mdl, x, caseInds = NULL,
                               moveThresh = NULL,
                               backTracking = NULL,
                               deepCopy = TRUE) {
  if (deepCopy) {
    x = MatrixSlice(x)
  }

  nr = MatrixNrow(x)
  if (is.null(caseInds)) {
    caseInds = matrix(0:(nr-1), nrow = 2)
  } else {
    caseInds = matrix(as.integer(caseInds-1), nrow = 2)
  }
  
  if (is.null(backTracking)) {
    if (!("backTrackingOpt" %in% names(mdl$ctrl))) {
      backTracking = TRUE
    } else {
      backTracking = mdl$ctrl$backTrackingOpt
    }
  }
  
  if (is.null(moveThresh)) moveThresh = mdl$ctrl$crossoverThresh
  .Call("PhaseEmCrossoverOpt_RI", mdl, x, caseInds, moveThresh, backTracking,
        PACKAGE = "PHASEEM")
  return(x)
}

PhaseEmAllOpt = function(mdl, x, nIter = 1, gtypes = NULL,
                         paramOpt = TRUE, crossOver = FALSE,
                         hapOpt = TRUE) {
  if (hapOpt && is.null(gtypes)) {
    gtypes = PhaseEmGenotypeCodes(x)
  }
  
  for (iterIdx in 1:nIter) {
    if (crossOver) {
      x = PhaseEmCrossoverOpt(mdl, x)
    }
    if (hapOpt) {
      x = PhaseEmHapOpt(mdl, x, gtypes = gtypes)
    }
    if (paramOpt) {
      mdl = PhaseEmParamUpd(mdl, x)
    }
  }
  return(list(mdl = mdl, x = x))
}