#include "aradmx_phased.hpp"

//#define EXACT_INF_DEBUG_CHECKS 1
namespace Aradmx {
namespace Phased {

void PhasedModelParams::Fill(SEXP drop_lookups_sxp, SEXP max_drop_sxp,
				          SEXP state_mat_sxp, SEXP win_sz_sxp,              // The state matrix
				          SEXP L2RMap, SEXP R2LMap,     // mappings to absorb messages
				          SEXP anc_transprb_sxp, SEXP anc_priors_sxp, // Recombination and ancestral prior parameters
				          SEXP bin_trees_sxp)    // Parent sets and their observed counts/multinomial parameters
{
	using namespace Util;
  this->ModelParams::Fill(drop_lookups_sxp, max_drop_sxp, win_sz_sxp,
  		                    L2RMap, R2LMap, anc_transprb_sxp,
  		                    anc_priors_sxp, bin_trees_sxp);

  GetMatrixDim(state_mat_sxp, &state_size_, &n_states_);
  int * state_mat_int = INTEGER(state_mat_sxp);

  state_mat_ = vector<State>(n_states_);
  for (int i = 0; i < n_states_; ++i, state_mat_int += state_size_) {
  	int ji = state_mat_int[0];
  	ji = (ji == 0) ? (win_sz_ - 1) : (ji-1);
  	state_mat_[i].anc_  = state_mat_int[1];
  	state_mat_[i].jump_idx_  = ji;

  	if (state_mat_int[0] < 0 || state_mat_int[1] < 0
  			|| state_mat_int[1] >= n_pops_) {
  		Util::ThrowError("PhasedEMModelParams : value in state matrix is out of range\n");
  	}
  }

  if (state_size_ != 2) {
  	Util::ThrowError("PhasedEMModelParams : state matrix has wrong dimensions\n");
  }

  this->CheckFwdBkwdMappings();
}

void PhasedModelParams::FirstSnpZero(double * r) {
	StateMat::const_iterator sm = state_mat_.begin();
  for (; sm != state_mat_.end(); ++r, ++sm) {
    if (sm->jump_idx_ != 0) {
      *r *= 0.0;
    }
	}
}

void PhasedModelParams::GetAncPrbs(int snp_idx, const double * cur_pttl, double * anc_prbs) const {
	if (anc_prbs) {
	  double * a0 = anc_prbs + n_pops_ * snp_idx;

	  for (int pop_idx = 0; pop_idx < n_pops_; ++pop_idx) {
	  	a0[pop_idx] = 0.0;
	  }

	  StateMat::const_iterator sm = state_mat_.begin();
	  for (; sm != state_mat_.end(); ++sm, ++cur_pttl) {
	  	a0[sm->anc_] += *cur_pttl;
	  }
	}
}

void PhasedModelParams::CondProbs(const IntVec& pop_lab, int snp_idx, double * r) {
  double * jump_pr = anc_jmp_ + (2 * snp_idx);

  StateMat::const_iterator sm = state_mat_.begin();
  for (; sm != state_mat_.end(); ++sm, ++r) {
  	if (!pop_lab[sm->anc_]) {
  		*r = 0.0;
  	} else {
	  	if (sm->jump_idx_ == 0) {
	  		// There was a ancestral state transition at this snp
	  		*r = jump_pr[1] * grp_prior_[sm->anc_];
	  	} else {
	  		*r = jump_pr[0];
	  	}
	    const int * tr = GetTree(snp_idx, sm->anc_, sm->jump_idx_);
	    int ct_idx = PhasedTreeTraverse(hap_, tr, win_sz_);
	    double p2 = *((double *)(tr + (ct_idx + 1)));

#ifdef EXACT_INF_DEBUG_CHECKS
	    if (p2 < 0.0 || p2 > 1.0) {
	    	Util::ThrowError("CondProbs : emission probability out of range\n");
	    }
#endif

	    *r *= (hap_[snp_idx]) ? p2 : (1.0 - p2);
  	}
  }
}

}  // namespace Aradmx {
}  // namespace Phased {


extern "C" {

SEXP AradmxPhasedExactFwdBkwd_RI(SEXP hap_sxp, SEXP drop_lookups_sxp, SEXP max_drop_sxp,
                              SEXP junc_tree_sxp, SEXP win_sz_sxp,              // The state matrix
                              SEXP anc_transprb_sxp, SEXP anc_priors_sxp, // Recombination and ancestral prior parameters
                              SEXP bin_trees_sxp,    // Parent sets and their observed counts/multinomial parameters
                              SEXP pop_lab_sxp,
                              SEXP checkpointed_sxp,
                              SEXP marg_lik_sxp, SEXP anc_prbs_sxp)           // The output of marginal likelihoods
{
	using namespace Aradmx::FwdBkwd;
	using namespace Aradmx::Phased;
	using namespace Util;

	// pop_lab_sxp is an integer vector of length K where K is the number of populations
	// pop_lab_sxp[j] = TRUE means that the sample can belong to the j'th population
	// and a value of FALSE indicates that it cannot belong to that population.
	//
	// If the population label is unknown pass in vector of TRUE's.

	IntVec pop_lab;
	SxpToIntVec(pop_lab_sxp, &pop_lab);
	Hap hap;
	HapFromSxp(hap_sxp, &hap);

	Util::SxpList jt_sxp(junc_tree_sxp);
	SEXP state_mat_sxp = jt_sxp.Sxp("stateMat");
	SEXP leftright_map_sxp = jt_sxp.Sxp("L2RMap");
	SEXP rightleft_map_sxp = jt_sxp.Sxp("R2LMap");

  double obs_wt = 1.0;
  PhasedModelParams cfg;
  cfg.Fill(drop_lookups_sxp, max_drop_sxp, state_mat_sxp, win_sz_sxp,
  		     leftright_map_sxp, rightleft_map_sxp, anc_transprb_sxp,
  		     anc_priors_sxp, bin_trees_sxp);
  cfg.set_hap(hap);

  if (pop_lab.size() != cfg.n_pops_) {
  	Util::ThrowError("pop_lab_sxp size is wrong\n");
  } else if (LENGTH(hap_sxp) != cfg.n_snp_) {
  	Util::ThrowError("AbbnUnphasedFwdBkwd : input haplotypes are wrong length\n");
  }

  double * anc_prbs = (anc_prbs_sxp == R_NilValue) ? NULL : REAL(anc_prbs_sxp);
  double * marg_lik = (marg_lik_sxp == R_NilValue) ? NULL : REAL(marg_lik_sxp);
  double ml = cfg.FwdBkwdAlgo(pop_lab, obs_wt, GetInt(checkpointed_sxp, 1, NULL), anc_prbs);
  if (marg_lik) *marg_lik = ml;
  return R_NilValue;
}

}  // extern "C"
