#ifndef ARADMX_UNPH_STATEMAT_HPP_
#define ARADMX_UNPH_STATEMAT_HPP_

#include "util.hpp"
#include "aradmx.hpp"

namespace Aradmx {
namespace UnphStatemat {

typedef int Field;

__inline bool NoJump(int jump_index) {
	return(jump_index == 0);
};

__inline bool RecentJump(int jump_index) {
	return(jump_index == 1);
};

class UnphApproxStateStrand {
public:
  // ancestral groups at the right-most position
	Field anc_;
  // indicator for whether or not there was an ancestral
  // jump
	Field anc_jmp_;

	void ToBuf(int * buf) {
	  buf[0] = anc_;
	  buf[1] = anc_jmp_;
	}
	void FromBuf(int * buf) {
	  anc_ = buf[0];
	  anc_jmp_ = buf[1];
	}

	void ConsolePrint() const {
	  Util::ConsolePrint("anc_ = %d\n", anc_);
	  Util::ConsolePrint("anc_jmp_ = %d\n", anc_jmp_);
	}

	static bool EltsCompat(int win_sz, int left_jmp_idx, int right_jmp_idx,
	                       const UnphApproxStateStrand& left_elt,
	                       const UnphApproxStateStrand& right_elt) {
	  if (left_jmp_idx == 0 || left_jmp_idx == win_sz - 1) {
	    if (right_elt.anc_jmp_ != 1 && left_elt.anc_ != right_elt.anc_) {
	      return false;
	    }
	  } else {
	    if (left_elt.anc_jmp_ != right_elt.anc_jmp_ || left_elt.anc_ != right_elt.anc_) {
	      return false;
	    }
	  }
	  return true;
	}

};

// The state matrix
class UnphApproxState {
public:

  // the two phasing orientations (0 or 1)
	Field phase_[2];
	UnphApproxStateStrand strands_[2];
  Field jmp_idx_;

  static const int kStateSize = 7;

  void Set(int o1, int o2, int jmpIdx, int g1, int g1jmp, int g2, int g2jmp) {
	  phase_[0] = o1;
	  phase_[1] = o2;
	  strands_[0].anc_ = g1;
	  strands_[1].anc_ = g2;
	  strands_[0].anc_jmp_ = g1jmp;
	  strands_[1].anc_jmp_ = g2jmp;
	  jmp_idx_ = jmpIdx;
  }
  void ToBuf(int * buf) {
    *buf = jmp_idx_; ++buf;
    *buf = phase_[0]; ++buf;
    *buf = phase_[1]; ++buf;
    strands_[0].ToBuf(buf);
    buf += 2;
    strands_[1].ToBuf(buf);
  }

  virtual void FromBuf(int * buf) {
    jmp_idx_ = *buf; ++buf;
    phase_[0] = *buf; ++buf;
    phase_[1] = *buf; ++buf;
    strands_[0].FromBuf(buf);
    buf += 2;
    strands_[1].FromBuf(buf);
  }
	void ConsolePrint() const {
	  Util::ConsolePrint("jmp_idx_ = %d\n", jmp_idx_);
	  Util::ConsolePrint("phase_ = (%d, %d)\n", phase_[0], phase_[1]);
	  strands_[0].ConsolePrint();
	  strands_[1].ConsolePrint();
	}

	static vector<UnphApproxState> StatesFromBuf(int * buf, int n_states) {
		vector<UnphApproxState> r(n_states);
	  for (int i = 0; i < n_states; ++i, buf += kStateSize) {
	    r[i].FromBuf(buf);
	  }
	  return r;
	}

	static bool StateValid(bool reduced, int o1, int o2, int jmpIdx, int g1,
			                   int g1jmp, int g2, int g2jmp) {
	  if (reduced && g1 > g2) {
	    return false;
	  }
	  if (jmpIdx == 0) {
	    return (((o1 == o2) && (g1jmp == 0) && (g2jmp == 0)) ? 1 : 0);
	  } else {
	    return true;
	  }
	}

	static int GetStates(bool reduced, int n_groups, int win_sz,
			                 vector<UnphApproxState> * vec) {
		// Call this function with vec = NULL to get the number of valid states
		// then create the buffer and call it again to fill out the state matrix
	  int n_elt = 0;
	  for (int o1 = 0; o1 < 2; ++o1) {
	  for (int o2 = 0; o2 < 2; ++o2) {
	  for (int jmp_idx = 0; jmp_idx < win_sz; ++jmp_idx) {
	  for (int g1 = 0; g1 < n_groups; ++g1) {
	  for (int g1_jmp = 0; g1_jmp < 2; ++g1_jmp) {
	  for (int g2 = 0; g2 < n_groups; ++g2) {
	  for (int g2_jmp = 0; g2_jmp < 2; ++g2_jmp) {
	    if (UnphApproxState::StateValid(reduced, o1, o2, jmp_idx, g1,
	    		                            g1_jmp, g2, g2_jmp)) {
	      if (vec) {
	        (*vec)[n_elt].Set(o1, o2, jmp_idx, g1, g1_jmp, g2, g2_jmp);
	      }
	      ++n_elt;
	    }
	  }}}}}}}

	  return n_elt;
	}

	static bool EltsCompat(int win_sz, const UnphApproxState& left_elt,
			                   const UnphApproxState& right_elt) {
	  int left_jmp = left_elt.jmp_idx_;
	  int right_jmp = right_elt.jmp_idx_;

	  int compat = 1;
	  for (int strd = 0; strd < 2; ++strd) {
	    if (0 == UnphApproxStateStrand::EltsCompat(win_sz, left_jmp, right_jmp,
	                                           left_elt.strands_[strd],
	                                           right_elt.strands_[strd])) {
	      compat = 0;
	    }
	  }

	  if (compat == 0) {
	    return false;
	  }
	  if (left_jmp == 0 || left_jmp == win_sz - 1) {
	    if (right_jmp != 0 && right_jmp != 1) {
	      return false;
	    } else if (right_elt.phase_[0] != left_elt.phase_[1]) {
	      return false;
	    }
	    return true;
	  } else {
	    if (left_jmp + 1 != right_jmp) {
	      return false;
	    } else if (left_elt.phase_[0] != right_elt.phase_[0]
	               || left_elt.phase_[1] != right_elt.phase_[1]) {
	      return false;
	    }
	    return true;
	  }
	}

	//
	static bool AreMirrorStates(UnphApproxState& s1, UnphApproxState& s2) {
	  for (int i = 0; i < 2; ++i) {
	    if (s1.phase_[i] != 1-s2.phase_[i]) return false;
	    if (s1.strands_[1-i].anc_ != s2.strands_[i].anc_) return false;
	    if (s1.strands_[1-i].anc_jmp_ != s2.strands_[i].anc_jmp_) return false;
	  }
	  return (s1.jmp_idx_ == s2.jmp_idx_);
	}

	static void GetMirrorStates(vector<UnphApproxState>& sm, int * smap) {
		int n_states = sm.size();
	  for (int i = 0; i < n_states; ++i) {
	    smap[i] = -1;
	  }
	  for (int i = 0; i < n_states; ++i) {
	    for (int j = 0; j < n_states; ++j) {
	      if (UnphApproxState::AreMirrorStates(sm[i], sm[j])) {
	        if (smap[i] != -1) {
	          Util::ThrowError("GetMirrorStates : multiple mirror states "
	          		             "(%d, %d, %d).  n_states = %d\n",
	          		              i, j, smap[i], n_states);
	        }
	        smap[i] = j;
	      }
	    }
	  }
	}
};

}  // namespace
}  // namespace

extern "C" {

SEXP UnphApproxStatesSxp_RI(SEXP nGroupsSxp, SEXP winSzSxp, SEXP reducedSxp,
		                        SEXP retSxp);

SEXP ReduceUnphApproxStatesSxp_RI(SEXP stateMat, SEXP l2rMapSxp,
		                                SEXP r2lMapSxp, SEXP retSxp);

}

#endif /*ARADMX_UNPH_STATEMAT_HPP_*/
