#include "aradmx_unphased_approx.hpp"

//#define EXACT_INF_DEBUG_CHECKS 1
namespace Aradmx {
namespace UnphApprox {

void UnphasedApproxModelParams::Fill(SEXP drop_lookups_sxp, SEXP max_drop_sxp,
				          SEXP state_mat_sxp, SEXP win_sz_sxp,              // The state matrix
				          SEXP L2RMap, SEXP R2LMap,     // mappings to absorb messages
				          SEXP anc_transprb_sxp, SEXP anc_priors_sxp, // Recombination and ancestral prior parameters
				          SEXP bin_trees_sxp)    // Parent sets and their observed counts/multinomial parameters
{
	using namespace Util;
  this->ModelParams::Fill(drop_lookups_sxp, max_drop_sxp, win_sz_sxp,
  		                    L2RMap, R2LMap, anc_transprb_sxp,
  		                    anc_priors_sxp, bin_trees_sxp);

  GetMatrixDim(state_mat_sxp, &state_size_, &n_states_);
  int * state_mat_int = INTEGER(state_mat_sxp);
  state_mat_ = State::StatesFromBuf(state_mat_int, n_states_, win_sz_);

  tree_lookup_ = vector<const int*>(win_sz_ * n_pops_, (const int*)NULL);
}

void UnphasedApproxModelParams::ClearTreeLookups() {
	vector<const int*>::iterator i = tree_lookup_.begin();
  for (; i != tree_lookup_.end(); ++i) {
  	*i = NULL;
  }
}

void UnphasedApproxModelParams::FirstSnpZero(double * r) {
	StateMat::const_iterator sm = state_mat_.begin();
  for (; sm != state_mat_.end(); ++r, ++sm) {
    if (sm->strands_[0].anc_jmp_ != 0 || sm->strands_[1].anc_jmp_ != 0) {
      *r *= 0.0;
    }
	}
}

void UnphasedApproxModelParams::GetAncPrbs(int snp_idx, const double * cur_pttl, double * anc_prbs) const {
	if (anc_prbs) {
	  double * a0 = anc_prbs + n_pops_ * snp_idx;

	  for (int pop_idx = 0; pop_idx < n_pops_; ++pop_idx) {
	  	a0[pop_idx] = 0.0;
	  }

	  StateMat::const_iterator sm = state_mat_.begin();
	  for (; sm != state_mat_.end(); ++sm, ++cur_pttl) {
	  	a0[sm->strands_[0].anc_] += *cur_pttl;
	  	a0[sm->strands_[1].anc_] += *cur_pttl;
	  }
	}
}

void UnphasedApproxModelParams::CondProbs(const IntVec& pop_lab, int snp_idx, double * r) {
	this->ClearTreeLookups();
  double * jump_pr = anc_jmp_ + (2 * snp_idx);

  int cur_base[4] = { haps_[0][snp_idx], haps_[1][snp_idx],
  		               haps_[1][snp_idx], haps_[0][snp_idx] };

  StateMat::const_iterator sm = state_mat_.begin();
  for (; sm != state_mat_.end(); ++sm, ++r) {
  	*r = 1.0;
  	for (int strand_idx = 0; strand_idx < 2; ++strand_idx) {
	  	int cur_anc = sm->strands_[strand_idx].anc_;
	  	if (!pop_lab[cur_anc]) {
	  		*r = 0.0;
	  	} else {
	  		int ph1 = sm->phase_[1];
	  		int anc_j = sm->anc_jump_idx_[strand_idx];

	  		double prb_mlt = (anc_j) ? jump_pr[0] : (jump_pr[1] * grp_prior_[cur_anc]);
		    const int * tr = tree_lookup_[cur_anc * win_sz_ + anc_j];
		    if (!tr) {
		    	tr = tree_lookup_[cur_anc * win_sz_ + anc_j] = GetTree(snp_idx, cur_anc, anc_j);
		    }
		    int ct_idx = TreeTraverseUnphApprox(
		    		             hap_pair_, strand_idx, sm->phase_[0], ph1,
		    		             snp_idx - sm->jump_idx_ - 1, tr);

		    double p2 = *((double *)(tr + (ct_idx + 1)));

//#ifdef EXACT_INF_DEBUG_CHECKS
//		    if (p2 < 0.0 || p2 > 1.0) {
//		    	Util::ThrowError("CondProbs : emission probability out of range\n");
//		    }
//#endif

		    prb_mlt *= (cur_base[2 * ph1 + strand_idx]) ? p2 : (1.0 - p2);
		    *r *= prb_mlt;
	  	}
  	}
  }
}

}  // namespace Aradmx {
}  // namespace Phased {


extern "C" {

SEXP AradmxUnphApproxFwdBkwd_RI(SEXP hap1_sxp, SEXP hap2_sxp,
		                          SEXP drop_lookups_sxp, SEXP max_drop_sxp,
                              SEXP junc_tree_sxp, SEXP win_sz_sxp,              // The state matrix
                              SEXP anc_transprb_sxp, SEXP anc_priors_sxp, // Recombination and ancestral prior parameters
                              SEXP bin_trees_sxp,    // Parent sets and their observed counts/multinomial parameters
                              SEXP pop_lab_sxp,
                              SEXP checkpointed_sxp,
                              SEXP marg_lik_sxp, SEXP anc_prbs_sxp)           // The output of marginal likelihoods
{
	using namespace Aradmx::FwdBkwd;
	using namespace Aradmx::UnphApprox;
	using namespace Util;

	// pop_lab_sxp is an integer vector of length K where K is the number of populations
	// pop_lab_sxp[j] = TRUE means that the sample can belong to the j'th population
	// and a value of FALSE indicates that it cannot belong to that population.
	//
	// If the population label is unknown pass in vector of TRUE's.

	IntVec pop_lab;
	SxpToIntVec(pop_lab_sxp, &pop_lab);
	Hap haps[2];
	HapFromSxp(hap1_sxp, &(haps[0]));
	HapFromSxp(hap2_sxp, &(haps[1]));


	Util::SxpList jt_sxp(junc_tree_sxp);
	SEXP state_mat_sxp = jt_sxp.Sxp("stateMat");
	SEXP leftright_map_sxp = jt_sxp.Sxp("L2RMap");
	SEXP rightleft_map_sxp = jt_sxp.Sxp("R2LMap");

  double obs_wt = 1.0;
  UnphasedApproxModelParams cfg;
  cfg.Fill(drop_lookups_sxp, max_drop_sxp, state_mat_sxp, win_sz_sxp,
  		     leftright_map_sxp, rightleft_map_sxp, anc_transprb_sxp,
  		     anc_priors_sxp, bin_trees_sxp);
  cfg.set_haps(haps[0], haps[1]);

  if (pop_lab.size() != cfg.n_pops_) {
  	Util::ThrowError("pop_lab_sxp size is wrong\n");
  } else if (LENGTH(hap1_sxp) != cfg.n_snp_ || LENGTH(hap2_sxp) != cfg.n_snp_) {
  	Util::ThrowError("AbbnUnphasedFwdBkwd : input haplotypes are wrong length\n");
  }

  double * anc_prbs = (anc_prbs_sxp == R_NilValue) ? NULL : REAL(anc_prbs_sxp);
  double * marg_lik = (marg_lik_sxp == R_NilValue) ? NULL : REAL(marg_lik_sxp);
  double ml = cfg.FwdBkwdAlgo(pop_lab, obs_wt, GetInt(checkpointed_sxp, 1, NULL), anc_prbs);
  if (marg_lik) *marg_lik = ml;
  return R_NilValue;
}

}  // extern "C"
