#include "util.hpp"
#include "bin_tree.hpp"

namespace BinTree {

using namespace Util;

BaseTree::BaseTree(const BinaryMat& dat, const IntVec& pop_labels,
		               int target_var) : dat_(dat), pop_labels_(pop_labels),
		                                 target_var_(target_var) {
	n_obs_ = dat_.ncol();
	split_pcount_ = 1.0;
	min_leaf_sz_ = 0;
	max_split_depth_ = 1000;
	pop_wtd_score_ = false;

	if (n_obs_ != (int)pop_labels_.size()) {
		Util::ThrowError("BaseTree::BaseTree : LENGTH(pop_labels) is wrong\n");
	}

	int max_pop_idx = 0;
	for (int i = 0; i < n_obs_; ++i) {
		if (pop_labels_[i] > max_pop_idx) {
			max_pop_idx = pop_labels_[i];
		}
	}

	pop_wts_ = RealVec(max_pop_idx + 1, 1.0 / (double)(max_pop_idx+1));

	this->ClearTree();
}

void BaseTree::ClearTree() {
	tree_.clear();
	tree_.push_back(TreeNode());

	tree_[0].pt_set_ = IntVec(n_obs_);

	for (int i = 0; i < n_obs_; ++i) {
		tree_[0].pt_set_[i] = i;
	}
}

double BaseTree::GetSinglePopNoSplitScore(int node_idx, int pop_idx) const {
	if (node_idx < 0 || node_idx >= (int)tree_.size()) {
		Util::ThrowError("BaseTree::SplitNode : node_idx not valid.\n");
	}

	const TreeNode& nd = tree_[node_idx];
	const IntVec& pts = nd.pt_set_;

	IntVec::const_iterator itr;

	// makes some formulas below a little shorter
	double pc = split_pcount_;
	double n[2] = {pc, pc};

	for (itr = pts.begin(); itr != pts.end(); ++itr) {
		if (pop_idx < 0 || pop_labels_[*itr] == pop_idx) {
			n[dat_.GetElt(target_var_, *itr)]++;
		}
	}

	double n0 = n[0], n1 = n[1];
	double llik0 = (n0 - pc) * log(n0 / (n0 + n1));
	llik0 += (n1 - pc) * log(n1 / (n0 + n1));

  return(llik0);
}

double BaseTree::GetNoSplitScore(int node_idx) const {
	double llik = 0.0;

	if (pop_wtd_score_) {
		for (int pop_idx = 0, n_pop = pop_wts_.size(); pop_idx < n_pop; ++pop_idx) {
			llik += pop_wts_[pop_idx] * this->GetSinglePopNoSplitScore(node_idx, pop_idx);
		}
	} else {
		llik = this->GetSinglePopNoSplitScore(node_idx, -1);
	}
	return(llik);
}


double BaseTree::GetSinglePopSplitScore(int node_idx, int split_var,
		                                    int pop_idx) const {
	if (node_idx < 0 || node_idx >= (int)tree_.size()) {
		Util::ThrowError("BaseTree::SplitNode : node_idx not valid.\n");
	}

	const TreeNode& nd = tree_[node_idx];
	const IntVec& pts = nd.pt_set_;

	if ((int)pts.size() < min_leaf_sz_ || nd.depth_ >= max_split_depth_) {
		return(R_NegInf);
	}

	IntVec::const_iterator itr;

	// makes some formulas below a little shorter
	double pc = split_pcount_;
	double n[2][2] = {{pc, pc}, {pc, pc}};

	if (this->IsPopSplitVar(split_var)) {
		int split_pop = this->SplitVarToPopIdx(split_var);

		for (itr = pts.begin(); itr != pts.end(); ++itr) {
			int i1 = dat_.GetElt(target_var_, *itr);
			int i2 = (pop_labels_[*itr] == split_pop) ? 1 : 0;
			n[i1][i2]++;
		}
	} else {
		for (itr = pts.begin(); itr != pts.end(); ++itr) {
			if (pop_idx >= 0 && pop_labels_[*itr] != pop_idx) continue;

			int i1 = dat_.GetElt(target_var_, *itr);
			int i2 = dat_.GetElt(split_var, *itr);
			n[i1][i2]++;
		}
	}

	double n00 = n[0][0], n01 = n[0][1], n10 = n[1][0], n11 = n[1][1];

	// 1e-8 is large enough for potential rounding error
	double min_sz = 2.0 * pc + 1e-8;

	double dnm0 = n00 + n10;
	double dnm1 = n01 + n11;

	// check that one of the leaves is not empty
	if (dnm0 < min_sz || dnm1 < min_sz) {
		return(R_NegInf);
	}

	double llik0 = (n00 - pc) * log(n00 / dnm0);
	llik0 += (n10 - pc) * log(n10 / dnm0);
	llik0 += (n01 - pc) * log(n01 / dnm1);
	llik0 += (n11 - pc) * log(n11 / dnm1);

	return(llik0);
}

double BaseTree::GetSplitScore(int node_idx, int split_var) const {
	const TreeNode& nd = tree_[node_idx];
	const IntVec& pts = nd.pt_set_;

	if (!nd.IsLeaf() || (int)pts.size() < min_leaf_sz_
			|| nd.depth_ >= max_split_depth_) {
		return(R_NegInf);
	}

	double llik = 0.0;

	if (pop_wtd_score_) {
		for (int pop_idx = 0, n_pop = pop_wts_.size(); pop_idx < n_pop; ++pop_idx) {
			llik += pop_wts_[pop_idx]
			        * this->GetSinglePopSplitScore(node_idx, split_var, pop_idx);
		}
	} else {
		llik += this->GetSinglePopSplitScore(node_idx, split_var, -1);
	}
	return(llik);
}

void BaseTree::SplitNode(int node_idx, int split_var) {
	if (node_idx < 0 || node_idx >= tree_.size()) {
		Util::ThrowError("BaseTree::SplitNode : node_idx not valid.\n");
	} else if (tree_[node_idx].pt_set_.size() == 0) {
		Util::ThrowError("attempting to split empty node\n");
	}
	int ch_idx0 = tree_.size();
	int ch_idx1 = tree_.size() + 1;

	tree_.push_back(TreeNode());
	tree_.push_back(TreeNode());

	TreeNode& nd = tree_[node_idx];
	TreeNode& ch0 = tree_[ch_idx0];
	TreeNode& ch1 = tree_[ch_idx1];

	ch0.depth_ = nd.depth_ + 1;
	ch1.depth_ = nd.depth_ + 1;

	const IntVec& pts = nd.pt_set_;

	IntVec::const_iterator itr;
	if (this->IsPopSplitVar(split_var)) {
		int split_pop = this->SplitVarToPopIdx(split_var);
		for (itr = pts.begin(); itr != pts.end(); ++itr) {
			if (pop_labels_[*itr] == split_pop) {
				ch1.pt_set_.push_back(*itr);
			} else {
				ch0.pt_set_.push_back(*itr);
			}
		}
	} else {
		for (itr = pts.begin(); itr != pts.end(); ++itr) {
			if (dat_.GetElt(split_var, *itr)) {
				ch1.pt_set_.push_back(*itr);
			} else {
				ch0.pt_set_.push_back(*itr);
			}
		}
	}
	if (ch0.pt_set_.size() == 0 || ch0.pt_set_.size() == pts.size()) {
		Util::ConsolePrint("BaseTree::SplitNode : split_var = %d.  "
				               "target_var_ = %d, pts.size() = %d\n",
				               split_var, target_var_, pts.size());
		Util::ConsolePrint("BaseTree::SplitNode : is_pop_split = %d\n",
				               this->IsPopSplitVar(split_var) ? 1 : 0);
		Util::ThrowError("BaseTree::SplitNode : made a split with empty leaf.\n");
	}
	nd.split_var_ = split_var;
	nd.ch0_idx_ = ch_idx0;
	nd.ch1_idx_ = ch_idx1;
}

void BaseTree::NodeTargetCounts(int node_idx, double * n0, double * n1) const {
	const TreeNode& nd = tree_[node_idx];
	const IntVec& pts = nd.pt_set_;

	double a = 0, b = 0;

	if (pts.size()) {
		IntVec::const_iterator itr;
		for (itr = pts.begin(); itr != pts.end(); ++itr) {
			if (dat_.GetElt(target_var_, *itr)) {
				++a;
			} else {
				++b;
			}
		}
	}
	if (n0) *n0 = a;
	if (n1) *n1 = b;
}

void BaseTree::GetSplitStream(IntVec * splits, RealVec * leaf_counts) const {
	for (int nd_idx = 0, sz = tree_.size(); nd_idx < sz; ++nd_idx) {
		if (this->IsPopSplitVar(tree_[nd_idx].split_var_)
				|| tree_[nd_idx].ch0_idx_ != BaseTree::kNoSplit) {
			// a -1 indicates this is a leaf node
			splits->push_back(tree_[nd_idx].split_var_);
			splits->push_back(3 * tree_[nd_idx].ch0_idx_);
			splits->push_back(3 * tree_[nd_idx].ch1_idx_);
		} else if (tree_[nd_idx].split_var_ == BaseTree::kNoSplit) {
			// it is a leaf node
			double x = 0, y = 0;
			this->NodeTargetCounts(nd_idx, &x, &y);
			// a -1 indicates this is a leaf node
			splits->push_back(-1);
			splits->push_back(leaf_counts->size());
			splits->push_back(leaf_counts->size() + 1);

			leaf_counts->push_back(x);
			leaf_counts->push_back(y);
		} else {
			error("BaseTree::GetSplitStream : tree_[%d].split_var_ = %d\n",
					  nd_idx, tree_[nd_idx].split_var_);
		}
	}
}

void BaseTree::GetSinglePopSplitStream(IntVec * splits, RealVec * leaf_counts,
		                                   int pop_idx) const {
	for (int nd_idx = 0, sz = tree_.size(); nd_idx < sz; ++nd_idx) {
		if (this->IsPopSplitVar(tree_[nd_idx].split_var_)) {
			if (pop_idx >= 0) {
				int ni = nd_idx;
				while (this->IsPopSplitVar(tree_[ni].split_var_)) {
					int sp = this->SplitVarToPopIdx(tree_[ni].split_var_);
					int ni2 = (sp == pop_idx) ? tree_[ni].ch1_idx_ : tree_[ni].ch0_idx_;

					if (ni2 < ni) {
						Util::ThrowError("BaseTree::GetSinglePopSplitStream : "
								             "splits are out of order\n");
					}
					ni = ni2;
				}

				if (tree_[ni].split_var_ == BaseTree::kNoSplit) {
					// it is a leaf node
					double x = 0, y = 0;
					this->NodeTargetCounts(ni, &x, &y);
					// a -1 indicates this is a leaf node
					splits->push_back(-1);
					splits->push_back(leaf_counts->size());
					splits->push_back(leaf_counts->size() + 1);

					leaf_counts->push_back(x);
					leaf_counts->push_back(y);
				} else {
					splits->push_back(tree_[ni].split_var_);
					splits->push_back(3 * tree_[ni].ch0_idx_);
					splits->push_back(3 * tree_[ni].ch1_idx_);
				}

			} else {
				splits->push_back(tree_[nd_idx].split_var_);
				splits->push_back(3 * tree_[nd_idx].ch0_idx_);
				splits->push_back(3 * tree_[nd_idx].ch1_idx_);
			}
		} else if (tree_[nd_idx].split_var_ == BaseTree::kNoSplit) {
			// it is a leaf node
			double x = 0, y = 0;
			this->NodeTargetCounts(nd_idx, &x, &y);
			// a -1 indicates this is a leaf node
			splits->push_back(-1);
			splits->push_back(leaf_counts->size());
			splits->push_back(leaf_counts->size() + 1);

			leaf_counts->push_back(x);
			leaf_counts->push_back(y);
		} else {
			// a -1 indicates this is a leaf node
			splits->push_back(tree_[nd_idx].split_var_);
			splits->push_back(3 * tree_[nd_idx].ch0_idx_);
			splits->push_back(3 * tree_[nd_idx].ch1_idx_);
		}
	}
}

void GreedyTreeBuilder::GrowTree() {
	for (int iter_idx = 0;; ++iter_idx) {
		bool made_split = false;
		for (int nd_idx = 0, sz = tree_.size(); nd_idx < sz; ++nd_idx) {
			int split_var = this->BestNodeSplit(nd_idx);

			if (split_var != BaseTree::kNoSplit) {
				this->SplitNode(nd_idx, split_var);
				made_split = true;
			}
		}
		if (!made_split) {
			break;
		}
	}
}


int GreedyTreeBuilder::BestNodeSplit(int node_idx) const {
	double best_score = this->GetNoSplitScore(node_idx);
	int best_split = BaseTree::kNoSplit;

	IntVec::const_iterator itr;
	for (itr = split_vars_.begin(); itr != split_vars_.end(); ++itr) {
		double cur_score = split_penalty_ + this->GetSplitScore(node_idx, *itr);

		if (cur_score != R_NegInf && cur_score > best_score) {
			best_split = *itr;
			best_score = cur_score;
		}
	}
	return(best_split);
}


int RandomTreeBuilder::SampleNodeSplit(int node_idx, RealVec * prb_buffer) const {
	int n_pts = tree_[node_idx].pt_set_.size();

	if (n_pts < min_leaf_sz_ || tree_[node_idx].depth_ >= max_split_depth_) {
		return(-1);
	}

	(*prb_buffer)[0] = this->GetNoSplitScore(node_idx);

	double mx = (*prb_buffer)[0];

	IntVec::const_iterator itr;
	int n_prbs = 1;
	for (itr = split_vars_.begin(); itr != split_vars_.end(); ++itr, ++n_prbs) {
		if (unif_rand() < 0.001) {
			(*prb_buffer)[n_prbs] = R_NegInf;
		} else {
		  (*prb_buffer)[n_prbs] = split_penalty_ + this->GetSplitScore(node_idx, *itr);
		}

		if ((*prb_buffer)[n_prbs] > mx) {
			mx = (*prb_buffer)[n_prbs];
		}
	}

	double sum = 0.0;
	for (int i = 0; i < n_prbs; ++i) {
		(*prb_buffer)[i] = exp((*prb_buffer)[i] - mx);
		sum += (*prb_buffer)[i];
	}

	double r1 = unif_rand();
	for (int i = 0; i < n_prbs; ++i) {
		r1 -= (*prb_buffer)[i] / sum;
		if (r1 < 0) {
			if (i == 0) {
				return(-1);
			} else {
				return(split_vars_[i-1]);
			}
		}
	}
	return(-1);
}

int MinimumParentIndex(int* split_list, int posn) {
	if (split_list[posn] < 0) {
	  return(-1);
	} else {
		int r1 = split_list[posn];
		int r2 = MinimumParentIndex(split_list, split_list[posn+1]);
		int r3 = MinimumParentIndex(split_list, split_list[posn+2]);

    if (r2 < r1 && r2 >= 0) {
      r1 = r2;
    }
    if (r3 < r1 && r3 >= 0) {
      r1 = r3;
    }
    return(r1);
	}
}

}


extern "C" {
SEXP GrowBinTrees_RI(SEXP bin_mat, SEXP p_pop_labels,
							    SEXP p_max_index_diffs, SEXP p_mdl_params_sxp,
							    SEXP p_split_penalty, SEXP p_pop_idx, SEXP p_verbose,
							    SEXP split_lists, SEXP split_counts) {
	using namespace BinTree;
	using namespace Util;
	using namespace BinaryMatrix;

#ifdef CONSOLE_STACK_TRACE
  Util::ConsolePrint("BinTree::GrowBinTrees : Entering\n");
#endif

  GetRNGstate();

	BinaryMatrix::BinaryMat bm(bin_mat);
	SxpList mdl_params(p_mdl_params_sxp);

	int sample_splits = mdl_params.Int("sampleSplits");
	int min_leaf_sz = mdl_params.Int("minNodeSize");
	double pcount = mdl_params.Dbl("pCount");
	double splitcrit_pcount = mdl_params.Dbl("splitCritPcount");
	double split_penalty = GetNumeric(p_split_penalty, 0.0, NULL);
	int max_tree_depth = mdl_params.Int("maxNumSplits");
	bool verbose = GetInt(p_verbose, 0, NULL) ? true : false;

	IntVec max_index_diffs;
	Util::SxpToIntVec(p_max_index_diffs, &max_index_diffs);

	int pop_idx = GetInt(p_pop_idx, -1, NULL);
	bool split_on_pop = mdl_params.Int("splitOnPops") ? true : false;
	bool pop_wtd_score = mdl_params.Int("popWtdScore") ? true : false;

	IntVec pop_labels;
	Util::SxpToIntVec(p_pop_labels, &pop_labels);

	if (pop_labels.size() != bm.ncol()) {
		Util::ThrowError("pop_labels.size() != ncol(bin_mat).\n");
	} else if (pcount <= 0.0 || splitcrit_pcount <= 0.0) {
		Util::ThrowError("min(pcount, splitcrit_pcount) <= 0.0.\n");
	}


	int n_pops = 1;
	for (IntVec::iterator itr = pop_labels.begin(); itr != pop_labels.end(); ++itr) {
		if (*itr >= n_pops) n_pops = 1 + *itr;
	}

	for (int var_i = 0, n_vars = bm.nrow(); var_i < n_vars; var_i++) {
		GreedyTreeBuilder * tree = NULL;
		if (sample_splits) {
			tree = new RandomTreeBuilder(bm, pop_labels, var_i);
		} else {
			tree = new GreedyTreeBuilder(bm, pop_labels, var_i);
		}
		tree->set_split_penalty(split_penalty);
		tree->set_max_split_depth(max_tree_depth);
		tree->set_min_leaf_sz(min_leaf_sz);
		tree->set_split_pcount(splitcrit_pcount);
		tree->set_pop_wtd_score(pop_wtd_score);

		for (int i = 1; i <= max_index_diffs[var_i]; ++i) {
			if (var_i - i >= 0) {
				tree->AddSplitVar(var_i - i);
			}
		}

		if (split_on_pop) {
			for (int idx = 0; idx < n_pops; ++idx) {
				tree->AddSplitVar(tree->PopIdxToSplitVar(idx));
			}
		}

		IntVec splits;
		RealVec counts;

		tree->GrowTree();
		if (pop_idx >= 0) {
		  tree->GetSinglePopSplitStream(&splits, &counts, pop_idx);
		} else {
		  tree->GetSplitStream(&splits, &counts);
		}

		delete tree;

		for (RealVec::iterator i = counts.begin(); i != counts.end(); ++i) {
			*i += pcount;
		}

		SaveIntVecToSxp(split_lists, var_i, splits);
		SaveDblVecToSxp(split_counts, var_i, counts);
	}

#ifdef CONSOLE_STACK_TRACE
	Util::ConsolePrint("BinTree::GrowBinTrees : Exiting\n");
#endif

  PutRNGstate();

	return R_NilValue;
}

SEXP IncLeafCounts_RI(SEXP bin_mat, SEXP wts_sxp, SEXP zero_counts_sxp,
		                   SEXP p_mdl_params_sxp, SEXP tree_sxp) {
	using namespace BinTree;
	using namespace Util;
	using namespace BinaryMatrix;

#ifdef CONSOLE_STACK_TRACE
  Util::ConsolePrint("BinTree::IncLeafCounts_RI : Entering\n");
#endif

	BinaryMatrix::BinaryMat bm(bin_mat);
	SxpList mdl_params(p_mdl_params_sxp);
	SxpList tree_obj(tree_sxp);

	double * wts = REAL(wts_sxp);
	bool zero_counts = GetInt(zero_counts_sxp, 0, NULL) ? true : false;
	double pcount = mdl_params.Dbl("pCount");

	if (pcount <= 0.0) {
    Util::ThrowError("pcount <= 0.0.\n");
  }

	SEXP split_lists = tree_obj.Sxp("splitLists");
	SEXP leaf_counts = tree_obj.Sxp("leafCounts");

	for (int var_i = 0, n_vars = bm.nrow(); var_i < n_vars; ++var_i) {
		SEXP split_list_sxp = VECTOR_ELT(split_lists, var_i);
		int split_list_len = LENGTH(split_list_sxp);
		int* split_list = INTEGER(split_list_sxp);

		SEXP leaf_counts_sxp = VECTOR_ELT(leaf_counts, var_i);
		int leaf_count_len = LENGTH(leaf_counts_sxp);
		double* cts = REAL(leaf_counts_sxp);

		if (zero_counts) {
			for (int i = 0; i < leaf_count_len; ++i) {
				cts[i] = pcount;
			}
		}

		for (int hap_i = 0, n_hap = bm.ncol(); hap_i < n_hap; ++hap_i) {
			int posn = 0;
			while (true) {
				if (split_list[posn] < 0) {
          if (bm.GetElt(var_i, hap_i)) {
          	cts[split_list[posn + 2]] += wts[hap_i];
          } else {
          	cts[split_list[posn + 1]] += wts[hap_i];
          }
        	break;
				} else {
          if (bm.GetElt(split_list[posn], hap_i)) {
          	posn = split_list[posn + 2];
          } else {
          	posn = split_list[posn + 1];
          }
				}
			}
		}
	}

#ifdef CONSOLE_STACK_TRACE
	Util::ConsolePrint("BinTree::IncLeafCounts_RI : Exiting\n");
#endif

	return R_NilValue;
}

SEXP PerMarkerLLiks_RI(SEXP bin_mat, SEXP wts_sxp, SEXP tree_sxp,
		                   SEXP lliks_sxp) {
	using namespace BinTree;
	using namespace Util;
	using namespace BinaryMatrix;

#ifdef CONSOLE_STACK_TRACE
  Util::ConsolePrint("BinTree::IncLeafCounts_RI : Entering\n");
#endif

	BinaryMatrix::BinaryMat bm(bin_mat);
	SxpList tree_obj(tree_sxp);

	double * wts = REAL(wts_sxp);

	SEXP split_lists = tree_obj.Sxp("splitLists");
	SEXP leaf_counts = tree_obj.Sxp("leafCounts");
	double* lliks = REAL(lliks_sxp);

	if (LENGTH(lliks_sxp) < bm.nrow()) {
		Util::ThrowError("PerMarkerLLiks_RI: lliks_sxp wrong length\n");
	}

	for (int var_i = 0, n_vars = bm.nrow(); var_i < n_vars; var_i++) {
		SEXP split_list_sxp = VECTOR_ELT(split_lists, var_i);
		int split_list_len = LENGTH(split_list_sxp);
		int* split_list = INTEGER(split_list_sxp);

		SEXP leaf_counts_sxp = VECTOR_ELT(leaf_counts, var_i);
		int leaf_count_len = LENGTH(leaf_counts_sxp);
		double* cts = REAL(leaf_counts_sxp);

		double llik = 0.0;
		for (int hap_i = 0, n_hap = bm.ncol(); hap_i < n_hap; hap_i++) {
			int posn = 0;
			while (true) {
				if (split_list[posn] < 0) {
					double ct0 = cts[split_list[posn + 1]];
					double ct1 = cts[split_list[posn + 2]];
					double pr1 = ct1 / (ct0 + ct1);
          if (bm.GetElt(var_i, hap_i)) {
          	llik += wts[hap_i] * log(pr1);
          } else {
          	llik += wts[hap_i] * log(1.0 - pr1);
          }
          break;
				} else {
          if (bm.GetElt(split_list[posn], hap_i)) {
          	posn = split_list[posn+2];
          } else {
          	posn = split_list[posn+1];
          }
				}
			}
		}
		lliks[var_i] = llik;
	}

#ifdef CONSOLE_STACK_TRACE
	Util::ConsolePrint("BinTree::IncLeafCounts_RI : Exiting\n");
#endif

	return R_NilValue;
}

SEXP FirstUnclippedTreeIndex_RI(SEXP trees_sxp, SEXP max_drop_sxp) {
	using namespace BinTree;
	using namespace Util;
	using namespace BinaryMatrix;
	using namespace Aradmx;

#ifdef CONSOLE_STACK_TRACE
  Rprintf("AbbnBuildDropLookups : Entering\n");
#endif

  // We return a matrix of dimensions [winSz, nSnp]
  // where [j, k] is the largest parent-set for which every parent
  // has index >= (k-j)

	SEXP tree_sxp0 = VECTOR_ELT(trees_sxp, 0);
	SxpList tree_obj0(tree_sxp0);

	int n_var0 = LENGTH(tree_obj0.Sxp("leafCounts"));
  int max_drop = GetInt(max_drop_sxp, 0, NULL);

  // The first argument is the number of populations -- just one at this point
  DropLookupIndexer indexer(1, n_var0, max_drop);

  SEXP r_sxp;
  int n_elt = n_var0 * max_drop;
  PROTECT(r_sxp = NEW_INTEGER(n_elt));
  int * ri = INTEGER(r_sxp);

  for (int i = 0; i < n_elt; ++i) {
  	ri[i] = -1;
  }

  int num_trees = LENGTH(trees_sxp);

  for (int tree_idx = 0; tree_idx < num_trees; ++tree_idx) {
  	SEXP tree_sxp = VECTOR_ELT(trees_sxp, tree_idx);
  	SxpList tree_obj(tree_sxp);

		SEXP split_lists_sxp = tree_obj.Sxp("splitLists");

  	int n_var = LENGTH(split_lists_sxp);
  	if (n_var != n_var0) {
  		Util::ThrowError("inconsistent number of implies snps\n");
  	}
	  for (int var_i = 0; var_i < n_var; ++var_i) {
			int* split_list = INTEGER(VECTOR_ELT(split_lists_sxp, var_i));
	  	int min_parent = MinimumParentIndex(split_list, 0);
	  	if (min_parent == -1) {
	  		// In the special case of no splits in the tree, we set this to
	  		// the current variable index, so the following checks will pass
	  		min_parent = var_i;
	  	}
	  	if (var_i - min_parent >= max_drop) {
        Util::ThrowError("var_i = %d, min_parent = %d, tree_idx = %d, max_drop = %d\n",
        		             var_i, min_parent, tree_idx, max_drop);
	  	}
	  	for (int cut_idx = 0; cut_idx < max_drop; ++cut_idx) {
	  		int idx = indexer.FlatIndex(0, var_i, cut_idx);
	  		if (ri[idx] < 0 && min_parent >= var_i - cut_idx) {
	  			ri[idx] = tree_idx;
	  		}
	  	}
	  }
  }
  UNPROTECT(1);

#ifdef CONSOLE_STACK_TRACE
  Rprintf("AbbnBuildDropLookups : Exiting\n");
#endif

  return(r_sxp);
}

SEXP PackIntsToUnsignedChars_RI(SEXP ints_sxp) {
	int* v = INTEGER(ints_sxp);
	int n = LENGTH(ints_sxp);
	int new_len = (n/4) + ((n % 4 == 0) ? 0 : 1);
	SEXP r_sxp;
	PROTECT(r_sxp = NEW_INTEGER(new_len));
	unsigned char* r_uchar = (unsigned char*)INTEGER(r_sxp);

	for (int i = 0; i < n; ++i) {
		r_uchar[i] = (unsigned char)v[i];
	}
	UNPROTECT(1);
	return(r_sxp);
}

}
