#include "binary_mat.hpp"

namespace BinaryMatrix {

void RbindMats(const CastableMatI& a, const CastableMatI& b,
               CastableMatI * c) {
  if (!c || c->nrow() != a.nrow() + b.nrow()) {
    error("RbindMats : incompatible number of rows\n");
  } else if (c->ncol() != a.ncol() || c->ncol() != b.ncol()) {
    error("RbindMats : incompatible number of columns\n");
  }

  for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
    int ri = 0;
    for (int ri_a = 0, a_nr = a.nrow(); ri_a < a_nr; ++ri_a, ++ri) {
      c->SetElt(ri, ci, a.GetElt(ri_a, ci));
    }
    for (int ri_b = 0, b_nr = b.nrow(); ri_b < b_nr; ++ri_b, ++ri) {
      c->SetElt(ri, ci, b.GetElt(ri_b, ci));
    }
  }
}

void CbindMats(const CastableMatI& a, const CastableMatI& b,
               CastableMatI * c) {
  if (!c || c->ncol() != a.ncol() + b.ncol()) {
    error("CbindMats : incompatible number of columns\n");
  } else if (c->nrow() != a.nrow() || c->nrow() != b.nrow()) {
    error("CbindMats : incompatible number of rows\n");
  }

  for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
    int ci = 0;
    for (int ci_a = 0, a_nc = a.ncol(); ci_a < a_nc; ++ci_a, ++ci) {
      c->SetElt(ri, ci, a.GetElt(ri, ci_a));
    }
    for (int ci_b = 0, b_nc = b.ncol(); ci_b < b_nc; ++ci_b, ++ci) {
      c->SetElt(ri, ci, b.GetElt(ri, ci_b));
    }
  }
}

bool CheckRowColIndices(const IntVec& row_inds, const IntVec& col_inds,
                        const CastableMatI& m) {
  for (int ci = 0, nc = col_inds.size(); ci < nc; ++ci) {
    if (col_inds[ci] < 0 || col_inds[ci] >= m.ncol()) {
      error("col_inds[%d] = %d.  out of range\n", ci, col_inds[ci]);
      return(false);
    }
  }

  for (int ri = 0, nr = row_inds.size(); ri < nr; ++ri) {
    if (row_inds[ri] < 0 || row_inds[ri] >= m.nrow()) {
      error("row_inds[%d] = %d.  out of range\n", ri, row_inds[ri]);
      return(false);
    }
  }
  return(true);
}

void ConvertMat(const CastableMatI& a, const IntVec& row_inds,
                const IntVec& col_inds, CastableMatI * b) {
  if (!b || b->nrow() != (int)row_inds.size() || b->ncol() != (int)col_inds.size()) {
    error("TransposeMats : incompatible number of rows or columns\n");
  }

  CheckRowColIndices(row_inds, col_inds, a);

  for (int ci = 0, nc = col_inds.size(); ci < nc; ++ci) {
    for (int ri = 0, nr = row_inds.size(); ri < nr; ++ri) {
      b->SetElt(ri, ci, a.GetElt(row_inds[ri], col_inds[ci]));
    }
  }
}

void ConvertMatDbl(const CastableMatI& a, const IntVec& row_inds,
                   const IntVec& col_inds, CastableMatI * b) {
  if (!b || b->nrow() != (int)row_inds.size() || b->ncol() != (int)col_inds.size()) {
    error("TransposeMats : incompatible number of rows or columns\n");
  }

  CheckRowColIndices(row_inds, col_inds, a);

  for (int ci = 0, nc = col_inds.size(); ci < nc; ++ci) {
    for (int ri = 0, nr = row_inds.size(); ri < nr; ++ri) {
      b->SetEltDbl(ri, ci, a.GetEltDbl(row_inds[ri], col_inds[ci]));
    }
  }
}

void SetSlice(const CastableMatI& x, const IntVec& row_inds,
              const IntVec& col_inds, CastableMatI * y) {
  if ((int)row_inds.size() != x.nrow() || (int)col_inds.size() != x.ncol()) {
    error("TransposeMats : incompatible number of rows or columns\n");
  }

  CheckRowColIndices(row_inds, col_inds, *y);

  for (int ci = 0, nc = col_inds.size(); ci < nc; ++ci) {
    for (int ri = 0, nr = row_inds.size(); ri < nr; ++ri) {
      y->SetElt(row_inds[ri], col_inds[ci], x.GetElt(ri, ci));
    }
  }
}

void TransposeMat(const CastableMatI& a, CastableMatI * b) {
  if (!b || b->ncol() != a.nrow() || b->nrow() != a.ncol()) {
    error("TransposeMats : incompatible number of rows or columns\n");
  }

  for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
    for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
      b->SetElt(ci, ri, a.GetElt(ri, ci));
    }
  }
}

int MatrixHash(const CastableMatI& a) {
  int h = 0;
  for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
    for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
      h = h * 33 + a.GetElt(ri, ci);
    }
  }
  return(h);
}

void RowSums(const CastableMatI& a, IntVec * v) {
  if (!v || (int)v->size() != a.nrow()) {
    error("RowSums : incompatible number of rows or columns\n");
  }

  for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
    int rs = 0;
    for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
      rs += a.GetElt(ri, ci);
    }
    (*v)[ri] = rs;
  }
}

void RowSumsDbl(const CastableMatI& a, Util::RealVec * v) {
  if (!v || (int)v->size() != a.nrow()) {
    error("RowSumsDbl : incompatible number of rows or columns\n");
  }

  for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
    double rs = 0;
    for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
      rs += a.GetEltDbl(ri, ci);
    }
    (*v)[ri] = rs;
  }
}

void ColSums(const CastableMatI& a, Util::IntVec * v) {
  if (!v || (int)v->size() != a.ncol()) {
    error("ColSums : incompatible number of rows or columns\n");
  }

  for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
  	int cs = 0;
    for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
      cs += a.GetElt(ri, ci);
    }
    (*v)[ci] = cs;
  }
}

void ColSumsDbl(const CastableMatI& a, Util::RealVec * v) {
  if (!v || (int)v->size() != a.ncol()) {
    error("ColSumsDbl : incompatible number of rows or columns\n");
  }

  for (int ci = 0, nc = a.ncol(); ci < nc; ++ci) {
  	double cs = 0;
    for (int ri = 0, nr = a.nrow(); ri < nr; ++ri) {
      cs += a.GetEltDbl(ri, ci);
    }
    (*v)[ci] = cs;
  }
}

void MatMult(const CastableMatI& a, const CastableMatI& b, CastableMatI * c) {
	int nr = a.nrow();
	int nc = a.ncol();

  if (!a.EqualDims(b) || !a.EqualDims(*c)) {
		error("MultMatMatMasked : a,b,c do not all have same dimensions\n");
	}
	for (int ri = 0; ri < nr; ++ri) {
  for (int ci = 0; ci < nc; ++ci) {
			c->SetEltDbl(ri, ci, a.GetEltDbl(ri, ci) * b.GetEltDbl(ri, ci));
	}}
}

void MatNegate(const CastableMatI& a, CastableMatI * b) {
	int nr = a.nrow();
	int nc = a.ncol();

  if (!a.EqualDims(*b)) {
		error("MultMatMatMasked : a and b do not have same dimensions\n");
	}
	for (int ri = 0; ri < nr; ++ri) {
  for (int ci = 0; ci < nc; ++ci) {
			b->SetElt(ri, ci, 1 - a.GetElt(ri, ci));
	}}
}

void MultMatMatMasked(const CastableMatI& a, const CastableMatI& b,
		                  const CastableMatI * msk,
		                  CastableMatI * r) {
	int nr = a.nrow();
	int nc = b.ncol();
  int m = a.ncol();

  if (a.ncol() != b.nrow()) {
		error("MultMatMatMasked : a and b dimensions incompatible\n");
	} else if (msk && !a.EqualDims(*msk)) {
		error("MultMatMatMasked : dim(a) != dim(msk)\n");
	} else if (a.nrow() != r->nrow() || b.ncol() != r->ncol()) {
		error("MultMatMatMasked : dim(r) incompatible\n");
	}

	for (int ri = 0; ri < nr; ++ri) {
		for (int ci = 0; ci < nc; ++ci) {
			double v = 0.0;
			if (msk) {
				for (int idx = 0; idx < m; ++idx) {
					if (msk->GetElt(ri, m)) {
						v += a.GetEltDbl(ri, m) * b.GetEltDbl(m, ci);
					}
				}
			} else {
				for (int idx = 0; idx < m; ++idx) {
				  v += a.GetEltDbl(ri, m) * b.GetEltDbl(m, ci);
				}
			}
			r->SetEltDbl(ri, ci, v);
		}
	}
}

CastableMatI * OwningCastableMatFactory(SEXP sxp_obj) {
  if (OnesMat::IsOnesMat(sxp_obj)) {
    return new OwningCastableMat<OnesMat>(new OnesMat(sxp_obj));
  } else if (DblMat::IsDblMat(sxp_obj)) {
    return new OwningCastableMat<DblMat>(new DblMat(sxp_obj));
  } else if (IntMat::IsIntMat(sxp_obj)) {
    return new OwningCastableMat<IntMat>(new IntMat(sxp_obj));
  } else if (ByteMat::IsByteMat(sxp_obj)) {
    return new OwningCastableMat<ByteMat>(new ByteMat(sxp_obj));
  } else if (BinaryMat::IsBinaryMat(sxp_obj)) {
    return new OwningCastableMat<BinaryMat>(new BinaryMat(sxp_obj));
  } else if (IsBinaryMatScaled(sxp_obj)) {
    return new OwningCastableMat<ScaledMat<BinaryMat> >(new ScaledMat<BinaryMat>(sxp_obj));
  } else if (IsByteMatScaled(sxp_obj)) {
    return new OwningCastableMat<ScaledMat<ByteMat> >(new ScaledMat<ByteMat>(sxp_obj));
  } else {
    error("OwningCastableMatFactory : sxp_obj is unknown type\n");
    return(NULL);
  }
}

}  // namespace


extern "C" {

SEXP MarginSums(SEXP mat_sxp, SEXP margin_idx_sxp, SEXP r_sxp) {
  namespace bm = BinaryMatrix;
  using namespace Util;

  int margin_idx = GetInt(margin_idx_sxp, 0, NULL);
  bm::CastableMatI * m1 = bm::OwningCastableMatFactory(mat_sxp);
  PointerDeleter<bm::CastableMatI> m1_deleter(m1);

  if (margin_idx < 1 || margin_idx > 2) {
  	error("MarginSums : invalid margin_idx\n");
  }

  SEXP r = R_NilValue;

  int ret_len = (margin_idx == 1) ? m1->nrow() : m1->ncol();
  if (r_sxp == R_NilValue || IS_NUMERIC(r_sxp)) {
	  RealVec v(ret_len, 0);
	  if (margin_idx == 1) bm::RowSumsDbl(*m1, &v);
	  if (margin_idx == 2) bm::ColSumsDbl(*m1, &v);

	  if (r_sxp == R_NilValue) r = NEW_NUMERIC(ret_len);

	  if (LENGTH(r) != ret_len) error("MarginSums : LENGTH(r_sxp) wrong\n");
	  for (int i = 0, sz = ret_len; i < sz; ++i) {
	    REAL(r)[i] = v[i];
	  }

  } else if (IS_INTEGER(r_sxp)) {
  	IntVec v(ret_len, 0);
	  if (margin_idx == 1) bm::RowSums(*m1, &v);
	  if (margin_idx == 2) bm::ColSums(*m1, &v);

	  if (LENGTH(r) != ret_len) error("MarginSums : LENGTH(r_sxp) wrong\n");
	  for (int i = 0, sz = ret_len; i < sz; ++i) {
	    INTEGER(r)[i] = v[i];
	  }
  } else {
  	error("MarginSums : invalid return type\n");
  }
  return(r);
}

SEXP MatrixHash(SEXP mat_sxp) {
  namespace bm = BinaryMatrix;
  using namespace Util;

  bm::CastableMatI * m1 = bm::OwningCastableMatFactory(mat_sxp);
  PointerDeleter<bm::CastableMatI> m1_deleter(m1);

  int h = bm::MatrixHash(*m1);

  SEXP r = NEW_INTEGER(1);
  INTEGER(r)[0] = h;
  return(r);
}

SEXP ConvertMatrix(SEXP mat1_sxp, SEXP row_inds_sxp, SEXP col_inds_sxp,
                   SEXP mat2_sxp) {
  using namespace BinaryMatrix;
  using namespace Util;

  IntVec ri, ci;
  SxpToIntVec(row_inds_sxp, &ri);
  SxpToIntVec(col_inds_sxp, &ci);

  CastableMatI * m1 = OwningCastableMatFactory(mat1_sxp);
  CastableMatI * m2 = OwningCastableMatFactory(mat2_sxp);
  PointerDeleter<CastableMatI> m1_deleter(m1);
  PointerDeleter<CastableMatI> m2_deleter(m2);

  if (m1 == NULL || m2 == NULL) {
    error("Failed to load matrices\n");
  }
  ConvertMat(*m1, ri, ci, m2);
  return(mat2_sxp);
}

SEXP ConvertMatrixDbl(SEXP mat1_sxp, SEXP row_inds_sxp, SEXP col_inds_sxp,
                      SEXP mat2_sxp) {
  using namespace BinaryMatrix;
  using namespace Util;

  IntVec ri, ci;
  SxpToIntVec(row_inds_sxp, &ri);
  SxpToIntVec(col_inds_sxp, &ci);

  CastableMatI * m1 = OwningCastableMatFactory(mat1_sxp);
  CastableMatI * m2 = OwningCastableMatFactory(mat2_sxp);
  PointerDeleter<CastableMatI> m1_deleter(m1);
  PointerDeleter<CastableMatI> m2_deleter(m2);

  if (m1 == NULL || m2 == NULL) {
    error("Failed to load matrices\n");
  }
  ConvertMatDbl(*m1, ri, ci, m2);
  return(mat2_sxp);
}

SEXP SetMatrixSlice(SEXP mat1_sxp, SEXP row_inds_sxp, SEXP col_inds_sxp,
                    SEXP mat2_sxp) {
  using namespace BinaryMatrix;
  using namespace Util;

  IntVec ri, ci;
  SxpToIntVec(row_inds_sxp, &ri);
  SxpToIntVec(col_inds_sxp, &ci);

  CastableMatI * m1 = OwningCastableMatFactory(mat1_sxp);
  CastableMatI * m2 = OwningCastableMatFactory(mat2_sxp);
  PointerDeleter<CastableMatI> m1_deleter(m1);
  PointerDeleter<CastableMatI> m2_deleter(m2);

  if (m1 == NULL || m2 == NULL) {
    error("Failed to load matrices\n");
  }
  SetSlice(*m1, ri, ci, m2);
  return(mat2_sxp);
}

SEXP TransposeMatrix(SEXP mat1_sxp, SEXP mat2_sxp) {
  using namespace BinaryMatrix;
  using namespace Util;

  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(mat1_sxp));
  PointerDeleter<CastableMatI> m2(OwningCastableMatFactory(mat2_sxp));

  if (m1.is_null() || m2.is_null()) {
    error("Failed to load matrices\n");
  }

  TransposeMat(*m1.get(), m2.get());
  return(mat2_sxp);
}

SEXP RbindMatrix(SEXP mat1_sxp, SEXP mat2_sxp, SEXP mat3_sxp) {
  using namespace BinaryMatrix;
  using namespace Util;

  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(mat1_sxp));
  PointerDeleter<CastableMatI> m2(OwningCastableMatFactory(mat2_sxp));
  PointerDeleter<CastableMatI> m3(OwningCastableMatFactory(mat3_sxp));

  if (m1.is_null() || m2.is_null() || m3.is_null()) {
    error("Failed to load matrices\n");
  }

  RbindMats(*m1.get(), *m2.get(), m3.get());
  return(mat3_sxp);
}

SEXP CbindMatrix(SEXP mat1_sxp, SEXP mat2_sxp, SEXP r_sxp) {
  using namespace BinaryMatrix;
  using namespace Util;

  CastableMatI * m1 = OwningCastableMatFactory(mat1_sxp);
  CastableMatI * m2 = OwningCastableMatFactory(mat2_sxp);
  CastableMatI * r = OwningCastableMatFactory(r_sxp);
  PointerDeleter<CastableMatI> m1_deleter(m1);
  PointerDeleter<CastableMatI> m2_deleter(m2);
  PointerDeleter<CastableMatI> r_deleter(r);

  if (m1 == NULL || m2 == NULL || r == NULL) {
    error("Failed to load matrices\n");
  }
  CbindMats(*m1, *m2, r);
  return(r_sxp);
}

SEXP MultMatMatMasked(SEXP mat1_sxp, SEXP mat2_sxp, SEXP msk_sxp,
		                  SEXP r_sxp) {
	// Let A, B and C be the matrices stored in mat1_sxp, mat2_sxp, and msk_sxp
	// respectively.  We compute AB and store it in r_sxp.  If msk_sxp is not
	// all one's, the result [AB]_{ij} is computed as
	//        \sum_k A_{ik}B_{kj}C_{kj}
	// so the elements in B are set to zero by the masking matrix C
  namespace bm = BinaryMatrix;
  using namespace Util;

  typedef bm::CastableMatI Mat;

  Mat * m1 = bm::OwningCastableMatFactory(mat1_sxp);
  Mat * m2 = bm::OwningCastableMatFactory(mat2_sxp);
  Mat * msk = bm::OwningCastableMatFactory(msk_sxp);
  Mat * r = bm::OwningCastableMatFactory(r_sxp);
  bm::PointerDeleter<Mat> m1_deleter(m1);
  bm::PointerDeleter<Mat> m2_deleter(m2);
  bm::PointerDeleter<Mat> msk_deleter(msk);
  bm::PointerDeleter<Mat> r_deleter(r);

  if (m1 == NULL || m2 == NULL || r == NULL) {
    error("Failed to load matrices\n");
  }
  bm::MultMatMatMasked(*m1, *m2, msk, r);

  return(r_sxp);
}

SEXP FillMatrix(SEXP val_sxp, SEXP mat_sxp) {
  using namespace BinaryMatrix;
  using namespace Util;

  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(mat_sxp));

  if (m1.get() == NULL) {
    error("Failed to load matrices\n");
  }
  m1.get()->Fill(GetInt(val_sxp, 0, NULL) != 0);
  return(mat_sxp);
}

SEXP MatrixCompMult(SEXP m1_sxp, SEXP m2_sxp, SEXP m3_sxp) {
  using namespace BinaryMatrix;
  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(m1_sxp));
  PointerDeleter<CastableMatI> m2(OwningCastableMatFactory(m2_sxp));
  PointerDeleter<CastableMatI> m3(OwningCastableMatFactory(m3_sxp));
  MatMult(*m1.get(), *m2.get(), m3.get());
  return(m3_sxp);
}

SEXP MatrixNegate(SEXP m1_sxp, SEXP m2_sxp) {
  using namespace BinaryMatrix;
  PointerDeleter<CastableMatI> m1(OwningCastableMatFactory(m1_sxp));
  PointerDeleter<CastableMatI> m2(OwningCastableMatFactory(m2_sxp));
  MatNegate(*m1.get(), m2.get());
  return(m2_sxp);
}

}  // extern "C"
