#include "phase_em.hpp"
#include "phase_em_hap_opt.hpp"

namespace PhaseEM {

using namespace Util;

double CrossoverHapOpt::LikRatio(int snp_idx) {
	const RealVec& fwd_m0 = inf_obj_[0]->ForwardMsg(snp_idx - 1);
	const RealVec& fwd_m1 = inf_obj_[1]->ForwardMsg(snp_idx - 1);

	double lik0 = this->Lik(bkwd_msgs_[0], fwd_m0) + this->Lik(bkwd_msgs_[1], fwd_m1);
	double lik1 = this->Lik(bkwd_msgs_[1], fwd_m0) + this->Lik(bkwd_msgs_[0], fwd_m1);
  return(lik1 - lik0);
}

bool CrossoverHapOpt::MakeMove(int * snp_idx) {
	// assumes backward messages are stored in 'bkwd_bufs_[0][0]'
	// and 'bkwd_bufs_[1][0]'.  These are swapped if the likelihood
	// ratio is larger than 'move_thresh_'

	int si = *snp_idx;

	if (si == 0) {
		move_liks_[si] = 0.0;
		made_cross_[si] = false;

	} else {
		move_liks_[si] = this->LikRatio(si);
		if (move_liks_[si] > move_thresh_) {
			// make the crossover move
			made_cross_[si] = true;
			bkwd_msgs_[0].swap(bkwd_msgs_[1]);
		} else {
			made_cross_[si] = false;
		}
	}
	return(made_cross_[si]);
}

bool BacktrackXOverHapOpt::MakeMove(int * snp_idx) {
	int si = *snp_idx;
	double cur_lr = this->LikRatio(si);
	if (verbose_) Rprintf("lik_ratio at snp %d = %g\n", si, cur_lr);

	move_liks_[si] = cur_lr;
	made_cross_[si] = false;

	if (run_tracker_.RunStarted(si, cur_lr, move_thresh_)) {
    this->MarkRun(si, cur_lr);
    if (verbose_) Rprintf("starting (lr = %g, thresh = %g)\n", cur_lr, move_thresh_);

	} else if (run_tracker_.RunEnded(si, cur_lr)) {
		if (verbose_) Rprintf("run_ending\n");
		run_tracker_.EndRun(snp_idx);
	  this->LoadBestMsgs();
		this->CrossoverHapOpt::MakeMove(snp_idx);
		return(true);
	}
	return(false);
}


CrossoverHapOpt * CrossoverHapOptFactory(PhaseEmModel& p,
		                                     bool back_tracking) {
	if (back_tracking) {
		return(new BacktrackXOverHapOpt(p));
	} else {
		return(new CrossoverHapOpt(p));
	}
}

}  // namespace PhaseEM {

extern "C" {

SEXP PhaseEmCrossoverScores_RI(SEXP params_sxp, SEXP cases_sxp,
					                     SEXP case_idx1_sxp, SEXP case_idx2_sxp) {
	using namespace PhaseEM;
	using namespace Util;

	PointerDeleter<PhaseEmModel> mdl(PhaseEmModelFactory(params_sxp, cases_sxp));
	CrossoverHapOpt opt_obj(*mdl.get());
  opt_obj.set_move_thresh(R_PosInf);

	int case_idx1 = GetInt(case_idx1_sxp, -1, NULL);
	int case_idx2 = GetInt(case_idx2_sxp, -1, NULL);

	opt_obj.set_case_inds(case_idx1, case_idx2);
	opt_obj.ForwardAlgo();
	opt_obj.BackwardAlgo();

	const RealVec& cross_liks = opt_obj.move_liks();
	SEXP ret_sxp = NEW_NUMERIC(cross_liks.size());
	double * crs_lks = REAL(ret_sxp);
	for (int i = 0, sz = mdl.get()->n_snp(); i < sz; ++i) {
		crs_lks[i] = cross_liks[i];
	}

	return(ret_sxp);
}

SEXP PhaseEmCrossoverOpt_RI(SEXP params_sxp, SEXP cases_sxp,
					                  SEXP case_inds_sxp, SEXP move_thresh_sxp,
					                  SEXP back_tracking_sxp) {
	using namespace PhaseEM;
	using namespace Util;

  bool back_tracking = (GetInt(back_tracking_sxp, 0, NULL) != 0);
  double move_thresh = GetNumeric(move_thresh_sxp, 0.0, NULL);

	PointerDeleter<PhaseEmModel> mdl(PhaseEmModelFactory(params_sxp, cases_sxp));

	CrossoverHapOpt * opt_obj = CrossoverHapOptFactory(*mdl.get(), back_tracking);
	PointerDeleter<CrossoverHapOpt> opt_obj_deleter(opt_obj);
  opt_obj->set_move_thresh(move_thresh);

	IntVec case_inds;
	SxpToIntVec(case_inds_sxp, &case_inds);
	if (case_inds.size() % 2 != 0) {
		error("case_inds_sxp must be a matrix with 2 rows\n");
	}

	for(int i = 0, ns = case_inds.size(); i < ns; i += 2) {
		opt_obj->set_case_inds(case_inds[i], case_inds[i + 1]);
		opt_obj->ForwardAlgo();
		opt_obj->BackwardAlgo();
		opt_obj->ApplyMoves();

		R_CheckUserInterrupt();
	}

	return(R_NilValue);
}

}  // extern "C"
