#include "phase_em.hpp"

namespace PhaseEM {

double PhaseEmInference::ForwardAlgo() {
	if (case_index_ < 0 || case_index_ >= params_.n_cases()) {
		error("PhaseEmInference::ForwardAlgo : case_index_ is invalid (%d)\n", case_index_);
	}

	double cur_plik = 0.0;
	double init_prb = 1.0 / (double)(params_.n_cases() - 2);
	//This is the forward pass
	RealVec tmp_msg(params_.n_cases(), init_prb);
	RealVec * last_msg = &init_fwd_msg_;

	for (int snp_idx = 0, ns = params_.n_snp(); snp_idx < ns; ++snp_idx) {
		RealVec * this_msg = (this->StoreMessage(snp_idx)
				                  ? this->FwdMsgBuffer(snp_idx)
		                      : &tmp_msg);
		this->ForwardStep(snp_idx, *last_msg, this_msg);
		if (snp_idx % rescale_freq_ == 0) {
			double x = VecSum(*this_msg);
			cur_plik += log(x);
			VecScale(1.0 / x, this_msg);
		}
		last_msg = this_msg;
	}

	double x = VecSum(*last_msg);
	cur_plik += log(x);
	VecScale(1.0 / x, last_msg);

	return(cur_plik);
}

void PhaseEmInference::BackwardAlgo() {
	double init_prb = 1.0 / (double)(params_.n_cases() - 2);
	//This is the forward pass
	RealVec bkwd_msg(params_.n_cases(), init_prb);

	for (int snp_idx = params_.n_snp() - 1; snp_idx >= 0; --snp_idx) {
		this->BackwardStep(snp_idx, bkwd_msg, &bkwd_msg);
		this->BackwardProcess(params_.GetAllele(case_index_, snp_idx), snp_idx, bkwd_msg);

		if (snp_idx % rescale_freq_ == 0) {
			double x = VecSum(bkwd_msg);
			VecScale(1.0 / x, &bkwd_msg);
		}
	}
}

void PhaseEmInference::FillForwardMsgs(int start_idx, int last_idx) {
	if (start_idx <= 0) {
		error("PhaseEmInference::FillForwardMsgs : invalid start index\n");
	}

	for (int snp_idx = start_idx + 1; snp_idx <= last_idx; ++snp_idx) {
		this->ForwardStep(snp_idx, *this->FwdMsgBuffer(snp_idx - 1), this->FwdMsgBuffer(snp_idx));
	}
}

int PhaseEmInference::FindForwardMsgs(int start_idx) {
	for (int snp_idx = start_idx; snp_idx >= 0; --snp_idx) {
		if (this->FwdMsgExists(snp_idx)) {
			return(snp_idx);
		}
	}
	error("FindForwardMsgs : failed to find saved message from index 1,..., %d\n", start_idx);
	return(-1);
}

void PhaseEmInference::ClearFwdMsgs() {
	vector<RealVec *>::iterator i = fwd_msgs_.begin();
	for (; i != fwd_msgs_.end(); ++i) {
		if (*i) {
			delete(*i);
			*i = NULL;
		}
	}
}

RealVec * PhaseEmInference::FwdMsgBuffer(int snp_idx) {
	if (snp_idx == -1) {
		return(&init_fwd_msg_);
	} else if (!this->FwdMsgExists(snp_idx)) {
		fwd_msgs_[snp_idx] = new RealVec(params_.n_cases());
	}
	return(fwd_msgs_[snp_idx]);
}

bool PhaseEmInference::StoreMessage(int snp_idx) {
	if (checkpoint_freq_ <= 1 || snp_idx == 0 || snp_idx + 1 == params_.n_snp() - 1) {
		return(true);
	} else {
		return((snp_idx % checkpoint_freq_) == 0);
	}
}

PhaseEmParamOpt::PhaseEmParamOpt(PhaseEmModel& p) : PhaseEmInference(p) {
	mut_counts_ = RealVec(params_.n_snp() * 2, 0.0);
	jump_counts_ = RealVec(params_.n_snp() * 2, 0.0);
	count_buf_ = RealVec(params_.n_cases());
	tmp_msg_ = RealVec(params_.n_cases());
}

void PhaseEmParamOpt::UpdateMutCounts(int cur_allele, int snp_idx,
	                                    const RealVec& bkwd_msg) {
	// allele-flipping ("mutation") counts
	const RealVec& fwd_msg = this->ForwardMsg(snp_idx - 1);
	RealVec::const_iterator m1 = bkwd_msg.begin();
	RealVec::const_iterator m2 = fwd_msg.begin();

	double ct_mlt = 1.0 / DotProd(bkwd_msg, fwd_msg);
	int delt = 2 * snp_idx;
	for (int ci = 0; m1 != bkwd_msg.end(); ++ci, ++m1, ++m2) {
		double ct = (*m1) * (*m2) * ct_mlt;
		int mut = (cur_allele != params_.GetAllele(ci, snp_idx)) ? 1 : 0;
		mut_counts_[delt + mut] += ct;
	}
}

void PhaseEmParamOpt::UpdateJumpCounts(int cur_allele, int snp_idx,
		                                   const RealVec& bkwd_msg) {
	if (snp_idx > 0) {
		const RealVec& fwd_msg = this->ForwardMsg(snp_idx - 2);
	  int prev_allele = params_.GetAllele(case_index_, snp_idx - 1);
		params_.InitPttl(fwd_msg, snp_idx - 1, case_index_, prev_allele, &tmp_msg_);

		double jp = params_.jump_prob(snp_idx - 1);
		double jump_ct = VecSum(bkwd_msg) * VecSum(tmp_msg_) * jp / (double)(params_.n_cases() - 2);
		double no_jump_ct = (1.0 - jp) * DotProd(tmp_msg_, bkwd_msg);
		double tot_ct = jump_ct + no_jump_ct;

		int delt = 2 * (snp_idx - 1);
	  jump_counts_[delt] += no_jump_ct / tot_ct;
	  jump_counts_[delt + 1] += jump_ct / tot_ct;
	}
}

PhaseEmModel * PhaseEmModelFactory(SEXP params_sxp, SEXP cases_sxp) {
	PhaseEmModel * mdl = NULL;
	if (BinaryMat::IsBinaryMat(cases_sxp)) {
		mdl = new PhaseEmModelBinMat();
	} else if (ByteMat::IsByteMat(cases_sxp)) {
		mdl = new PhaseEmModelByteMat();
	} else {
		mdl = new PhaseEmModelIntMat();
	}
	mdl->LoadParams(params_sxp);
	mdl->LoadCases(cases_sxp);
	return(mdl);
}

}  // namespace PhaseEM {



extern "C" {

SEXP PhaseEmExpCounts_RI(SEXP params_sxp, SEXP cases_sxp, SEXP case_inds_sxp,
				                 SEXP jump_counts_sxp, SEXP mut_counts_sxp,
					               SEXP pliks_sxp) {
	using namespace PhaseEM;
	using namespace Util;

	PhaseEmModel * mdl = PhaseEmModelFactory(params_sxp, cases_sxp);
	PointerDeleter<PhaseEmModel> mdl_deleter(mdl);
	PhaseEmParamOpt inf_obj(*mdl);

	IntVec case_inds;
	SxpToIntVec(case_inds_sxp, &case_inds);

	double * pliks = (pliks_sxp == R_NilValue) ? NULL : REAL(pliks_sxp);
	if (pliks && LENGTH(pliks_sxp) != (int)case_inds.size()) {
		error("PhaseEMParamUpd : input buffers wrong length\n");
	}

	for(int case_idx = 0, ns = case_inds.size(); case_idx < ns; ++case_idx) {
		inf_obj.set_case_index(case_inds[case_idx]);
		double cur_plik = inf_obj.ForwardAlgo();
		inf_obj.BackwardAlgo();
		if (pliks) pliks[case_idx] = cur_plik;

		R_CheckUserInterrupt();
	}

	// Save the counts

	const RealVec& mut_counts = inf_obj.mut_counts();
	const RealVec& jump_counts = inf_obj.jump_counts();

	if (LENGTH(jump_counts_sxp) != (int)jump_counts.size()
			|| LENGTH(mut_counts_sxp) != (int)mut_counts.size()
			|| (int)jump_counts.size() != 2 * mdl->n_snp()) {
		error("PhaseEMParamUpd : input buffers wrong length\n");
	}

	double * jump_counts_ret = REAL(jump_counts_sxp);
	double * mut_counts_ret = REAL(mut_counts_sxp);

	for (int i = 0, sz = 2 * mdl->n_snp(); i < sz; ++i) {
		jump_counts_ret[i] = jump_counts[i];
		mut_counts_ret[i] = mut_counts[i];
	}

	return R_NilValue;
}

}  // extern "C"
