#ifndef PHASEEM_C_SRC_PHASEEM_HPP__
#define PHASEEM_C_SRC_PHASEEM_HPP__ 1

#include <R.h>
#include <Rdefines.h>
#include <Rmath.h>
#include <R_ext/Print.h>

#include "util.hpp"
#include "binary_mat.hpp"

namespace PhaseEM {

using namespace Util;
using namespace BinaryMatrix;

class PhaseEmModel;
PhaseEmModel * PhaseEmModelFactory(SEXP params_sxp, SEXP cases_sxp);

class PhaseEmModel {
public:
	PhaseEmModel() {
		n_snp_ = -1;
		rescale_freq_ = 3;
	}

	int n_snp() const { return(n_snp_); }
	virtual int n_cases() const = 0;
	double jump_prob(int snp_idx) const { return(jump_probs_[snp_idx]); }

	void set_rescale_freq(int x) { rescale_freq_ = x; }

	virtual int GetAllele(int case_idx, int snp_idx) const = 0;
	virtual void SetAllele(int case_idx, int snp_idx, int val) = 0;

	virtual void InitPttl(const RealVec& inp_pttl, int snp_idx, int case_idx,
				                  int cur_allele, RealVec * pttl) = 0;
	virtual void InitPttl(int snp_idx, int case_idx, int cur_allele, RealVec * pttl) = 0;

	virtual void LoadParams(SEXP params_sxp) {
		int n_snp_idx = ListEltIdxByName(params_sxp, "n_snp");
		n_snp_ = GetInt(VECTOR_ELT(params_sxp, n_snp_idx), -1, NULL);
		SxpVecEltToDblVec(params_sxp, ListEltIdxByName(params_sxp, "mut_probs"), &mut_probs_);
		SxpVecEltToDblVec(params_sxp, ListEltIdxByName(params_sxp, "jump_probs"), &jump_probs_);

		if ((int)jump_probs_.size() != n_snp_ || (int)mut_probs_.size() != n_snp_) {
			error("LoadParams : mut_probs or jump_probs is wrong length\n");
		}
	}

	void MakeMessage(const RealVec& v, double jump_mlt, double no_jump_mlt,
			             RealVec * m) const {
		double sm = 0.0;
		RealVec::const_iterator i;
		RealVec::iterator j;
		for (i = v.begin(); i != v.end(); ++i) {
			sm += *i;
		}
		double add_amt = sm * jump_mlt;
		for (i = v.begin(), j = m->begin(); i != v.end(); ++i, ++j) {
			*j = (*i) * no_jump_mlt + add_amt;
		}
	}

	virtual void LoadCases(SEXP cases_sxp) = 0;
	virtual ~PhaseEmModel() { }
private:
	void LoadIntVec(SEXP sxp_obj, const char * nm, IntVec * r) {
	  SxpVecEltToIntVec(sxp_obj, ListEltIdxByName(sxp_obj, nm), r);
	}

public:
	// jump_probs_[0] is the probability of a jump between markers with
	// indices 0 and 1.  jump_probs[n_snp_ - 1] should be 0.0 or 1.0.  Any
	// other value would work, but since there is no n_snp_'th marker, there
	// is no meaningful value either.  However, jump_probs_.size() must be
	// at least n_snp_
	int n_snp_;
	RealVec mut_probs_;
	RealVec jump_probs_;
	int rescale_freq_;
};

class PhaseEmModelBinMat : public PhaseEmModel {
public:
	PhaseEmModelBinMat() : PhaseEmModel(), n_cases_(0), cases_(NULL) { }

	virtual int GetAllele(int case_idx, int snp_idx) const {
		return(cases_->GetElt(case_idx, snp_idx));
	}

	virtual void SetAllele(int case_idx, int snp_idx, int val) {
		cases_->SetElt(case_idx, snp_idx, val);
	}

	virtual int n_cases() const {
	  return(cases_->nrow());
	}

	virtual void InitPttl(const RealVec& inp_pttl, int snp_idx, int case_idx,
			                  int cur_allele, RealVec * pttl) {
		double mut_pr = mut_probs_[snp_idx];
		double no_mut_pr = 1.0 - mut_pr;

    RealVec::iterator r = pttl->begin();
    RealVec::const_iterator i = inp_pttl.begin();
    BinaryMat::RowIterator v(*cases_, snp_idx);

		for (; r != pttl->end(); ++r, ++i, v.Inc()) {
			*r = (*i) * ((cur_allele == v.Val()) ? no_mut_pr : mut_pr);
		}

		int case_i2 = (case_idx % 2) ? (case_idx - 1) : (case_idx + 1);
		(*pttl)[case_idx] = (*pttl)[case_i2] = 0.0;
	}

	virtual void InitPttl(int snp_idx, int case_idx, int cur_allele, RealVec * pttl) {
		double mut_pr = mut_probs_[snp_idx];
		double no_mut_pr = 1.0 - mut_pr;

    RealVec::iterator r = pttl->begin();
    BinaryMat::RowIterator v(*cases_, snp_idx);
		for (; r != pttl->end(); ++r, v.Inc()) {
			*r *= ((cur_allele == v.Val()) ? no_mut_pr : mut_pr);
		}

		int case_i2 = (case_idx % 2) ? (case_idx - 1) : (case_idx + 1);
		(*pttl)[case_idx] = (*pttl)[case_i2] = 0.0;
	}

	virtual void LoadCases(SEXP cases_sxp) {
		if (cases_) delete cases_;
		cases_ = new BinaryMat(cases_sxp);
		n_cases_ = cases_->nrow();
	}

	virtual ~PhaseEmModelBinMat() {
		if (cases_) delete cases_;
		n_cases_ = 0;
	}

public:
	int n_cases_;
	BinaryMat * cases_;
};

class PhaseEmModelByteMat : public PhaseEmModel {
public:
	PhaseEmModelByteMat() : PhaseEmModel(), n_cases_(0), cases_(NULL) { }

	virtual int GetAllele(int case_idx, int snp_idx) const {
		return(cases_->GetElt(case_idx, snp_idx));
	}

	virtual void SetAllele(int case_idx, int snp_idx, int val) {
		cases_->SetElt(case_idx, snp_idx, val);
	}

	virtual int n_cases() const {
	  return(cases_->nrow());
	}

	virtual void InitPttl(const RealVec& inp_pttl, int snp_idx, int case_idx,
			                  int cur_allele, RealVec * pttl) {
		double mut_pr = mut_probs_[snp_idx];
		double no_mut_pr = 1.0 - mut_pr;

    RealVec::iterator r = pttl->begin();
    RealVec::const_iterator i = inp_pttl.begin();
    unsigned char * v = cases_->dat_ + (cases_->bytes_per_col_ * snp_idx);
    unsigned char a0 = (unsigned char)cur_allele;

		for (; r != pttl->end(); ++r, ++i, ++v) {
			*r = (*i) * ((a0 == *v) ? no_mut_pr : mut_pr);
		}

		int case_i2 = (case_idx % 2) ? (case_idx - 1) : (case_idx + 1);
		(*pttl)[case_idx] = (*pttl)[case_i2] = 0.0;
	}

	virtual void InitPttl(int snp_idx, int case_idx, int cur_allele, RealVec * pttl) {
		double mut_pr = mut_probs_[snp_idx];
		double no_mut_pr = 1.0 - mut_pr;

    RealVec::iterator r = pttl->begin();
    unsigned char * v = cases_->dat_ + (cases_->bytes_per_col_ * snp_idx);
    unsigned char a0 = (unsigned char)cur_allele;

		for (; r != pttl->end(); ++r, ++v) {
			*r *= ((a0 == *v) ? no_mut_pr : mut_pr);
		}

		int case_i2 = (case_idx % 2) ? (case_idx - 1) : (case_idx + 1);
		(*pttl)[case_idx] = (*pttl)[case_i2] = 0.0;
	}

	virtual void LoadCases(SEXP cases_sxp) {
		if (cases_) delete cases_;
		cases_ = new ByteMat(cases_sxp);
		n_cases_ = cases_->nrow();
	}

	virtual ~PhaseEmModelByteMat() {
		if (cases_) delete cases_;
		n_cases_ = 0;
	}

public:
	int n_cases_;
	ByteMat * cases_;
};

class PhaseEmModelIntMat : public PhaseEmModel {
public:
	PhaseEmModelIntMat() : PhaseEmModel(), cases_(NULL), n_cases_(-1) { }

	virtual int GetAllele(int case_idx, int snp_idx) const {
		//if (snp_idx < 0 || snp_idx >= this->n_snp()) error("invalid snp_idx (%d)\n", snp_idx);
		//if (case_idx < 0 || case_idx >= n_cases_) error("invalid snp_idx\n");
    return(cases_[snp_idx * n_cases_ + case_idx]);
	}

	virtual void SetAllele(int case_idx, int snp_idx, int val) {
		cases_[snp_idx * n_cases_ + case_idx] = val;
	}

	virtual int n_cases() const { return(n_cases_); }

	virtual void InitPttl(const RealVec& inp_pttl, int snp_idx, int case_idx,
			                  int cur_allele, RealVec * pttl) {
		double mut_pr = mut_probs_[snp_idx];
		double no_mut_pr = 1.0 - mut_pr;
		int * case_col = cases_ + (snp_idx * n_cases_);

    RealVec::iterator r = pttl->begin();
    RealVec::const_iterator i = inp_pttl.begin();

		for (; r != pttl->end(); ++r, ++i, ++case_col) {
			*r = (*i) * ((cur_allele == *case_col) ? no_mut_pr : mut_pr);
		}

		int case_i2 = (case_idx % 2) ? (case_idx - 1) : (case_idx + 1);
		(*pttl)[case_idx] = (*pttl)[case_i2] = 0.0;
	}

	virtual void InitPttl(int snp_idx, int case_idx, int cur_allele, RealVec * pttl) {
		double mut_pr = mut_probs_[snp_idx];
		double no_mut_pr = 1.0 - mut_pr;

		int * case_col = cases_ + (snp_idx * n_cases_);
    RealVec::iterator r = pttl->begin();

		for (; r != pttl->end(); ++r, ++case_col) {
			*r *= ((cur_allele == *case_col) ? no_mut_pr : mut_pr);
		}

		int case_i2 = (case_idx % 2) ? (case_idx - 1) : (case_idx + 1);
		(*pttl)[case_idx] = (*pttl)[case_i2] = 0.0;
	}

	virtual void LoadCases(SEXP cases_sxp) {
		int n_snp = -1;
		GetMatrixDim(cases_sxp, &n_cases_, &n_snp);
		cases_ = INTEGER(cases_sxp);

		if (n_snp != n_snp_) error("ncol(cases_sxp) is wrong\n");
	}

	virtual ~PhaseEmModelIntMat() {
		cases_ = NULL;
	}

public:
	int * cases_;
	int n_cases_;
};

class PhaseEmInference {
public:
	PhaseEmInference(PhaseEmModel& p)
	  : params_(p), checkpoint_freq_(1), rescale_freq_(3), case_index_(-1),
		  fwd_msgs_(params_.n_snp(), (RealVec *)NULL),
		  init_fwd_msg_(params_.n_cases(), 1.0 / (double)(params_.n_cases() - 2)) { }

	void set_checkpoint_freq(int n) {
		// pass in checkpoint_freq <= 1 for no check-pointing
		checkpoint_freq_ = n;
		this->ClearFwdMsgs();
	}

	void set_case_index(int i) { case_index_ = i; }
	int get_checkpoint_freq() const { return checkpoint_freq_; }
	void set_rescale_freq(int x) { rescale_freq_ = x; }

	virtual double ForwardAlgo();
	virtual void BackwardAlgo();

	const RealVec& ForwardMsg(int snp_idx) {
		if (snp_idx == -1) {
			return(init_fwd_msg_);
		} else if (!this->FwdMsgExists(snp_idx)) {
    	this->FillForwardMsgs(this->FindForwardMsgs(snp_idx), snp_idx);
    }
    return(*this->FwdMsgBuffer(snp_idx));
	}

	virtual ~PhaseEmInference() {
		this->ClearFwdMsgs();
	}

public:
	void ForwardStep(int snp_idx, const RealVec& inp_msg, RealVec * ret_msg) {
		if (snp_idx < 0 || snp_idx >= params_.n_snp()) error("ForwardStep : bad input (snp_idx = %d)\n", snp_idx);
		if ((int)inp_msg.size() != params_.n_cases()) {
			error("ForwardStep : bad input (2) (%d vs %d)\n", inp_msg.size(), params_.n_cases());
		}

		int cur_allele = params_.GetAllele(case_index_, snp_idx);
		params_.InitPttl(inp_msg, snp_idx, case_index_, cur_allele, ret_msg);

		double jp = params_.jump_prob(snp_idx);
		double jump_pr = jp / (double)(params_.n_cases() - 2);
		params_.MakeMessage(*ret_msg, jump_pr, 1.0 - jp, ret_msg);
	}

	void BackwardStep(int snp_idx, const RealVec& inp_msg, RealVec * ret_msg) {
		if (snp_idx < 0 || snp_idx >= params_.n_snp()) error("BackwardStep : bad input.  snp_idx = %d\n", snp_idx);
		if ((int)inp_msg.size() != params_.n_cases()) error("BackwardStep : bad input (2)\n");

		int cur_allele = params_.GetAllele(case_index_, snp_idx);
		double jp = params_.jump_prob(snp_idx);
		double jump_pr = jp / (double)(params_.n_cases() - 2);
		params_.MakeMessage(inp_msg, jump_pr, 1.0 - jp, ret_msg);
		params_.InitPttl(snp_idx, case_index_, cur_allele, ret_msg);
	}

	void BackwardStepSetAllele(int snp_idx, int cur_allele,
			                       const RealVec& inp_msg, RealVec * ret_msg) {
		double jp = params_.jump_prob(snp_idx);
		double jump_pr = jp / (double)(params_.n_cases() - 2);
		params_.MakeMessage(inp_msg, jump_pr, 1.0 - jp, ret_msg);
		params_.InitPttl(snp_idx, case_index_, cur_allele, ret_msg);
	}

protected:
	// deletes all buffers
	void ClearFwdMsgs();

	bool FwdMsgExists(int snp_idx) const {
		return((snp_idx == -1) || (fwd_msgs_[snp_idx] != NULL));
	}

	virtual void BackwardProcess(int cur_allele, int snp_idx, const RealVec& bkwd_msg) {
		// do nothing
	}

	void FillForwardMsgs(int start_idx, int last_idx);
	int FindForwardMsgs(int start_idx);

	// allocates the buffer if it does not exist
	RealVec * FwdMsgBuffer(int snp_idx);

	// indicates whether the forward message should be stored or not
	// This is only false if we are check-pointing
	bool StoreMessage(int snp_idx);
public:
	PhaseEmModel& params_;

	int checkpoint_freq_, rescale_freq_;
	int case_index_;
  vector<RealVec *> fwd_msgs_;
  RealVec init_fwd_msg_;
};

class PhaseEmParamOpt : public PhaseEmInference {
public:
	PhaseEmParamOpt(PhaseEmModel& p);

	virtual void BackwardProcess(int cur_allele, int snp_idx,
			                         const RealVec& bkwd_msg) {
		this->UpdateMutCounts(cur_allele, snp_idx, bkwd_msg);
		this->UpdateJumpCounts(cur_allele, snp_idx, bkwd_msg);
	}

	const RealVec& mut_counts() const { return(mut_counts_); }
	const RealVec& jump_counts() const { return(jump_counts_); }

public:
	void UpdateMutCounts(int cur_allele, int snp_idx, const RealVec& bkwd_msg);
	void UpdateJumpCounts(int cur_allele, int snp_idx, const RealVec& bkwd_msg);

public:
	RealVec mut_counts_, jump_counts_, count_buf_, tmp_msg_;
};

}  // namespace PhaseEM


extern "C" {
SEXP PhaseEmExpCounts_RI(SEXP params_sxp, SEXP cases_sxp, SEXP case_inds_sxp,
				                 SEXP jump_counts_sxp, SEXP mut_counts_sxp,
					               SEXP pliks_sxp);

}  // extern "C" {

#endif