#include "phase_em.hpp"
#include "phase_em_hap_opt.hpp"

namespace PhaseEM {

using namespace Util;

void PhaseEmHapOpt::NoMove(int snp_idx, double move_lik) {
	this->SetMove(snp_idx, false,
			          params_.GetAllele(case_inds_[0], snp_idx),
                params_.GetAllele(case_inds_[1], snp_idx),
                move_lik);
}

void PhaseEmHapOpt::SetMove(int snp_idx, int made_cross, int a0, int a1, double move_lik) {
	if (made_cross) {
	  bkwd_msgs_[0][0].swap(bkwd_msgs_[1][a0]);
	  bkwd_msgs_[1][0].swap(bkwd_msgs_[0][a1]);
	} else {
		if (a0) bkwd_msgs_[0][0].swap(bkwd_msgs_[0][a0]);
		if (a1) bkwd_msgs_[1][0].swap(bkwd_msgs_[1][a1]);
	}

	opt_alleles_[snp_idx].first = a0;
	opt_alleles_[snp_idx].second = a1;

	made_cross_[snp_idx] = made_cross;
	move_liks_[snp_idx] = move_lik;
}

double PhaseEmHapOpt::LikRatio(int snp_idx) {
	const RealVec& fwd_m0 = inf_obj_[0]->ForwardMsg(snp_idx - 1);
	const RealVec& fwd_m1 = inf_obj_[1]->ForwardMsg(snp_idx - 1);

	int gtype = gtype_codes_.GetElt(case_inds_[0], snp_idx);

	if (!allow_crossover_ && (gtype == 1 || gtype == 2 || gtype == 4 || gtype == 8)) {
		return(0.0);
	}

	int cur_code = params_.GetAllele(case_inds_[0], snp_idx)
	               + 2 * params_.GetAllele(case_inds_[1], snp_idx);

	if (!allow_single_site_) {
		gtype = (1 << cur_code);
	}

	double liks[2][4] = {{R_NegInf, R_NegInf, R_NegInf, R_NegInf},
		                   {R_NegInf, R_NegInf, R_NegInf, R_NegInf}};
  int mx_idx1 = -1, mx_idx2 = -1;
  double mx_lik = R_NegInf;

	for (int gi = 0, bi = 1; gi < 4; ++gi, bi <<= 1) {
		if (gtype & bi) {
			int a0 = gi % 2;
			int a1 = gi / 2;
			liks[0][gi] = this->Lik(bkwd_msgs_[0][a0], fwd_m0) + this->Lik(bkwd_msgs_[1][a1], fwd_m1);

			if (liks[0][gi] > mx_lik) {
				mx_lik = liks[0][gi];
				mx_idx1 = 0;
				mx_idx2 = gi;
			}

			if (allow_crossover_) {
				liks[1][gi] = this->Lik(bkwd_msgs_[1][a0], fwd_m0) + this->Lik(bkwd_msgs_[0][a1], fwd_m1);

				if (liks[1][gi] > mx_lik) {
					mx_lik = liks[1][gi];
					mx_idx1 = 1;
					mx_idx2 = gi;
				}
			}
		}
	}

	double delt = mx_lik - liks[0][cur_code];
	return(delt);
}

bool PhaseEmHapOpt::MakeMove(int * snp_idx) {
	int si = *snp_idx;
	const RealVec& fwd_m0 = inf_obj_[0]->ForwardMsg(si - 1);
	const RealVec& fwd_m1 = inf_obj_[1]->ForwardMsg(si - 1);

	int gtype = gtype_codes_.GetElt(case_inds_[0], si);

	if (!allow_crossover_ && (gtype == 1 || gtype == 2 || gtype == 4 || gtype == 8)) {
		// only a single phased genotype is allowed and we cannot make a crossover,
		// so we do not evaluate any moves
		this->NoMove(si, 0.0);
  	return(false);
	}

	int cur_code = params_.GetAllele(case_inds_[0], si)
	               + 2 * params_.GetAllele(case_inds_[1], si);

	if (!((1 << cur_code) & gtype)) {
		error("genotype codes do not match input phasing (gtype = %d, a0 = %d, a1 = %d)\n",
				  gtype, params_.GetAllele(case_inds_[0], si), params_.GetAllele(case_inds_[1], si));
	}

	if (!allow_single_site_) {
		gtype = (1 << cur_code);
	}

	double liks[2][4] = {{R_NegInf, R_NegInf, R_NegInf, R_NegInf},
		                   {R_NegInf, R_NegInf, R_NegInf, R_NegInf}};
  int mx_idx1 = -1, mx_idx2 = -1;
  double mx_lik = R_NegInf;

	for (int gi = 0, bi = 1; gi < 4; ++gi, bi <<= 1) {
		if (gtype & bi) {
			int a0 = gi % 2;
			int a1 = gi / 2;
			liks[0][gi] = this->Lik(bkwd_msgs_[0][a0], fwd_m0) + this->Lik(bkwd_msgs_[1][a1], fwd_m1);

			if (liks[0][gi] > mx_lik) {
				mx_lik = liks[0][gi];
				mx_idx1 = 0;
				mx_idx2 = gi;
			}

			if (allow_crossover_) {
				liks[1][gi] = this->Lik(bkwd_msgs_[1][a0], fwd_m0) + this->Lik(bkwd_msgs_[0][a1], fwd_m1);

				if (liks[1][gi] > mx_lik) {
					mx_lik = liks[1][gi];
					mx_idx1 = 1;
					mx_idx2 = gi;
				}
			}
		}
	}

	double delt = mx_lik - liks[0][cur_code];
	//Rprintf("delt = %g, cur_code = %d, new code = %d, crossover = %d\n", delt, cur_code, mx_idx2, mx_idx1);

  if (delt > move_thresh_) {
  	this->SetMove(si, (mx_idx1 == 1), mx_idx2 % 2, mx_idx2 / 2, delt);
  	return(true);
  } else {
  	this->NoMove(si, delt);
  	return(false);
  }
}

void PhaseEmMakeGtypeCodes(const CastableMatI& allele_mat, CastableMatI * gtypes) {
	if (!allele_mat.EqualDims(*gtypes)) {
		error("PhaseEmMakeGtypeCodes : dim(gtypes) != dim(allele_mat)\n");
		return;
	}

	int nr = allele_mat.nrow();
	int nc = allele_mat.ncol();

	for (int ri = 0; ri < nr; ri += 2) {
		for (int ci = 0; ci < nc; ++ci) {
			int c0 = allele_mat.GetElt(ri, ci) + 2 * allele_mat.GetElt(ri + 1, ci);
			int c1 = allele_mat.GetElt(ri + 1, ci) + 2 * allele_mat.GetElt(ri, ci);
			int cur_code = ((1 << c0) | (1 << c1));
			gtypes->SetElt(ri, ci, cur_code);
			gtypes->SetElt(ri + 1, ci, cur_code);
		}
	}
}

void PhaseEmSetMissGtypeCodes(const CastableMatI& miss_mat, CastableMatI * gtypes) {
	if (!miss_mat.EqualDims(*gtypes)) {
		error("PhaseEmMakeGtypeCodes : dim(gtypes) != dim(allele_mat)\n");
		return;
	}

	int nr = miss_mat.nrow();
	int nc = miss_mat.ncol();
  int c0 = 0, c1 = (1 | 2 | 4 | 8);

	for (int ri = 0; ri < nr; ri += 2) {
		for (int ci = 0; ci < nc; ++ci) {
			int cur_code = (miss_mat.GetElt(ri, ci) || miss_mat.GetElt(ri + 1, ci))
			               ? c1 : c0;
			gtypes->SetElt(ri, ci, cur_code);
			gtypes->SetElt(ri + 1, ci, cur_code);
		}
	}
}

PhaseEmHapOpt * HapOptFactory(PhaseEmModel& p, const CastableMatI& gtype_codes,
		                          bool back_tracking) {
	if (back_tracking) {
		return(new BacktrackHapOpt(p, gtype_codes));
	} else {
		return(new PhaseEmHapOpt(p, gtype_codes));
	}
}

}  // namespace PhaseEM {


extern "C" {

SEXP PhaseEmGenotypeCodes_RI(SEXP allele_mat_sxp, SEXP gtype_mat_sxp) {
	using namespace PhaseEM;
	using namespace Util;

	PointerDeleter<CastableMatI> allele_mat(OwningCastableMatFactory(allele_mat_sxp));
	PointerDeleter<CastableMatI> gtypes(OwningCastableMatFactory(gtype_mat_sxp));
	PhaseEmMakeGtypeCodes(*allele_mat.get(), gtypes.get());

	return(gtype_mat_sxp);
}

SEXP PhaseEmSetMissGenotypeCodes_RI(SEXP miss_mat_sxp, SEXP gtype_mat_sxp) {
	using namespace PhaseEM;
	using namespace Util;

	PointerDeleter<CastableMatI> miss_mat(OwningCastableMatFactory(miss_mat_sxp));
	PointerDeleter<CastableMatI> gtypes(OwningCastableMatFactory(gtype_mat_sxp));
	PhaseEmSetMissGtypeCodes(*miss_mat.get(), gtypes.get());

	return(gtype_mat_sxp);
}

SEXP PhaseEmHapOptScores_RI(SEXP params_sxp, SEXP cases_sxp,
					                  SEXP case_idx1_sxp, SEXP case_idx2_sxp,
					                  SEXP allow_crossover_sxp,
					                  SEXP gtype_codes_sxp) {
	using namespace PhaseEM;
	using namespace Util;

	PointerDeleter<PhaseEmModel> mdl(PhaseEmModelFactory(params_sxp, cases_sxp));
	PointerDeleter<CastableMatI> gtypes(OwningCastableMatFactory(gtype_codes_sxp));
	PhaseEmHapOpt opt_obj(*mdl.get(), *gtypes.get());
  opt_obj.set_move_thresh(R_PosInf);

	int case_idx1 = GetInt(case_idx1_sxp, -1, NULL);
	int case_idx2 = GetInt(case_idx2_sxp, -1, NULL);

	opt_obj.set_case_inds(case_idx1, case_idx2);
	opt_obj.ForwardAlgo();
	opt_obj.BackwardAlgo();

	const RealVec& move_liks = opt_obj.move_liks();
	SEXP ret_sxp = NEW_NUMERIC(move_liks.size());
	double * mv_lks = REAL(ret_sxp);
	for (int i = 0, sz = mdl.get()->n_snp(); i < sz; ++i) {
		mv_lks[i] = move_liks[i];
	}

	return(ret_sxp);
}

SEXP PhaseEmHapOpt_RI(SEXP params_sxp, SEXP cases_sxp,
					            SEXP case_inds_sxp, SEXP move_thresh_sxp,
					            SEXP allow_crossover_sxp,
					            SEXP allow_single_site_sxp,
					            SEXP gtype_codes_sxp,
					            SEXP back_tracking_sxp) {
	using namespace PhaseEM;
	using namespace Util;

	bool back_tracking = (GetInt(back_tracking_sxp, 0, NULL) != 0);
	double move_thresh = GetNumeric(move_thresh_sxp, 0.0, NULL);
  bool allow_crossover = (GetInt(allow_crossover_sxp, 0, NULL) != 0);
  bool allow_single_site = (GetInt(allow_single_site_sxp, 0, NULL) != 0);

	IntVec case_inds;
	SxpToIntVec(case_inds_sxp, &case_inds);
	if (case_inds.size() % 2 != 0) {
		error("case_inds_sxp must be a matrix with 2 rows\n");
	}

	PointerDeleter<PhaseEmModel> mdl(PhaseEmModelFactory(params_sxp, cases_sxp));
	PointerDeleter<CastableMatI> gtypes(OwningCastableMatFactory(gtype_codes_sxp));

	PhaseEmHapOpt * opt_obj = HapOptFactory(*mdl.get(), *gtypes.get(), back_tracking);
	PointerDeleter<PhaseEmHapOpt> opt_obj_deleter(opt_obj);

  opt_obj->set_move_thresh(move_thresh);
  opt_obj->set_allow_crossover(allow_crossover);
  opt_obj->set_allow_single_site(allow_single_site);

	for(int i = 0, ns = case_inds.size(); i < ns; i += 2) {
		opt_obj->set_case_inds(case_inds[i], case_inds[i + 1]);
		opt_obj->ForwardAlgo();
		opt_obj->BackwardAlgo();
		opt_obj->ApplyMoves();

		R_CheckUserInterrupt();
	}

	return(R_NilValue);
}

}  // extern "C"
